<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/9/bcm/dma.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/9/bcm/dma.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * bcm2835 dma controller
 *
 * simplest to use only channels 0-6
 *	channels 7-14 have reduced functionality
 *	channel 15 is at a weird address
 *	channels 0 and 15 have an "external 128 bit 8 word read FIFO"
 *	  for memory to memory transfers
 *
 * Experiments show that only channels 2-5,11-12 work with mmc
 */

#include "u.h"
#include "../port/lib.h"
#include "../port/error.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"

#define DMAREGS	(VIRTIO+0x7000)

#define DBG	if(Dbg)

enum {
	Nchan		= 7,		/* number of dma channels */
	Regsize		= 0x100,	/* size of regs for each chan */
	Cbalign		= 32,		/* control block byte alignment */
	Dbg		= 0,
	
	/* registers for each dma controller */
	Cs		= 0x00&gt;&gt;2,
	Conblkad	= 0x04&gt;&gt;2,
	Ti		= 0x08&gt;&gt;2,
	Sourcead	= 0x0c&gt;&gt;2,
	Destad		= 0x10&gt;&gt;2,
	Txfrlen		= 0x14&gt;&gt;2,
	Stride		= 0x18&gt;&gt;2,
	Nextconbk	= 0x1c&gt;&gt;2,
	Debug		= 0x20&gt;&gt;2,

	/* collective registers */
	Intstatus	= 0xfe0&gt;&gt;2,
	Enable		= 0xff0&gt;&gt;2,

	/* Cs */
	Reset		= 1&lt;&lt;31,
	Abort		= 1&lt;&lt;30,
	Error		= 1&lt;&lt;8,
	Waitwrite	= 1&lt;&lt;6,
	Waitdreq	= 1&lt;&lt;5,
	Paused		= 1&lt;&lt;4,
	Dreq		= 1&lt;&lt;3,
	Int		= 1&lt;&lt;2,
	End		= 1&lt;&lt;1,
	Active		= 1&lt;&lt;0,

	/* Ti */
	Permapshift= 16,
	Srcignore	= 1&lt;&lt;11,
	Srcdreq		= 1&lt;&lt;10,
	Srcwidth128	= 1&lt;&lt;9,
	Srcinc		= 1&lt;&lt;8,
	Destignore	= 1&lt;&lt;7,
	Destdreq	= 1&lt;&lt;6,
	Destwidth128	= 1&lt;&lt;5,
	Destinc		= 1&lt;&lt;4,
	Waitresp	= 1&lt;&lt;3,
	Tdmode		= 1&lt;&lt;1,
	Inten		= 1&lt;&lt;0,

	/* Debug */
	Lite		= 1&lt;&lt;28,
	Clrerrors	= 7&lt;&lt;0,
};

typedef struct Ctlr Ctlr;
typedef struct Cb Cb;

struct Ctlr {
	u32int	*regs;
	Cb	*cb;
	Rendez	r;
	int	dmadone;
};

struct Cb {
	u32int	ti;
	u32int	sourcead;
	u32int	destad;
	u32int	txfrlen;
	u32int	stride;
	u32int	nextconbk;
	u32int	reserved[2];
};

static Ctlr dma[Nchan];
static u32int *dmaregs = (u32int*)DMAREGS;

static void
dump(char *msg, uchar *p, int n)
{
	print("%s", msg);
	while(n-- &gt; 0)
		print(" %2.2x", *p++);
	print("\n");
}

static void
dumpdregs(char *msg, u32int *r)
{
	int i;

	print("%s: %#p =", msg, r);
	for(i = 0; i &lt; 9; i++)
		print(" %8.8uX", r[i]);
	print("\n");
}

static int
dmadone(void *a)
{
	return ((Ctlr*)a)-&gt;dmadone;
}

static void
dmainterrupt(Ureg*, void *a)
{
	Ctlr *ctlr;

	ctlr = a;
	ctlr-&gt;regs[Cs] = Int;
	ctlr-&gt;dmadone = 1;
	wakeup(&amp;ctlr-&gt;r);
}

void
dmastart(int chan, int dev, int dir, void *src, void *dst, int len)
{
	Ctlr *ctlr;
	Cb *cb;
	int ti;

	ctlr = &amp;dma[chan];
	if(ctlr-&gt;regs == nil){
		ctlr-&gt;regs = (u32int*)(DMAREGS + chan*Regsize);
		ctlr-&gt;cb = xspanalloc(sizeof(Cb), Cbalign, 0);
		assert(ctlr-&gt;cb != nil);
		dmaregs[Enable] |= 1&lt;&lt;chan;
		ctlr-&gt;regs[Cs] = Reset;
		while(ctlr-&gt;regs[Cs] &amp; Reset)
			;
		intrenable(IRQDMA(chan), dmainterrupt, ctlr, 0, "dma");
	}
	cb = ctlr-&gt;cb;
	ti = 0;
	switch(dir){
	case DmaD2M:
		cachedwbinvse(dst, len);
		ti = Srcdreq | Destinc;
		cb-&gt;sourcead = DMAIO(src);
		cb-&gt;destad = DMAADDR(dst);
		break;
	case DmaM2D:
		cachedwbse(src, len);
		ti = Destdreq | Srcinc;
		cb-&gt;sourcead = DMAADDR(src);
		cb-&gt;destad = DMAIO(dst);
		break;
	case DmaM2M:
		cachedwbse(src, len);
		cachedwbinvse(dst, len);
		ti = Srcinc | Destinc;
		cb-&gt;sourcead = DMAADDR(src);
		cb-&gt;destad = DMAADDR(dst);
		break;
	}
	cb-&gt;ti = ti | dev&lt;&lt;Permapshift | Inten;
	cb-&gt;txfrlen = len;
	cb-&gt;stride = 0;
	cb-&gt;nextconbk = 0;
	cachedwbse(cb, sizeof(Cb));
	ctlr-&gt;regs[Cs] = 0;
	microdelay(1);
	ctlr-&gt;regs[Conblkad] = DMAADDR(cb);
	DBG print("dma start: %ux %ux %ux %ux %ux %ux\n",
		cb-&gt;ti, cb-&gt;sourcead, cb-&gt;destad, cb-&gt;txfrlen,
		cb-&gt;stride, cb-&gt;nextconbk);
	DBG print("intstatus %ux\n", dmaregs[Intstatus]);
	dmaregs[Intstatus] = 0;
	ctlr-&gt;regs[Cs] = Int;
	microdelay(1);
	coherence();
	DBG dumpdregs("before Active", ctlr-&gt;regs);
	ctlr-&gt;regs[Cs] = Active;
	DBG dumpdregs("after Active", ctlr-&gt;regs);
}

int
dmawait(int chan)
{
	Ctlr *ctlr;
	u32int *r;
	int s;

	ctlr = &amp;dma[chan];
	tsleep(&amp;ctlr-&gt;r, dmadone, ctlr, 3000);
	ctlr-&gt;dmadone = 0;
	r = ctlr-&gt;regs;
	DBG dumpdregs("after sleep", r);
	s = r[Cs];
	if((s &amp; (Active|End|Error)) != End){
		print("dma chan %d %s Cs %ux Debug %ux\n", chan,
			(s&amp;End)? "error" : "timeout", s, r[Debug]);
		r[Cs] = Reset;
		r[Debug] = Clrerrors;
		return -1;
	}
	r[Cs] = Int|End;
	return 0;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

