<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/9/bcm/trap.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/9/bcm/trap.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * traps, exceptions, interrupts, system calls.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "ureg.h"
#include "../port/error.h"

#include "arm.h"

#define INTREGS		(VIRTIO+0xB200)

typedef struct Intregs Intregs;
typedef struct Vctl Vctl;

enum {
	Debug = 0,

	Nvec = 8,		/* # of vectors at start of lexception.s */
	Fiqenable = 1&lt;&lt;7,
};

/*
 *   Layout at virtual address KZERO (double mapped at HVECTORS).
 */
typedef struct Vpage0 {
	void	(*vectors[Nvec])(void);
	u32int	vtable[Nvec];
} Vpage0;

/*
 * interrupt control registers
 */
struct Intregs {
	u32int	ARMpending;
	u32int	GPUpending[2];
	u32int	FIQctl;
	u32int	GPUenable[2];
	u32int	ARMenable;
	u32int	GPUdisable[2];
	u32int	ARMdisable;
};

struct Vctl {
	Vctl	*next;
	int	irq;
	u32int	*reg;
	u32int	mask;
	void	(*f)(Ureg*, void*);
	void	*a;
};

static Vctl *vctl, *vfiq;

static char *trapnames[PsrMask+1] = {
	[ PsrMusr ] "user mode",
	[ PsrMfiq ] "fiq interrupt",
	[ PsrMirq ] "irq interrupt",
	[ PsrMsvc ] "svc/swi exception",
	[ PsrMabt ] "prefetch abort/data abort",
	[ PsrMabt+1 ] "data abort",
	[ PsrMund ] "undefined instruction",
	[ PsrMsys ] "sys trap",
};

extern int notify(Ureg*);

/*
 *  set up for exceptions
 */
void
trapinit(void)
{
	Vpage0 *vpage0;

	/* disable everything */
	intrsoff();

	/* set up the exception vectors */
	vpage0 = (Vpage0*)HVECTORS;
	memmove(vpage0-&gt;vectors, vectors, sizeof(vpage0-&gt;vectors));
	memmove(vpage0-&gt;vtable, vtable, sizeof(vpage0-&gt;vtable));
	cacheuwbinv();

	/* set up the stacks for the interrupt modes */
	setr13(PsrMfiq, (u32int*)(FIQSTKTOP));
	setr13(PsrMirq, m-&gt;sirq);
	setr13(PsrMabt, m-&gt;sabt);
	setr13(PsrMund, m-&gt;sund);
	setr13(PsrMsys, m-&gt;ssys);

	coherence();
}

void
intrsoff(void)
{
	Intregs *ip;
	int disable;

	ip = (Intregs*)INTREGS;
	disable = ~0;
	ip-&gt;GPUdisable[0] = disable;
	ip-&gt;GPUdisable[1] = disable;
	ip-&gt;ARMdisable = disable;
	ip-&gt;FIQctl = 0;
}

/*
 *  called by trap to handle irq interrupts.
 *  returns true iff a clock interrupt, thus maybe reschedule.
 */
static int
irq(Ureg* ureg)
{
	Vctl *v;
	int clockintr;

	clockintr = 0;
	for(v = vctl; v; v = v-&gt;next)
		if(*v-&gt;reg &amp; v-&gt;mask){
			coherence();
			v-&gt;f(ureg, v-&gt;a);
			coherence();
			if(v-&gt;irq == IRQclock)
				clockintr = 1;
		}
	return clockintr;
}

/*
 * called direct from lexception.s to handle fiq interrupt.
 */
void
fiq(Ureg *ureg)
{
	Vctl *v;

	v = vfiq;
	if(v == nil)
		panic("unexpected item in bagging area");
	m-&gt;intr++;
	ureg-&gt;pc -= 4;
	coherence();
	v-&gt;f(ureg, v-&gt;a);
	coherence();
}

void
irqenable(int irq, void (*f)(Ureg*, void*), void* a)
{
	Vctl *v;
	Intregs *ip;
	u32int *enable;

	ip = (Intregs*)INTREGS;
	v = (Vctl*)malloc(sizeof(Vctl));
	if(v == nil)
		panic("irqenable: no mem");
	v-&gt;irq = irq;
	if(irq &gt;= IRQbasic){
		enable = &amp;ip-&gt;ARMenable;
		v-&gt;reg = &amp;ip-&gt;ARMpending;
		v-&gt;mask = 1 &lt;&lt; (irq - IRQbasic);
	}else{
		enable = &amp;ip-&gt;GPUenable[irq/32];
		v-&gt;reg = &amp;ip-&gt;GPUpending[irq/32];
		v-&gt;mask = 1 &lt;&lt; (irq % 32);
	}
	v-&gt;f = f;
	v-&gt;a = a;
	if(irq == IRQfiq){
		assert((ip-&gt;FIQctl &amp; Fiqenable) == 0);
		assert((*enable &amp; v-&gt;mask) == 0);
		vfiq = v;
		ip-&gt;FIQctl = Fiqenable | irq;
	}else{
		v-&gt;next = vctl;
		vctl = v;
		*enable = v-&gt;mask;
	}
}

static char *
trapname(int psr)
{
	char *s;

	s = trapnames[psr &amp; PsrMask];
	if(s == nil)
		s = "unknown trap number in psr";
	return s;
}

/* this is quite helpful during mmu and cache debugging */
static void
ckfaultstuck(uintptr va)
{
	static int cnt, lastpid;
	static uintptr lastva;

	if (va == lastva &amp;&amp; up-&gt;pid == lastpid) {
		++cnt;
		if (cnt &gt;= 2)
			/* fault() isn't fixing the underlying cause */
			panic("fault: %d consecutive faults for va %#p",
				cnt+1, va);
	} else {
		cnt = 0;
		lastva = va;
		lastpid = up-&gt;pid;
	}
}

/*
 *  called by trap to handle access faults
 */
static void
faultarm(Ureg *ureg, uintptr va, int user, int read)
{
	int n, insyscall;
	char buf[ERRMAX];

	if(up == nil) {
		dumpregs(ureg);
		panic("fault: nil up in faultarm, accessing %#p", va);
	}
	insyscall = up-&gt;insyscall;
	up-&gt;insyscall = 1;
	if (Debug)
		ckfaultstuck(va);

	n = fault(va, read);
	if(n &lt; 0){
		if(!user){
			dumpregs(ureg);
			panic("fault: kernel accessing %#p", va);
		}
		/* don't dump registers; programs suicide all the time */
		snprint(buf, sizeof buf, "sys: trap: fault %s va=%#p",
			read? "read": "write", va);
		postnote(up, 1, buf, NDebug);
	}
	up-&gt;insyscall = insyscall;
}

/*
 *  returns 1 if the instruction writes memory, 0 otherwise
 */
int
writetomem(ulong inst)
{
	/* swap always write memory */
	if((inst &amp; 0x0FC00000) == 0x01000000)
		return 1;

	/* loads and stores are distinguished by bit 20 */
	if(inst &amp; (1&lt;&lt;20))
		return 0;

	return 1;
}

/*
 *  here on all exceptions other than syscall (SWI) and fiq
 */
void
trap(Ureg *ureg)
{
	int clockintr, user, x, rv, rem;
	ulong inst, fsr;
	uintptr va;
	char buf[ERRMAX];

	assert(!islo());
	if(up != nil)
		rem = ((char*)ureg)-up-&gt;kstack;
	else
		rem = ((char*)ureg)-((char*)m+sizeof(Mach));
	if(rem &lt; 256) {
		iprint("trap: %d stack bytes left, up %#p ureg %#p at pc %#lux\n",
			rem, up, ureg, ureg-&gt;pc);
		delay(1000);
		dumpstack();
		panic("trap: %d stack bytes left, up %#p ureg %#p at pc %#lux",
			rem, up, ureg, ureg-&gt;pc);
	}

	user = (ureg-&gt;psr &amp; PsrMask) == PsrMusr;
	if(user){
		up-&gt;dbgreg = ureg;
		cycles(&amp;up-&gt;kentry);
	}

	/*
	 * All interrupts/exceptions should be resumed at ureg-&gt;pc-4,
	 * except for Data Abort which resumes at ureg-&gt;pc-8.
	 */
	if(ureg-&gt;type == (PsrMabt+1))
		ureg-&gt;pc -= 8;
	else
		ureg-&gt;pc -= 4;

	clockintr = 0;		/* if set, may call sched() before return */
	switch(ureg-&gt;type){
	default:
		panic("unknown trap; type %#lux, psr mode %#lux", ureg-&gt;type,
			ureg-&gt;psr &amp; PsrMask);
		break;
	case PsrMirq:
		clockintr = irq(ureg);
		m-&gt;intr++;
		break;
	case PsrMabt:			/* prefetch fault */
		x = ifsrget();
		fsr = (x&gt;&gt;7) &amp; 0x8 | x &amp; 0x7;
		switch(fsr){
		case 0x02:		/* instruction debug event (BKPT) */
			if(user){
				snprint(buf, sizeof buf, "sys: breakpoint");
				postnote(up, 1, buf, NDebug);
			}else{
				iprint("kernel bkpt: pc %#lux inst %#ux\n",
					ureg-&gt;pc, *(u32int*)ureg-&gt;pc);
				panic("kernel bkpt");
			}
			break;
		default:
			faultarm(ureg, ureg-&gt;pc, user, 1);
			break;
		}
		break;
	case PsrMabt+1:			/* data fault */
		va = farget();
		inst = *(ulong*)(ureg-&gt;pc);
		/* bits 12 and 10 have to be concatenated with status */
		x = fsrget();
		fsr = (x&gt;&gt;7) &amp; 0x20 | (x&gt;&gt;6) &amp; 0x10 | x &amp; 0xf;
		switch(fsr){
		default:
		case 0xa:		/* ? was under external abort */
			panic("unknown data fault, 6b fsr %#lux", fsr);
			break;
		case 0x0:
			panic("vector exception at %#lux", ureg-&gt;pc);
			break;
		case 0x1:		/* alignment fault */
		case 0x3:		/* access flag fault (section) */
			if(user){
				snprint(buf, sizeof buf,
					"sys: alignment: pc %#lux va %#p\n",
					ureg-&gt;pc, va);
				postnote(up, 1, buf, NDebug);
			} else
				panic("kernel alignment: pc %#lux va %#p", ureg-&gt;pc, va);
			break;
		case 0x2:
			panic("terminal exception at %#lux", ureg-&gt;pc);
			break;
		case 0x4:		/* icache maint fault */
		case 0x6:		/* access flag fault (page) */
		case 0x8:		/* precise external abort, non-xlat'n */
		case 0x28:
		case 0xc:		/* l1 translation, precise ext. abort */
		case 0x2c:
		case 0xe:		/* l2 translation, precise ext. abort */
		case 0x2e:
		case 0x16:		/* imprecise ext. abort, non-xlt'n */
		case 0x36:
			panic("external abort %#lux pc %#lux addr %#p",
				fsr, ureg-&gt;pc, va);
			break;
		case 0x1c:		/* l1 translation, precise parity err */
		case 0x1e:		/* l2 translation, precise parity err */
		case 0x18:		/* imprecise parity or ecc err */
			panic("translation parity error %#lux pc %#lux addr %#p",
				fsr, ureg-&gt;pc, va);
			break;
		case 0x5:		/* translation fault, no section entry */
		case 0x7:		/* translation fault, no page entry */
			faultarm(ureg, va, user, !writetomem(inst));
			break;
		case 0x9:
		case 0xb:
			/* domain fault, accessing something we shouldn't */
			if(user){
				snprint(buf, sizeof buf,
					"sys: access violation: pc %#lux va %#p\n",
					ureg-&gt;pc, va);
				postnote(up, 1, buf, NDebug);
			} else
				panic("kernel access violation: pc %#lux va %#p",
					ureg-&gt;pc, va);
			break;
		case 0xd:
		case 0xf:
			/* permission error, copy on write or real permission error */
			faultarm(ureg, va, user, !writetomem(inst));
			break;
		}
		break;
	case PsrMund:			/* undefined instruction */
		if(user){
			if(seg(up, ureg-&gt;pc, 0) != nil &amp;&amp;
			   *(u32int*)ureg-&gt;pc == 0xD1200070)
				postnote(up, 1, "sys: breakpoint", NDebug);
			else{
				/* look for floating point instructions to interpret */
				rv = fpuemu(ureg);
				if(rv == 0){
					snprint(buf, sizeof buf,
						"undefined instruction: pc %#lux\n",
						ureg-&gt;pc);
					postnote(up, 1, buf, NDebug);
				}
			}
		}else{
			if (ureg-&gt;pc &amp; 3) {
				iprint("rounding fault pc %#lux down to word\n",
					ureg-&gt;pc);
				ureg-&gt;pc &amp;= ~3;
			}
			iprint("undefined instruction: pc %#lux inst %#ux\n",
				ureg-&gt;pc, *(u32int*)ureg-&gt;pc);
			panic("undefined instruction");
		}
		break;
	}
	splhi();

	/* delaysched set because we held a lock or because our quantum ended */
	if(up &amp;&amp; up-&gt;delaysched &amp;&amp; clockintr){
		sched();		/* can cause more traps */
		splhi();
	}

	if(user){
		if(up-&gt;procctl || up-&gt;nnote)
			notify(ureg);
		kexit(ureg);
	}
}

int
isvalidaddr(void *v)
{
	return (uintptr)v &gt;= KZERO;
}

static void
dumplongs(char *msg, ulong *v, int n)
{
	int i, l;

	l = 0;
	iprint("%s at %.8p: ", msg, v);
	for(i=0; i&lt;n; i++){
		if(l &gt;= 4){
			iprint("\n    %.8p: ", v);
			l = 0;
		}
		if(isvalidaddr(v)){
			iprint(" %.8lux", *v++);
			l++;
		}else{
			iprint(" invalid");
			break;
		}
	}
	iprint("\n");
}

static void
dumpstackwithureg(Ureg *ureg)
{
	uintptr l, i, v, estack;
	u32int *p;
	char *s;

	if((s = getconf("*nodumpstack")) != nil &amp;&amp; strcmp(s, "0") != 0){
		iprint("dumpstack disabled\n");
		return;
	}
	iprint("ktrace /kernel/path %#.8lux %#.8lux %#.8lux # pc, sp, link\n",
		ureg-&gt;pc, ureg-&gt;sp, ureg-&gt;r14);
	delay(2000);
	i = 0;
	if(up != nil &amp;&amp; (uintptr)&amp;l &lt;= (uintptr)up-&gt;kstack+KSTACK)
		estack = (uintptr)up-&gt;kstack+KSTACK;
	else if((uintptr)&amp;l &gt;= (uintptr)m-&gt;stack
	     &amp;&amp; (uintptr)&amp;l &lt;= (uintptr)m+MACHSIZE)
		estack = (uintptr)m+MACHSIZE;
	else{
		if(up != nil)
			iprint("&amp;up-&gt;kstack %#p &amp;l %#p\n", up-&gt;kstack, &amp;l);
		else
			iprint("&amp;m %#p &amp;l %#p\n", m, &amp;l);
		return;
	}
	for(l = (uintptr)&amp;l; l &lt; estack; l += sizeof(uintptr)){
		v = *(uintptr*)l;
		if(KTZERO &lt; v &amp;&amp; v &lt; (uintptr)etext &amp;&amp; !(v &amp; 3)){
			v -= sizeof(u32int);		/* back up an instr */
			p = (u32int*)v;
			if((*p &amp; 0x0f000000) == 0x0b000000){	/* BL instr? */
				iprint("%#8.8lux=%#8.8lux ", l, v);
				i++;
			}
		}
		if(i == 4){
			i = 0;
			iprint("\n");
		}
	}
	if(i)
		iprint("\n");
}

/*
 * Fill in enough of Ureg to get a stack trace, and call a function.
 * Used by debugging interface rdb.
 */
void
callwithureg(void (*fn)(Ureg*))
{
	Ureg ureg;

	ureg.pc = getcallerpc(&amp;fn);
	ureg.sp = PTR2UINT(&amp;fn);
	fn(&amp;ureg);
}

void
dumpstack(void)
{
	callwithureg(dumpstackwithureg);
}

void
dumpregs(Ureg* ureg)
{
	int s;

	if (ureg == nil) {
		iprint("trap: no user process\n");
		return;
	}
	s = splhi();
	iprint("trap: %s", trapname(ureg-&gt;type));
	if(ureg != nil &amp;&amp; (ureg-&gt;psr &amp; PsrMask) != PsrMsvc)
		iprint(" in %s", trapname(ureg-&gt;psr));
	iprint("\n");
	iprint("psr %8.8lux type %2.2lux pc %8.8lux link %8.8lux\n",
		ureg-&gt;psr, ureg-&gt;type, ureg-&gt;pc, ureg-&gt;link);
	iprint("R14 %8.8lux R13 %8.8lux R12 %8.8lux R11 %8.8lux R10 %8.8lux\n",
		ureg-&gt;r14, ureg-&gt;r13, ureg-&gt;r12, ureg-&gt;r11, ureg-&gt;r10);
	iprint("R9  %8.8lux R8  %8.8lux R7  %8.8lux R6  %8.8lux R5  %8.8lux\n",
		ureg-&gt;r9, ureg-&gt;r8, ureg-&gt;r7, ureg-&gt;r6, ureg-&gt;r5);
	iprint("R4  %8.8lux R3  %8.8lux R2  %8.8lux R1  %8.8lux R0  %8.8lux\n",
		ureg-&gt;r4, ureg-&gt;r3, ureg-&gt;r2, ureg-&gt;r1, ureg-&gt;r0);
	iprint("stack is at %#p\n", ureg);
	iprint("pc %#lux link %#lux\n", ureg-&gt;pc, ureg-&gt;link);

	if(up)
		iprint("user stack: %#p-%#p\n", up-&gt;kstack, up-&gt;kstack+KSTACK-4);
	else
		iprint("kernel stack: %8.8lux-%8.8lux\n",
			(ulong)(m+1), (ulong)m+BY2PG-4);
	dumplongs("stack", (ulong *)(ureg + 1), 16);
	delay(2000);
	dumpstack();
	splx(s);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

