sed: Can't open /usr/web/plan9/sources.html
/*
 *  Reliable User Datagram Protocol, currently only for IPv4.
 *  This protocol is compatible with UDP's packet format.
 *  It could be done over UDP if need be.
 */
#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"../port/error.h"

#include	"ip.h"

#define DEBUG	0
#define DPRINT if(DEBUG)print

#define SEQDIFF(a,b) ( (a)&gt;=(b)?\
			(a)-(b):\
			0xffffffffUL-((b)-(a)) )
#define INSEQ(a,start,end) ( (start)&lt;=(end)?\
				((a)&gt;(start)&amp;&amp;(a)&lt;=(end)):\
				((a)&gt;(start)||(a)&lt;=(end)) )
#define UNACKED(r) SEQDIFF(r-&gt;sndseq, r-&gt;ackrcvd)
#define NEXTSEQ(a) ( (a)+1 == 0 ? 1 : (a)+1 )

enum
{
	UDP_PHDRSIZE	= 12,	/* pseudo header */
//	UDP_HDRSIZE	= 20,	/* pseudo header + udp header */
	UDP_RHDRSIZE	= 36,	/* pseudo header + udp header + rudp header */
	UDP_IPHDR	= 8,	/* ip header */
	IP_UDPPROTO	= 254,
	UDP_USEAD7	= 52,	/* size of new ipv6 headers struct */

	Rudprxms	= 200,
	Rudptickms	= 50,
	Rudpmaxxmit	= 10,
	Maxunacked	= 100,
};

#define Hangupgen	0xffffffff	/* used only in hangup messages */

typedef struct Udphdr Udphdr;
struct Udphdr
{
	/* ip header */
	uchar	vihl;		/* Version and header length */
	uchar	tos;		/* Type of service */
	uchar	length[2];	/* packet length */
	uchar	id[2];		/* Identification */
	uchar	frag[2];	/* Fragment information */

	/* pseudo header starts here */
	uchar	Unused;
	uchar	udpproto;	/* Protocol */
	uchar	udpplen[2];	/* Header plus data length */
	uchar	udpsrc[4];	/* Ip source */
	uchar	udpdst[4];	/* Ip destination */

	/* udp header */
	uchar	udpsport[2];	/* Source port */
	uchar	udpdport[2];	/* Destination port */
	uchar	udplen[2];	/* data length */
	uchar	udpcksum[2];	/* Checksum */
};

typedef struct Rudphdr Rudphdr;
struct Rudphdr
{
	/* ip header */
	uchar	vihl;		/* Version and header length */
	uchar	tos;		/* Type of service */
	uchar	length[2];	/* packet length */
	uchar	id[2];		/* Identification */
	uchar	frag[2];	/* Fragment information */

	/* pseudo header starts here */
	uchar	Unused;
	uchar	udpproto;	/* Protocol */
	uchar	udpplen[2];	/* Header plus data length */
	uchar	udpsrc[4];	/* Ip source */
	uchar	udpdst[4];	/* Ip destination */

	/* udp header */
	uchar	udpsport[2];	/* Source port */
	uchar	udpdport[2];	/* Destination port */
	uchar	udplen[2];	/* data length (includes rudp header) */
	uchar	udpcksum[2];	/* Checksum */

	/* rudp header */
	uchar	relseq[4];	/* id of this packet (or 0) */
	uchar	relsgen[4];	/* generation/time stamp */
	uchar	relack[4];	/* packet being acked (or 0) */
	uchar	relagen[4];	/* generation/time stamp */
};


/*
 *  one state structure per destination
 */
typedef struct Reliable Reliable;
struct Reliable
{
	Ref;

	Reliable *next;

	uchar	addr[IPaddrlen];	/* always V6 when put here */
	ushort	port;

	Block	*unacked;	/* unacked msg list */
	Block	*unackedtail;	/*  and its tail */

	int	timeout;	/* time since first unacked msg sent */
	int	xmits;		/* number of times first unacked msg sent */

	ulong	sndseq;		/* next packet to be sent */
	ulong	sndgen;		/*  and its generation */

	ulong	rcvseq;		/* last packet received */
	ulong	rcvgen;		/*  and its generation */

	ulong	acksent;	/* last ack sent */
	ulong	ackrcvd;	/* last msg for which ack was rcvd */

	/* flow control */
	QLock	lock;
	Rendez	vous;
	int	blocked;
};



/* MIB II counters */
typedef struct Rudpstats Rudpstats;
struct Rudpstats
{
	ulong	rudpInDatagrams;
	ulong	rudpNoPorts;
	ulong	rudpInErrors;
	ulong	rudpOutDatagrams;
};

typedef struct Rudppriv Rudppriv;
struct Rudppriv
{
	Ipht	ht;

	/* MIB counters */
	Rudpstats	ustats;

	/* non-MIB stats */
	ulong	csumerr;		/* checksum errors */
	ulong	lenerr;			/* short packet */
	ulong	rxmits;			/* # of retransmissions */
	ulong	orders;			/* # of out of order pkts */

	/* keeping track of the ack kproc */
	int	ackprocstarted;
	QLock	apl;
};


static ulong generation = 0;
static Rendez rend;

/*
 *  protocol specific part of Conv
 */
typedef struct Rudpcb Rudpcb;
struct Rudpcb
{
	QLock;
	uchar	headers;
	uchar	randdrop;
	Reliable *r;
};

/*
 * local functions
 */
void	relsendack(Conv*, Reliable*, int);
int	reliput(Conv*, Block*, uchar*, ushort);
Reliable *relstate(Rudpcb*, uchar*, ushort, char*);
void	relput(Reliable*);
void	relforget(Conv *, uchar*, int, int);
void	relackproc(void *);
void	relackq(Reliable *, Block*);
void	relhangup(Conv *, Reliable*);
void	relrexmit(Conv *, Reliable*);
void	relput(Reliable*);
void	rudpkick(void *x);

static void
rudpstartackproc(Proto *rudp)
{
	Rudppriv *rpriv;
	char kpname[KNAMELEN];

	rpriv = rudp-&gt;priv;
	if(rpriv-&gt;ackprocstarted == 0){
		qlock(&amp;rpriv-&gt;apl);
		if(rpriv-&gt;ackprocstarted == 0){
			snprint(kpname, sizeof kpname, "#I%drudpack",
				rudp-&gt;f-&gt;dev);
			kproc(kpname, relackproc, rudp);
			rpriv-&gt;ackprocstarted = 1;
		}
		qunlock(&amp;rpriv-&gt;apl);
	}
}

static char*
rudpconnect(Conv *c, char **argv, int argc)
{
	char *e;
	Rudppriv *upriv;

	upriv = c-&gt;p-&gt;priv;
	rudpstartackproc(c-&gt;p);
	e = Fsstdconnect(c, argv, argc);
	Fsconnected(c, e);
	iphtadd(&amp;upriv-&gt;ht, c);

	return e;
}


static int
rudpstate(Conv *c, char *state, int n)
{
	Rudpcb *ucb;
	Reliable *r;
	int m;

	m = snprint(state, n, "%s", c-&gt;inuse?"Open":"Closed");
	ucb = (Rudpcb*)c-&gt;ptcl;
	qlock(ucb);
	for(r = ucb-&gt;r; r; r = r-&gt;next)
		m += snprint(state+m, n-m, " %I/%ld", r-&gt;addr, UNACKED(r));
	m += snprint(state+m, n-m, "\n");
	qunlock(ucb);
	return m;
}

static char*
rudpannounce(Conv *c, char** argv, int argc)
{
	char *e;
	Rudppriv *upriv;

	upriv = c-&gt;p-&gt;priv;
	rudpstartackproc(c-&gt;p);
	e = Fsstdannounce(c, argv, argc);
	if(e != nil)
		return e;
	Fsconnected(c, nil);
	iphtadd(&amp;upriv-&gt;ht, c);

	return nil;
}

static void
rudpcreate(Conv *c)
{
	c-&gt;rq = qopen(64*1024, Qmsg, 0, 0);
	c-&gt;wq = qopen(64*1024, Qkick, rudpkick, c);
}

static void
rudpclose(Conv *c)
{
	Rudpcb *ucb;
	Reliable *r, *nr;
	Rudppriv *upriv;

	upriv = c-&gt;p-&gt;priv;
	iphtrem(&amp;upriv-&gt;ht, c);

	/* force out any delayed acks */
	ucb = (Rudpcb*)c-&gt;ptcl;
	qlock(ucb);
	for(r = ucb-&gt;r; r; r = r-&gt;next){
		if(r-&gt;acksent != r-&gt;rcvseq)
			relsendack(c, r, 0);
	}
	qunlock(ucb);

	qclose(c-&gt;rq);
	qclose(c-&gt;wq);
	qclose(c-&gt;eq);
	ipmove(c-&gt;laddr, IPnoaddr);
	ipmove(c-&gt;raddr, IPnoaddr);
	c-&gt;lport = 0;
	c-&gt;rport = 0;

	ucb-&gt;headers = 0;
	ucb-&gt;randdrop = 0;
	qlock(ucb);
	for(r = ucb-&gt;r; r; r = nr){
		if(r-&gt;acksent != r-&gt;rcvseq)
			relsendack(c, r, 0);
		nr = r-&gt;next;
		relhangup(c, r);
		relput(r);
	}
	ucb-&gt;r = 0;

	qunlock(ucb);
}

/*
 *  randomly don't send packets
 */
static void
doipoput(Conv *c, Fs *f, Block *bp, int x, int ttl, int tos)
{
	Rudpcb *ucb;

	ucb = (Rudpcb*)c-&gt;ptcl;
	if(ucb-&gt;randdrop &amp;&amp; nrand(100) &lt; ucb-&gt;randdrop)
		freeblist(bp);
	else
		ipoput4(f, bp, x, ttl, tos, nil);
}

int
flow(void *v)
{
	Reliable *r = v;

	return UNACKED(r) &lt;= Maxunacked;
}

void
rudpkick(void *x)
{
	Conv *c = x;
	Udphdr *uh;
	ushort rport;
	uchar laddr[IPaddrlen], raddr[IPaddrlen];
	Block *bp;
	Rudpcb *ucb;
	Rudphdr *rh;
	Reliable *r;
	int dlen, ptcllen;
	Rudppriv *upriv;
	Fs *f;

	upriv = c-&gt;p-&gt;priv;
	f = c-&gt;p-&gt;f;

	netlog(c-&gt;p-&gt;f, Logrudp, "rudp: kick\n");
	bp = qget(c-&gt;wq);
	if(bp == nil)
		return;

	ucb = (Rudpcb*)c-&gt;ptcl;
	switch(ucb-&gt;headers) {
	case 7:
		/* get user specified addresses */
		bp = pullupblock(bp, UDP_USEAD7);
		if(bp == nil)
			return;
		ipmove(raddr, bp-&gt;rp);
		bp-&gt;rp += IPaddrlen;
		ipmove(laddr, bp-&gt;rp);
		bp-&gt;rp += IPaddrlen;
		/* pick interface closest to dest */
		if(ipforme(f, laddr) != Runi)
			findlocalip(f, laddr, raddr);
		bp-&gt;rp += IPaddrlen;		/* Ignore ifc address */
		rport = nhgets(bp-&gt;rp);
		bp-&gt;rp += 2+2;			/* Ignore local port */
		break;
	default:
		ipmove(raddr, c-&gt;raddr);
		ipmove(laddr, c-&gt;laddr);
		rport = c-&gt;rport;
		break;
	}

	dlen = blocklen(bp);

	/* Make space to fit rudp &amp; ip header */
	bp = padblock(bp, UDP_IPHDR+UDP_RHDRSIZE);
	if(bp == nil)
		return;

	uh = (Udphdr *)(bp-&gt;rp);
	uh-&gt;vihl = IP_VER4;

	rh = (Rudphdr*)uh;

	ptcllen = dlen + (UDP_RHDRSIZE-UDP_PHDRSIZE);
	uh-&gt;Unused = 0;
	uh-&gt;udpproto = IP_UDPPROTO;
	uh-&gt;frag[0] = 0;
	uh-&gt;frag[1] = 0;
	hnputs(uh-&gt;udpplen, ptcllen);
	switch(ucb-&gt;headers){
	case 7:
		v6tov4(uh-&gt;udpdst, raddr);
		hnputs(uh-&gt;udpdport, rport);
		v6tov4(uh-&gt;udpsrc, laddr);
		break;
	default:
		v6tov4(uh-&gt;udpdst, c-&gt;raddr);
		hnputs(uh-&gt;udpdport, c-&gt;rport);
		if(ipcmp(c-&gt;laddr, IPnoaddr) == 0)
			findlocalip(f, c-&gt;laddr, c-&gt;raddr);
		v6tov4(uh-&gt;udpsrc, c-&gt;laddr);
		break;
	}
	hnputs(uh-&gt;udpsport, c-&gt;lport);
	hnputs(uh-&gt;udplen, ptcllen);
	uh-&gt;udpcksum[0] = 0;
	uh-&gt;udpcksum[1] = 0;

	qlock(ucb);
	r = relstate(ucb, raddr, rport, "kick");
	r-&gt;sndseq = NEXTSEQ(r-&gt;sndseq);
	hnputl(rh-&gt;relseq, r-&gt;sndseq);
	hnputl(rh-&gt;relsgen, r-&gt;sndgen);

	hnputl(rh-&gt;relack, r-&gt;rcvseq);  /* ACK last rcvd packet */
	hnputl(rh-&gt;relagen, r-&gt;rcvgen);

	if(r-&gt;rcvseq != r-&gt;acksent)
		r-&gt;acksent = r-&gt;rcvseq;

	hnputs(uh-&gt;udpcksum, ptclcsum(bp, UDP_IPHDR, dlen+UDP_RHDRSIZE));

	relackq(r, bp);
	qunlock(ucb);

	upriv-&gt;ustats.rudpOutDatagrams++;

	DPRINT("sent: %lud/%lud, %lud/%lud\n",
		r-&gt;sndseq, r-&gt;sndgen, r-&gt;rcvseq, r-&gt;rcvgen);

	doipoput(c, f, bp, 0, c-&gt;ttl, c-&gt;tos);

	if(waserror()) {
		relput(r);
		qunlock(&amp;r-&gt;lock);
		nexterror();
	}

	/* flow control of sorts */
	qlock(&amp;r-&gt;lock);
	if(UNACKED(r) &gt; Maxunacked){
		r-&gt;blocked = 1;
		sleep(&amp;r-&gt;vous, flow, r);
		r-&gt;blocked = 0;
	}

	qunlock(&amp;r-&gt;lock);
	relput(r);
	poperror();
}

void
rudpiput(Proto *rudp, Ipifc *ifc, Block *bp)
{
	int len, olen, ottl;
	Udphdr *uh;
	Conv *c;
	Rudpcb *ucb;
	uchar raddr[IPaddrlen], laddr[IPaddrlen];
	ushort rport, lport;
	Rudppriv *upriv;
	Fs *f;
	uchar *p;

	upriv = rudp-&gt;priv;
	f = rudp-&gt;f;

	upriv-&gt;ustats.rudpInDatagrams++;

	uh = (Udphdr*)(bp-&gt;rp);

	/* Put back pseudo header for checksum
	 * (remember old values for icmpnoconv())
	 */
	ottl = uh-&gt;Unused;
	uh-&gt;Unused = 0;
	len = nhgets(uh-&gt;udplen);
	olen = nhgets(uh-&gt;udpplen);
	hnputs(uh-&gt;udpplen, len);

	v4tov6(raddr, uh-&gt;udpsrc);
	v4tov6(laddr, uh-&gt;udpdst);
	lport = nhgets(uh-&gt;udpdport);
	rport = nhgets(uh-&gt;udpsport);

	if(nhgets(uh-&gt;udpcksum)) {
		if(ptclcsum(bp, UDP_IPHDR, len+UDP_PHDRSIZE)) {
			upriv-&gt;ustats.rudpInErrors++;
			upriv-&gt;csumerr++;
			netlog(f, Logrudp, "rudp: checksum error %I\n", raddr);
			DPRINT("rudp: checksum error %I\n", raddr);
			freeblist(bp);
			return;
		}
	}

	qlock(rudp);

	c = iphtlook(&amp;upriv-&gt;ht, raddr, rport, laddr, lport);
	if(c == nil){
		/* no conversation found */
		upriv-&gt;ustats.rudpNoPorts++;
		qunlock(rudp);
		netlog(f, Logudp, "udp: no conv %I!%d -&gt; %I!%d\n", raddr, rport,
			laddr, lport);
		uh-&gt;Unused = ottl;
		hnputs(uh-&gt;udpplen, olen);
		icmpnoconv(f, bp);
		freeblist(bp);
		return;
	}
	ucb = (Rudpcb*)c-&gt;ptcl;
	qlock(ucb);
	qunlock(rudp);

	if(reliput(c, bp, raddr, rport) &lt; 0){
		qunlock(ucb);
		freeb(bp);
		return;
	}

	/*
	 * Trim the packet down to data size
	 */

	len -= (UDP_RHDRSIZE-UDP_PHDRSIZE);
	bp = trimblock(bp, UDP_IPHDR+UDP_RHDRSIZE, len);
	if(bp == nil) {
		netlog(f, Logrudp, "rudp: len err %I.%d -&gt; %I.%d\n",
			raddr, rport, laddr, lport);
		DPRINT("rudp: len err %I.%d -&gt; %I.%d\n",
			raddr, rport, laddr, lport);
		upriv-&gt;lenerr++;
		return;
	}

	netlog(f, Logrudpmsg, "rudp: %I.%d -&gt; %I.%d l %d\n",
		raddr, rport, laddr, lport, len);

	switch(ucb-&gt;headers){
	case 7:
		/* pass the src address */
		bp = padblock(bp, UDP_USEAD7);
		p = bp-&gt;rp;
		ipmove(p, raddr); p += IPaddrlen;
		ipmove(p, laddr); p += IPaddrlen;
		ipmove(p, ifc-&gt;lifc-&gt;local); p += IPaddrlen;
		hnputs(p, rport); p += 2;
		hnputs(p, lport);
		break;
	default:
		/* connection oriented rudp */
		if(ipcmp(c-&gt;raddr, IPnoaddr) == 0){
			/* save the src address in the conversation */
		 	ipmove(c-&gt;raddr, raddr);
			c-&gt;rport = rport;

			/* reply with the same ip address (if not broadcast) */
			if(ipforme(f, laddr) == Runi)
				ipmove(c-&gt;laddr, laddr);
			else
				v4tov6(c-&gt;laddr, ifc-&gt;lifc-&gt;local);
		}
		break;
	}
	if(bp-&gt;next)
		bp = concatblock(bp);

	if(qfull(c-&gt;rq)) {
		netlog(f, Logrudp, "rudp: qfull %I.%d -&gt; %I.%d\n", raddr, rport,
			laddr, lport);
		freeblist(bp);
	}
	else
		qpass(c-&gt;rq, bp);

	qunlock(ucb);
}

static char *rudpunknown = "unknown rudp ctl request";

char*
rudpctl(Conv *c, char **f, int n)
{
	Rudpcb *ucb;
	uchar ip[IPaddrlen];
	int x;

	ucb = (Rudpcb*)c-&gt;ptcl;
	if(n &lt; 1)
		return rudpunknown;

	if(strcmp(f[0], "headers") == 0){
		ucb-&gt;headers = 7;		/* new headers format */
		return nil;
	} else if(strcmp(f[0], "hangup") == 0){
		if(n &lt; 3)
			return "bad syntax";
		if (parseip(ip, f[1]) == -1)
			return Ebadip;
		x = atoi(f[2]);
		qlock(ucb);
		relforget(c, ip, x, 1);
		qunlock(ucb);
		return nil;
	} else if(strcmp(f[0], "randdrop") == 0){
		x = 10;			/* default is 10% */
		if(n &gt; 1)
			x = atoi(f[1]);
		if(x &gt; 100 || x &lt; 0)
			return "illegal rudp drop rate";
		ucb-&gt;randdrop = x;
		return nil;
	}
	return rudpunknown;
}

void
rudpadvise(Proto *rudp, Block *bp, char *msg)
{
	Udphdr *h;
	uchar source[IPaddrlen], dest[IPaddrlen];
	ushort psource, pdest;
	Conv *s, **p;

	h = (Udphdr*)(bp-&gt;rp);

	v4tov6(dest, h-&gt;udpdst);
	v4tov6(source, h-&gt;udpsrc);
	psource = nhgets(h-&gt;udpsport);
	pdest = nhgets(h-&gt;udpdport);

	/* Look for a connection */
	for(p = rudp-&gt;conv; *p; p++) {
		s = *p;
		if(s-&gt;rport == pdest)
		if(s-&gt;lport == psource)
		if(ipcmp(s-&gt;raddr, dest) == 0)
		if(ipcmp(s-&gt;laddr, source) == 0){
			qhangup(s-&gt;rq, msg);
			qhangup(s-&gt;wq, msg);
			break;
		}
	}
	freeblist(bp);
}

int
rudpstats(Proto *rudp, char *buf, int len)
{
	Rudppriv *upriv;

	upriv = rudp-&gt;priv;
	return snprint(buf, len, "%lud %lud %lud %lud %lud %lud\n",
		upriv-&gt;ustats.rudpInDatagrams,
		upriv-&gt;ustats.rudpNoPorts,
		upriv-&gt;ustats.rudpInErrors,
		upriv-&gt;ustats.rudpOutDatagrams,
		upriv-&gt;rxmits,
		upriv-&gt;orders);
}

void
rudpinit(Fs *fs)
{

	Proto *rudp;

	rudp = smalloc(sizeof(Proto));
	rudp-&gt;priv = smalloc(sizeof(Rudppriv));
	rudp-&gt;name = "rudp";
	rudp-&gt;connect = rudpconnect;
	rudp-&gt;announce = rudpannounce;
	rudp-&gt;ctl = rudpctl;
	rudp-&gt;state = rudpstate;
	rudp-&gt;create = rudpcreate;
	rudp-&gt;close = rudpclose;
	rudp-&gt;rcv = rudpiput;
	rudp-&gt;advise = rudpadvise;
	rudp-&gt;stats = rudpstats;
	rudp-&gt;ipproto = IP_UDPPROTO;
	rudp-&gt;nc = 32;
	rudp-&gt;ptclsize = sizeof(Rudpcb);

	Fsproto(fs, rudp);
}

/*********************************************/
/* Here starts the reliable helper functions */
/*********************************************/
/*
 *  Enqueue a copy of an unacked block for possible retransmissions
 */
void
relackq(Reliable *r, Block *bp)
{
	Block *np;

	np = copyblock(bp, blocklen(bp));
	if(r-&gt;unacked)
		r-&gt;unackedtail-&gt;list = np;
	else {
		/* restart timer */
		r-&gt;timeout = 0;
		r-&gt;xmits = 1;
		r-&gt;unacked = np;
	}
	r-&gt;unackedtail = np;
	np-&gt;list = nil;
}

/*
 *  retransmit unacked blocks
 */
void
relackproc(void *a)
{
	Rudpcb *ucb;
	Proto *rudp;
	Reliable *r;
	Conv **s, *c;

	rudp = (Proto *)a;

loop:
	tsleep(&amp;up-&gt;sleep, return0, 0, Rudptickms);

	for(s = rudp-&gt;conv; *s; s++) {
		c = *s;
		ucb = (Rudpcb*)c-&gt;ptcl;
		qlock(ucb);

		for(r = ucb-&gt;r; r; r = r-&gt;next) {
			if(r-&gt;unacked != nil){
				r-&gt;timeout += Rudptickms;
				if(r-&gt;timeout &gt; Rudprxms*r-&gt;xmits)
					relrexmit(c, r);
			}
			if(r-&gt;acksent != r-&gt;rcvseq)
				relsendack(c, r, 0);
		}
		qunlock(ucb);
	}
	goto loop;
}

/*
 *  get the state record for a conversation
 */
Reliable*
relstate(Rudpcb *ucb, uchar *addr, ushort port, char *from)
{
	Reliable *r, **l;

	l = &amp;ucb-&gt;r;
	for(r = *l; r; r = *l){
		if(memcmp(addr, r-&gt;addr, IPaddrlen) == 0 &amp;&amp;
		    port == r-&gt;port)
			break;
		l = &amp;r-&gt;next;
	}

	/* no state for this addr/port, create some */
	if(r == nil){
		while(generation == 0)
			generation = rand();

		DPRINT("from %s new state %lud for %I!%ud\n",
		        from, generation, addr, port);

		r = smalloc(sizeof(Reliable));
		memmove(r-&gt;addr, addr, IPaddrlen);
		r-&gt;port = port;
		r-&gt;unacked = 0;
		if(generation == Hangupgen)
			generation++;
		r-&gt;sndgen = generation++;
		r-&gt;sndseq = 0;
		r-&gt;ackrcvd = 0;
		r-&gt;rcvgen = 0;
		r-&gt;rcvseq = 0;
		r-&gt;acksent = 0;
		r-&gt;xmits = 0;
		r-&gt;timeout = 0;
		r-&gt;ref = 0;
		incref(r);	/* one reference for being in the list */

		*l = r;
	}

	incref(r);
	return r;
}

void
relput(Reliable *r)
{
	if(decref(r) == 0)
		free(r);
}

/*
 *  forget a Reliable state
 */
void
relforget(Conv *c, uchar *ip, int port, int originator)
{
	Rudpcb *ucb;
	Reliable *r, **l;

	ucb = (Rudpcb*)c-&gt;ptcl;

	l = &amp;ucb-&gt;r;
	for(r = *l; r; r = *l){
		if(ipcmp(ip, r-&gt;addr) == 0 &amp;&amp; port == r-&gt;port){
			*l = r-&gt;next;
			if(originator)
				relsendack(c, r, 1);
			relhangup(c, r);
			relput(r);	/* remove from the list */
			break;
		}
		l = &amp;r-&gt;next;
	}
}

/*
 *  process a rcvd reliable packet. return -1 if not to be passed to user process,
 *  0 therwise.
 *
 *  called with ucb locked.
 */
int
reliput(Conv *c, Block *bp, uchar *addr, ushort port)
{
	Block *nbp;
	Rudpcb *ucb;
	Rudppriv *upriv;
	Udphdr *uh;
	Reliable *r;
	Rudphdr *rh;
	ulong seq, ack, sgen, agen, ackreal;
	int rv = -1;

	/* get fields */
	uh = (Udphdr*)(bp-&gt;rp);
	rh = (Rudphdr*)uh;
	seq = nhgetl(rh-&gt;relseq);
	sgen = nhgetl(rh-&gt;relsgen);
	ack = nhgetl(rh-&gt;relack);
	agen = nhgetl(rh-&gt;relagen);

	upriv = c-&gt;p-&gt;priv;
	ucb = (Rudpcb*)c-&gt;ptcl;
	r = relstate(ucb, addr, port, "input");

	DPRINT("rcvd %lud/%lud, %lud/%lud, r-&gt;sndgen = %lud\n",
		seq, sgen, ack, agen, r-&gt;sndgen);

	/* if acking an incorrect generation, ignore */
	if(ack &amp;&amp; agen != r-&gt;sndgen)
		goto out;

	/* Look for a hangup */
	if(sgen == Hangupgen) {
		if(agen == r-&gt;sndgen)
			relforget(c, addr, port, 0);
		goto out;
	}

	/* make sure we're not talking to a new remote side */
	if(r-&gt;rcvgen != sgen){
		if(seq != 0 &amp;&amp; seq != 1)
			goto out;

		/* new connection */
		if(r-&gt;rcvgen != 0){
			DPRINT("new con r-&gt;rcvgen = %lud, sgen = %lud\n", r-&gt;rcvgen, sgen);
			relhangup(c, r);
		}
		r-&gt;rcvgen = sgen;
	}

	/* dequeue acked packets */
	if(ack &amp;&amp; agen == r-&gt;sndgen){
		ackreal = 0;
		while(r-&gt;unacked != nil &amp;&amp; INSEQ(ack, r-&gt;ackrcvd, r-&gt;sndseq)){
			nbp = r-&gt;unacked;
			r-&gt;unacked = nbp-&gt;list;
			DPRINT("%lud/%lud acked, r-&gt;sndgen = %lud\n",
			       ack, agen, r-&gt;sndgen);
			freeb(nbp);
			r-&gt;ackrcvd = NEXTSEQ(r-&gt;ackrcvd);
			ackreal = 1;
		}

		/* flow control */
		if(UNACKED(r) &lt; Maxunacked/8 &amp;&amp; r-&gt;blocked)
			wakeup(&amp;r-&gt;vous);

		/*
		 *  retransmit next packet if the acked packet
		 *  was transmitted more than once
		 */
		if(ackreal &amp;&amp; r-&gt;unacked != nil){
			r-&gt;timeout = 0;
			if(r-&gt;xmits &gt; 1){
				r-&gt;xmits = 1;
				relrexmit(c, r);
			}
		}

	}

	/* no message or input queue full */
	if(seq == 0 || qfull(c-&gt;rq))
		goto out;

	/* refuse out of order delivery */
	if(seq != NEXTSEQ(r-&gt;rcvseq)){
		relsendack(c, r, 0);	/* tell him we got it already */
		upriv-&gt;orders++;
		DPRINT("out of sequence %lud not %lud\n", seq, NEXTSEQ(r-&gt;rcvseq));
		goto out;
	}
	r-&gt;rcvseq = seq;

	rv = 0;
out:
	relput(r);
	return rv;
}

void
relsendack(Conv *c, Reliable *r, int hangup)
{
	Udphdr *uh;
	Block *bp;
	Rudphdr *rh;
	int ptcllen;
	Fs *f;

	bp = allocb(UDP_IPHDR + UDP_RHDRSIZE);
	if(bp == nil)
		return;
	bp-&gt;wp += UDP_IPHDR + UDP_RHDRSIZE;
	f = c-&gt;p-&gt;f;
	uh = (Udphdr *)(bp-&gt;rp);
	uh-&gt;vihl = IP_VER4;
	rh = (Rudphdr*)uh;

	ptcllen = (UDP_RHDRSIZE-UDP_PHDRSIZE);
	uh-&gt;Unused = 0;
	uh-&gt;udpproto = IP_UDPPROTO;
	uh-&gt;frag[0] = 0;
	uh-&gt;frag[1] = 0;
	hnputs(uh-&gt;udpplen, ptcllen);

	v6tov4(uh-&gt;udpdst, r-&gt;addr);
	hnputs(uh-&gt;udpdport, r-&gt;port);
	hnputs(uh-&gt;udpsport, c-&gt;lport);
	if(ipcmp(c-&gt;laddr, IPnoaddr) == 0)
		findlocalip(f, c-&gt;laddr, c-&gt;raddr);
	v6tov4(uh-&gt;udpsrc, c-&gt;laddr);
	hnputs(uh-&gt;udplen, ptcllen);

	if(hangup)
		hnputl(rh-&gt;relsgen, Hangupgen);
	else
		hnputl(rh-&gt;relsgen, r-&gt;sndgen);
	hnputl(rh-&gt;relseq, 0);
	hnputl(rh-&gt;relagen, r-&gt;rcvgen);
	hnputl(rh-&gt;relack, r-&gt;rcvseq);

	if(r-&gt;acksent &lt; r-&gt;rcvseq)
		r-&gt;acksent = r-&gt;rcvseq;

	uh-&gt;udpcksum[0] = 0;
	uh-&gt;udpcksum[1] = 0;
	hnputs(uh-&gt;udpcksum, ptclcsum(bp, UDP_IPHDR, UDP_RHDRSIZE));

	DPRINT("sendack: %lud/%lud, %lud/%lud\n", 0L, r-&gt;sndgen, r-&gt;rcvseq, r-&gt;rcvgen);
	doipoput(c, f, bp, 0, c-&gt;ttl, c-&gt;tos);
}


/*
 *  called with ucb locked (and c locked if user initiated close)
 */
void
relhangup(Conv *c, Reliable *r)
{
	int n;
	Block *bp;
	char hup[ERRMAX];

	n = snprint(hup, sizeof(hup), "hangup %I!%d", r-&gt;addr, r-&gt;port);
	qproduce(c-&gt;eq, hup, n);

	/*
	 *  dump any unacked outgoing messages
	 */
	for(bp = r-&gt;unacked; bp != nil; bp = r-&gt;unacked){
		r-&gt;unacked = bp-&gt;list;
		bp-&gt;list = nil;
		freeb(bp);
	}

	r-&gt;rcvgen = 0;
	r-&gt;rcvseq = 0;
	r-&gt;acksent = 0;
	if(generation == Hangupgen)
		generation++;
	r-&gt;sndgen = generation++;
	r-&gt;sndseq = 0;
	r-&gt;ackrcvd = 0;
	r-&gt;xmits = 0;
	r-&gt;timeout = 0;
	wakeup(&amp;r-&gt;vous);
}

/*
 *  called with ucb locked
 */
void
relrexmit(Conv *c, Reliable *r)
{
	Rudppriv *upriv;
	Block *np;
	Fs *f;

	upriv = c-&gt;p-&gt;priv;
	f = c-&gt;p-&gt;f;
	r-&gt;timeout = 0;
	if(r-&gt;xmits++ &gt; Rudpmaxxmit){
		relhangup(c, r);
		return;
	}

	upriv-&gt;rxmits++;
	np = copyblock(r-&gt;unacked, blocklen(r-&gt;unacked));
	DPRINT("rxmit r-&gt;ackrvcd+1 = %lud\n", r-&gt;ackrcvd+1);
	doipoput(c, f, np, 0, c-&gt;ttl, c-&gt;tos);
}
sed: Can't open /usr/web/plan9/sources.html
