<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/9/omap/clock.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/9/omap/clock.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * omap3530 clocks
 *
 * timers count up to zero.
 *
 * the source clock signals for the timers are sometimes selectable.  for
 * WDTIMER[23] and GPTIMER12, it's always the 32kHz clock.  for the
 * others, it can be the 32kHz clock or the system clock.  we use only
 * WDTIMER2 and GPTIMER[12], and configure GPTIMER[12] in archomap.c to
 * use the 32kHZ clock.  WDTIMER1 is not accessible to us on GP
 * (general-purpose) omaps.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "arm.h"

enum {
	Debug		= 0,

	Tn0		= PHYSTIMER1,
	Tn1		= PHYSTIMER2,

	/* irq 36 is watchdog timer module 3 overflow */
	Tn0irq		= 37,			/* base IRQ for all timers */

	Freebase	= 1,			/* base of free-running timer */

	/*
	 * clock is 32K (32,768) Hz, so one tick is 30.517µs,
	 * so 327.68 ticks is 10ms, 32.768 ticks is 1ms.
	 */
	Clockfreqbase	= 32 * 1024,		/* base rate in Hz */
	Tcycles		= Clockfreqbase / HZ,	/* cycles per clock tick */

	MinPeriod	= (Tcycles / 100 &lt; 2? 2: Tcycles / 100),
	MaxPeriod	= Tcycles,

	Dogtimeout	= 20 * Clockfreqbase,	/* was 4 s.; must be ≤ 21 s. */
};

enum {
	/* ticpcfg bits */
	Noidle		= 1&lt;&lt;3,
	Softreset	= 1&lt;&lt;1,

	/* tistat bits */
	Resetdone	= 1&lt;&lt;0,

	/* tisr/tier bits */
	Ovf_it		= 1&lt;&lt;1,		/* gp: overflow intr */
	Mat_it		= 1&lt;&lt;0,		/* gp: match intr */
	Wdovf_it	= 1&lt;&lt;0,		/* wdog: overflow intr */

	/* tclr bits */
	Ar		= 1&lt;&lt;1,		/* gp only: autoreload mode overflow */
	St		= 1&lt;&lt;0,		/* gp only: start the timer */
};

/* omap35x timer registers */
typedef struct Timerregs Timerregs;
struct Timerregs {
	/* common to all timers, gp and watchdog */
	uchar	pad0[0x10];
	ulong	ticpcfg;
	ulong	tistat;		/* ro: low bit: reset done */
	ulong	tisr;
	ulong	tier;
	ulong	twer;
	ulong	tclr;
	ulong	tcrr;		/* counter: cycles to zero */
	ulong	tldr;
	ulong	ttgr;		/* trigger */
	ulong	twps;		/* ro: write posted pending */

	/* gp timers only, unused by us */
	ulong	tmar;		/* value to compare with counter */
	ulong	tcar1;		/* ro */
	ulong	tsicr;
	ulong	tcar2;		/* ro */
	union {
		ulong	tpir;	/* gp: 1 ms tick generation: +ve */
		ulong	wspr;	/* wdog: start/stop control */
	};
	ulong	tnir;		/* 1 ms tick generation: -ve */
	ulong	tcvr;		/* 1 ms tick generation: next counter value */
	ulong	tocr;		/* intr mask for n ticks */
	ulong	towr;
};

static int ticks; /* for sanity checking; m-&gt;ticks doesn't always get called */
static Lock clklck;

static ulong	rdcycles(void), rdbaseticks(void);

/* write a watchdog timer's start/stop register */
static void
wdogwrss(Timerregs *tn, ulong val)
{
	while (tn-&gt;twps &amp; (1 &lt;&lt; 4))	/* pending write to start/stop reg? */
		;
	tn-&gt;wspr = val;
	coherence();
	while (tn-&gt;twps &amp; (1 &lt;&lt; 4))	/* pending write to start/stop reg? */
		;
}

static void
resetwait(Timerregs *tn)
{
	long bound;

	for (bound = 400*Mhz; !(tn-&gt;tistat &amp; Resetdone) &amp;&amp; bound &gt; 0; bound--)
		;
	if (bound &lt;= 0)
		iprint("clock reset didn't complete\n");
}


static void
wdogoff(Timerregs *tn)
{
	resetwait(tn);

	wdogwrss(tn, 0xaaaa);		/* magic off sequence */
	wdogwrss(tn, 0x5555);

	tn-&gt;tldr = 1;
	coherence();
	tn-&gt;tcrr = 1;			/* paranoia */
	coherence();
}

static void wdogassure(void);

static void
wdogon(Timerregs *tn)
{
	static int beenhere;

	resetwait(tn);
	tn-&gt;tldr = -Dogtimeout;
	tn-&gt;tcrr = -Dogtimeout;
	coherence();
	wdogwrss(tn, 0xbbbb);		/* magic on sequence */
	wdogwrss(tn, 0x4444);		/* magic on sequence */

	if (!beenhere) {
		beenhere = 1;
		/* touching the dog is not quick, so do it infrequently */
		addclock0link(wdogassure, HZ);
	}
}

static void
wdogassure(void)		/* reset the watch dog's counter */
{
	Timerregs *tn;

	tn = (Timerregs *)PHYSWDOG;
	wdogoff(tn);

	tn-&gt;tcrr = -Dogtimeout;
	coherence();

	wdogon(tn);
}

static void
clockintr(Ureg* ureg, void *arg)
{
	Timerregs *tn;
	static int nesting;

	ticks++;
	coherence();

	if (nesting == 0) {	/* if the clock interrupted itself, bail out */
		++nesting;
		timerintr(ureg, 0);
		--nesting;
	}

	tn = arg;
	tn-&gt;tisr = Ovf_it;			/* dismiss the interrupt */
	coherence();
}

static void
clockreset(Timerregs *tn)
{
	if (probeaddr((uintptr)&amp;tn-&gt;ticpcfg) &lt; 0)
		panic("no clock at %#p", tn);
	tn-&gt;ticpcfg = Softreset | Noidle;
	coherence();
	resetwait(tn);
	tn-&gt;tier = tn-&gt;tclr = 0;
	coherence();
}

/* stop clock interrupts and disable the watchdog timer */
void
clockshutdown(void)
{
	clockreset((Timerregs *)PHYSWDT2);
	wdogoff((Timerregs *)PHYSWDT2);
	clockreset((Timerregs *)PHYSWDT3);
	wdogoff((Timerregs *)PHYSWDT3);

	clockreset((Timerregs *)Tn0);
	clockreset((Timerregs *)Tn1);
}

enum {
	Instrs		= 10*Mhz,
};

static long
issue1loop(void)
{
	register int i;
	long st;

	st = rdbaseticks();
	i = Instrs;
	do {
		--i; --i; --i; --i; --i;
		--i; --i; --i; --i;
	} while(--i &gt;= 0);
	return rdbaseticks() - st;
}

static long
issue2loop(void)
{
	register int i, j;
	long st;

	st = rdbaseticks();
	i = Instrs / 2;
	j = 0;
	do {
		--i; --j; --i; --j;
		--i; --j; --i; --j;
		--j;
	} while(--i &gt;= 0);
	return rdbaseticks() - st;
}

/* estimate instructions/s. using 32kHz clock */
static void
guessmips(long (*loop)(void), char *lab)
{
	int s;
	long tcks;

	do {
		s = splhi();
		tcks = loop();
		splx(s);
		if (tcks &lt; 0)
			iprint("again...");
	} while (tcks &lt; 0);
	/*
	 * Instrs instructions took tcks ticks @ Clockfreqbase Hz.
	 */
	s = ((vlong)Clockfreqbase * Instrs) / tcks / 1000000;
	if (Debug)
		iprint("%ud mips (%s-issue)", s, lab);
	USED(s);
}

void
clockinit(void)
{
	int i, s;
	Timerregs *tn;

	clockshutdown();

	/* turn cycle counter on */
	cpwrsc(0, CpCLD, CpCLDena, CpCLDenacyc, 1&lt;&lt;31);

	/* turn all counters on and clear the cycle counter */
	cpwrsc(0, CpCLD, CpCLDena, CpCLDenapmnc, 1&lt;&lt;2 | 1);

	/* let users read the cycle counter directly */
	cpwrsc(0, CpCLD, CpCLDena, CpCLDenapmnc, 1);

	ilock(&amp;clklck);
	m-&gt;fastclock = 1;
	m-&gt;ticks = ticks = 0;

	/*
	 * T0 is a freerunning timer (cycle counter); it wraps,
	 * automatically reloads, and does not dispatch interrupts.
	 */
	tn = (Timerregs *)Tn0;
	tn-&gt;tcrr = Freebase;			/* count up to 0 */
	tn-&gt;tldr = Freebase;
	coherence();
	tn-&gt;tclr = Ar | St;
	iunlock(&amp;clklck);

	/*
	 * T1 is the interrupting timer and does not participate
	 * in measuring time.  It is initially set to HZ.
	 */
	tn = (Timerregs *)Tn1;
	irqenable(Tn0irq+1, clockintr, tn, "clock");
	ilock(&amp;clklck);
	tn-&gt;tcrr = -Tcycles;			/* approx.; count up to 0 */
	tn-&gt;tldr = -Tcycles;
	coherence();
	tn-&gt;tclr = Ar | St;
	coherence();
	tn-&gt;tier = Ovf_it;
	coherence();
	iunlock(&amp;clklck);

	/*
	 * verify sanity of timer1
	 */
	s = spllo();			/* risky */
	for (i = 0; i &lt; 5 &amp;&amp; ticks == 0; i++) {
		delay(10);
		cachedwbinvse(&amp;ticks, sizeof ticks);
	}
	splx(s);
	if (ticks == 0) {
		if (tn-&gt;tcrr == 0)
			panic("clock not interrupting");
		else if (tn-&gt;tcrr == tn-&gt;tldr)
			panic("clock not ticking at all");
#ifdef PARANOID
		else
			panic("clock running very slowly");
#endif
	}

	guessmips(issue1loop, "single");
	if (Debug)
		iprint(", ");
	guessmips(issue2loop, "dual");
	if (Debug)
		iprint("\n");

	/*
	 * m-&gt;delayloop should be the number of delay loop iterations
	 * needed to consume 1 ms.  2 is min. instructions in the delay loop.
	 */
	m-&gt;delayloop = m-&gt;cpuhz / (1000 * 2);
//	iprint("m-&gt;delayloop = %lud\n", m-&gt;delayloop);

	/*
	 *  desynchronize the processor clocks so that they all don't
	 *  try to resched at the same time.
	 */
	delay(m-&gt;machno*2);
}

void
watchdoginit(void)
{
	wdogassure();
}

ulong
µs(void)
{
	return fastticks2us(fastticks(nil));
}

void
timerset(Tval next)
{
	long offset;
	Timerregs *tn = (Timerregs *)Tn1;
	static Lock setlck;

	ilock(&amp;setlck);
	offset = next - fastticks(nil);
	if(offset &lt; MinPeriod)
		offset = MinPeriod;
	else if(offset &gt; MaxPeriod)
		offset = MaxPeriod;
	tn-&gt;tcrr = -offset;
	coherence();
	iunlock(&amp;setlck);
}

static ulong
rdcycles(void)
{
	ulong v;

	/* reads 32-bit cycle counter (counting up) */
	v = cprdsc(0, CpCLD, CpCLDcyc, 0);
	/* keep it positive; prevent m-&gt;fastclock ever going to 0 */
	return v == 0? 1: v;
}

static ulong
rdbaseticks(void)
{
	ulong v;

	v = ((Timerregs *)Tn0)-&gt;tcrr;		/* tcrr should be counting up */
	/* keep it positive; prevent m-&gt;fastclock ever going to 0 */
	return v == 0? 1: v;
}

ulong
perfticks(void)
{
	return rdcycles();
}

long
lcycles(void)
{
	return perfticks();
}

/*
 * until 5[cal] inline vlong ops, avoid them where possible,
 * they are currently slow function calls.
 */
typedef union Counter Counter;
union Counter {
	uvlong	uvl;
	struct {			/* little-endian */
		ulong	low;
		ulong	high;
	};
};

enum {
	Fastvlongops	= 0,
};

uvlong
fastticks(uvlong *hz)
{
	Counter now, sclnow;

	if(hz)
		*hz = m-&gt;cpuhz;
	ilock(&amp;clklck);
	if (m-&gt;ticks &gt; HZ/10 &amp;&amp; m-&gt;fastclock == 0)
		panic("fastticks: zero m-&gt;fastclock; ticks %lud fastclock %#llux",
			m-&gt;ticks, m-&gt;fastclock);

	now.uvl = m-&gt;fastclock;
	now.low = rdcycles();
	if(now.uvl &lt; m-&gt;fastclock)	/* low bits must have wrapped */
		now.high++;
	m-&gt;fastclock = now.uvl;
	coherence();

	sclnow.uvl = now.uvl;
	iunlock(&amp;clklck);
	return sclnow.uvl;
}

void
microdelay(int l)
{
	int i;

	l = l * (vlong)m-&gt;delayloop / 1000;
	if(l &lt;= 0)
		l = 1;
	for(i = 0; i &lt; l; i++)
		;
}

void
delay(int l)
{
	ulong i, j;

	j = m-&gt;delayloop;
	while(l-- &gt; 0)
		for(i=0; i &lt; j; i++)
			;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

