<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/9/pc/mtrr.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/9/pc/mtrr.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * memory-type region registers.
 *
 * due to the possibility of extended addresses (for PAE)
 * as large as 36 bits coming from the e820 memory map and the like,
 * we'll use vlongs to hold addresses and lengths, even though we don't
 * implement PAE in Plan 9.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"

enum {
	/*
	 * MTRR Physical base/mask are indexed by
	 *	MTRRPhys{Base|Mask}N = MTRRPhys{Base|Mask}0 + 2*N
	 */
	MTRRPhysBase0 = 0x200,
	MTRRPhysMask0 = 0x201,
	MTRRDefaultType = 0x2FF,
	MTRRCap = 0xFE,
	Nmtrr = 8,

	/* cpuid extended function codes */
	Exthighfunc = 1ul &lt;&lt; 31,
	Extprocsigamd,
	Extprocname0,
	Extprocname1,
	Extprocname2,
	Exttlbl1,
	Extl2,
	Extapm,
	Extaddrsz,

	Paerange = 1LL &lt;&lt; 36,
};

enum {
	CR4PageGlobalEnable	= 1 &lt;&lt; 7,
	CR0CacheDisable		= 1 &lt;&lt; 30,
};

enum {
	Uncacheable	= 0,
	Writecomb	= 1,
	Unknown1	= 2,
	Unknown2	= 3,
	Writethru	= 4,
	Writeprot	= 5,
	Writeback	= 6,
};

enum {
	Capvcnt = 0xff,		/* mask: # of variable-range MTRRs we have */
	Capwc	= 1&lt;&lt;8,		/* flag: have write combining? */
	Capfix	= 1&lt;&lt;10,	/* flag: have fixed MTRRs? */
	Deftype = 0xff,		/* default MTRR type */
	Deffixena = 1&lt;&lt;10,	/* fixed-range MTRR enable */
	Defena	= 1&lt;&lt;11,	/* MTRR enable */
};

typedef struct Mtrreg Mtrreg;
typedef struct Mtrrop Mtrrop;

struct Mtrreg {
	vlong	base;
	vlong	mask;
};
struct Mtrrop {
	Mtrreg	*reg;
	int	slot;
};

static char *types[] = {
[Uncacheable]	"uc",
[Writecomb]	"wc",
[Unknown1]	"uk1",
[Unknown2]	"uk2",
[Writethru]	"wt",
[Writeprot]	"wp",
[Writeback]	"wb",
		nil
};
static Mtrrop *postedop;
static Rendez oprend;

static char *
type2str(int type)
{
	if(type &lt; 0 || type &gt;= nelem(types))
		return nil;
	return types[type];
}

static int
str2type(char *str)
{
	char **p;

	for(p = types; *p != nil; p++)
		if (strcmp(str, *p) == 0)
			return p - types;
	return -1;
}

static uvlong
physmask(void)
{
	ulong regs[4];
	static vlong mask = -1;

	if (mask != -1)
		return mask;
	cpuid(Exthighfunc, regs);
	if(regs[0] &gt;= Extaddrsz) {			/* ax */
		cpuid(Extaddrsz, regs);
		mask = (1LL &lt;&lt; (regs[0] &amp; 0xFF)) - 1;	/* ax */
	}
	mask &amp;= Paerange - 1;				/* x86 sanity */
	return mask;
}

/* limit physical addresses to 36 bits on the x86 */
static void
sanity(Mtrreg *mtrr)
{
	mtrr-&gt;base &amp;= Paerange - 1;
	mtrr-&gt;mask &amp;= Paerange - 1;
}

static int
ispow2(uvlong ul)
{
	return (ul &amp; (ul - 1)) == 0;
}

/* true if mtrr is valid */
static int
mtrrdec(Mtrreg *mtrr, uvlong *ptr, uvlong *size, int *type)
{
	sanity(mtrr);
	*ptr =  mtrr-&gt;base &amp; ~(BY2PG-1);
	*type = mtrr-&gt;base &amp; 0xff;
	*size = (physmask() ^ (mtrr-&gt;mask &amp; ~(BY2PG-1))) + 1;
	return (mtrr-&gt;mask &gt;&gt; 11) &amp; 1;
}

static void
mtrrenc(Mtrreg *mtrr, uvlong ptr, uvlong size, int type, int ok)
{
	mtrr-&gt;base = ptr | (type &amp; 0xff);
	mtrr-&gt;mask = (physmask() &amp; ~(size - 1)) | (ok? 1&lt;&lt;11: 0);
	sanity(mtrr);
}

/*
 * i is the index of the MTRR, and is multiplied by 2 because
 * mask and base offsets are interleaved.
 */
static void
mtrrget(Mtrreg *mtrr, uint i)
{
	if (i &gt;= Nmtrr)
		error("mtrr index out of range");
	rdmsr(MTRRPhysBase0 + 2*i, &amp;mtrr-&gt;base);
	rdmsr(MTRRPhysMask0 + 2*i, &amp;mtrr-&gt;mask);
	sanity(mtrr);
}

static void
mtrrput(Mtrreg *mtrr, uint i)
{
	if (i &gt;= Nmtrr)
		error("mtrr index out of range");
	sanity(mtrr);
	wrmsr(MTRRPhysBase0 + 2*i, mtrr-&gt;base);
	wrmsr(MTRRPhysMask0 + 2*i, mtrr-&gt;mask);
}

static void
mtrrop(Mtrrop **op)
{
	int s;
	ulong cr0, cr4;
	vlong def;
	static long bar1, bar2;

	s = splhi();		/* avoid race with mtrrclock */

	/*
	 * wait for all CPUs to sync here, so that the MTRR setup gets
	 * done at roughly the same time on all processors.
	 */
	_xinc(&amp;bar1);
	while(bar1 &lt; conf.nmach)
		microdelay(10);

	cr4 = getcr4();
	putcr4(cr4 &amp; ~CR4PageGlobalEnable);
	cr0 = getcr0();
	wbinvd();
	putcr0(cr0 | CR0CacheDisable);
	wbinvd();
	rdmsr(MTRRDefaultType, &amp;def);
	wrmsr(MTRRDefaultType, def &amp; ~(vlong)Defena);

	mtrrput((*op)-&gt;reg, (*op)-&gt;slot);

	wbinvd();
	wrmsr(MTRRDefaultType, def);
	putcr0(cr0);
	putcr4(cr4);

	/*
	 * wait for all CPUs to sync up again, so that we don't continue
	 * executing while the MTRRs are still being set up.
	 */
	_xinc(&amp;bar2);
	while(bar2 &lt; conf.nmach)
		microdelay(10);
	*op = nil;
	_xdec(&amp;bar1);
	while(bar1 &gt; 0)
		microdelay(10);
	_xdec(&amp;bar2);
	wakeup(&amp;oprend);
	splx(s);
}

void
mtrrclock(void)				/* called from clock interrupt */
{
	if(postedop != nil)
		mtrrop(&amp;postedop);
}

/* if there's an operation still pending, keep sleeping */
static int
opavail(void *)
{
	return postedop == nil;
}

int
mtrr(uvlong base, uvlong size, char *tstr)
{
	int i, vcnt, slot, type, mtype, mok;
	vlong def, cap;
	uvlong mp, msize;
	Mtrreg entry, mtrr;
	Mtrrop op;
	static int tickreg;
	static QLock mtrrlk;

	if(!(m-&gt;cpuiddx &amp; Mtrr))
		error("mtrrs not supported");
	if(base &amp; (BY2PG-1) || size &amp; (BY2PG-1) || size == 0)
		error("mtrr base or size not 4k aligned or zero size");
	if(base + size &gt;= Paerange)
		error("mtrr range exceeds 36 bits");
	if(!ispow2(size))
		error("mtrr size not power of 2");
	if(base &amp; (size - 1))
		error("mtrr base not naturally aligned");

	if((type = str2type(tstr)) == -1)
		error("mtrr bad type");

	rdmsr(MTRRCap, &amp;cap);
	rdmsr(MTRRDefaultType, &amp;def);

	switch(type){
	default:
		error("mtrr unknown type");
		break;
	case Writecomb:
		if(!(cap &amp; Capwc))
			error("mtrr type wc (write combining) unsupported");
		/* fallthrough */
	case Uncacheable:
	case Writethru:
	case Writeprot:
	case Writeback:
		break;
	}

	qlock(&amp;mtrrlk);
	slot = -1;
	vcnt = cap &amp; Capvcnt;
	for(i = 0; i &lt; vcnt &amp;&amp; i &lt; Nmtrr; i++){
		mtrrget(&amp;mtrr, i);
		mok = mtrrdec(&amp;mtrr, &amp;mp, &amp;msize, &amp;mtype);
		/* reuse any entry for addresses above 4GB */
		if(!mok || mp == base &amp;&amp; msize == size || mp &gt;= (1LL&lt;&lt;32)){
			slot = i;
			break;
		}
	}
	if(slot == -1)
		error("no free mtrr slots");

	while(postedop != nil)
		sleep(&amp;oprend, opavail, 0);
	mtrrenc(&amp;entry, base, size, type, 1);
	op.reg = &amp;entry;
	op.slot = slot;
	postedop = &amp;op;
	mtrrop(&amp;postedop);
	qunlock(&amp;mtrrlk);
	return 0;
}

int
mtrrprint(char *buf, long bufsize)
{
	int i, vcnt, type;
	long n;
	uvlong base, size;
	vlong cap, def;
	Mtrreg mtrr;

	n = 0;
	if(!(m-&gt;cpuiddx &amp; Mtrr))
		return 0;
	rdmsr(MTRRCap, &amp;cap);
	rdmsr(MTRRDefaultType, &amp;def);
	n += snprint(buf+n, bufsize-n, "cache default %s\n",
		type2str(def &amp; Deftype));
	vcnt = cap &amp; Capvcnt;
	for(i = 0; i &lt; vcnt &amp;&amp; i &lt; Nmtrr; i++){
		mtrrget(&amp;mtrr, i);
		if (mtrrdec(&amp;mtrr, &amp;base, &amp;size, &amp;type))
			n += snprint(buf+n, bufsize-n,
				"cache 0x%llux %llud %s\n",
				base, size, type2str(type));
	}
	return n;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

