<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/9/port/devfs.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/9/port/devfs.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * File system devices.
 * Follows device config in Ken's file server.
 * Builds mirrors, concatenations, interleavings, and partitions
 * of devices out of other (inner) devices.
 * It is ok if inner devices are provided by this driver.
 *
 * Built files are grouped on different directories
 * (called trees, and used to represent disks).
 * The "#k/fs" tree is always available and never goes away.
 * Configuration changes happen only while no I/O is in progress.
 *
 * Default sector size is one byte unless changed by the "disk" ctl.
 */

#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "ureg.h"
#include "../port/error.h"

enum
{
	Fnone,
	Fmirror,		/* mirror of others */
	Fcat,			/* catenation of others */
	Finter,			/* interleaving of others */
	Fpart,			/* part of other */
	Fclear,			/* start over */
	Fdel,			/* delete a configure device */
	Fdisk,			/* set default tree and sector sz*/

	Sectorsz = 1,
	Blksize	= 8*1024,	/* for Finter only */

	Incr = 5,		/* Increments for the dev array */

	/*
	 * All qids are decorated with the tree number.
	 * #k/fs is tree number 0, is automatically added and
	 * its first qid is for the ctl file. It never goes away.
	 */
	Qtop	= 0,		/* #k */
	Qdir,			/* directory (#k/fs) */
	Qctl,			/* ctl, only for #k/fs/ctl */
	Qfirst,			/* first qid assigned for device */

	Iswrite = 0,
	Isread,

	Optional = 0,
	Mustexist,

	/* tunable parameters */
	Maxconf	= 4*1024,	/* max length for config */
	Ndevs	= 32,		/* max. inner devs per command */
	Ntrees	= 128,		/* max. number of trees */
	Maxretries = 3,		/* max. retries of i/o errors */
	Retrypause = 5000,	/* ms. to pause between retries */
};

typedef struct Inner Inner;
typedef struct Fsdev Fsdev;
typedef struct Tree Tree;

struct Inner
{
	char	*iname;		/* inner device name */
	vlong	isize;		/* size of inner device */
	Chan	*idev;		/* inner device */
};

struct Fsdev
{
	Ref;			/* one per Chan doing I/O */
	int	gone;		/* true if removed */
	int	vers;		/* qid version for this device */
	int	type;		/* Fnone, Fmirror, ... */
	char	*name;		/* name for this fsdev */
	Tree*	tree;		/* where the device is kept */
	vlong	size;		/* min(inner[X].isize) */
	vlong	start;		/* start address (for Fpart) */
	uint	ndevs;		/* number of inner devices */
	int	perm;		/* minimum of inner device perms */
	Inner	*inner[Ndevs];	/* inner devices */
};

struct Tree
{
	char	*name;		/* name for #k/&lt;name&gt; */
	Fsdev	**devs;		/* devices in dir. */
	uint	ndevs;		/* number of devices */
	uint	nadevs;		/* number of allocated devices in devs */
};

#define dprint if(debug)print

extern Dev fsdevtab;		/* forward */

static RWlock lck;		/* r: use devices; w: change config  */
static Tree fstree;		/* The main "fs" tree. Never goes away */
static Tree *trees[Ntrees];	/* internal representation of config */
static int ntrees;		/* max number of trees */
static int qidvers;

static char *disk;		/* default tree name used */
static char *source;		/* default inner device used */
static int sectorsz = Sectorsz;	/* default sector size */

static char confstr[Maxconf];	/* textual configuration */

static int debug;

static char cfgstr[] = "fsdev:\n";

static Qid tqid = {Qtop, 0, QTDIR};
static Qid cqid = {Qctl, 0, 0};

static char* tnames[] = {
	[Fmirror]	"mirror",
	[Fcat]		"cat",
	[Finter]	"inter",
	[Fpart]		"part",
};

static Cmdtab configs[] = {
	Fmirror,"mirror",	0,
	Fcat,	"cat",		0,
	Finter,	"inter",	0,
	Fpart,	"part",		0,
	Fclear,	"clear",	1,
	Fdel,	"del",		2,
	Fdisk,	"disk",		0,
};

static char Egone[] = "device is gone";		/* file has been removed */

static char*
seprintdev(char *s, char *e, Fsdev *mp)
{
	int i;

	if(mp == nil)
		return seprint(s, e, "&lt;null Fsdev&gt;");
	if(mp-&gt;type &lt; 0 || mp-&gt;type &gt;= nelem(tnames) || tnames[mp-&gt;type] == nil)
		return seprint(s, e, "bad device type %d\n", mp-&gt;type);

	s = strecpy(s, e, tnames[mp-&gt;type]);
	if(mp-&gt;tree != &amp;fstree)
		s = seprint(s, e, " %s/%s", mp-&gt;tree-&gt;name, mp-&gt;name);
	else
		s = seprint(s, e, " %s", mp-&gt;name);
	for(i = 0; i &lt; mp-&gt;ndevs; i++)
		s = seprint(s, e, " %s", mp-&gt;inner[i]-&gt;iname);
	switch(mp-&gt;type){
	case Fmirror:
	case Fcat:
	case Finter:
		s = strecpy(s, e, "\n");
		break;
	case Fpart:
		s = seprint(s, e, " %ulld %ulld\n", mp-&gt;start, mp-&gt;size);
		break;
	default:
		panic("#k: seprintdev bug");
	}
	return s;
}

static vlong
mkpath(int tree, int devno)
{
	return (tree&amp;0xFFFF)&lt;&lt;16 | devno&amp;0xFFFF;
}

static int
path2treeno(int q)
{
	return q&gt;&gt;16 &amp; 0xFFFF;
}

static int
path2devno(int q)
{
	return q &amp; 0xFFFF;
}

static Tree*
gettree(int i, int mustexist)
{
	dprint("gettree %d\n", i);
	if(i &lt; 0)
		panic("#k: bug: bad tree index %d in gettree", i);
	if(i &gt;= ntrees || trees[i] == nil)
		if(mustexist)
			error(Enonexist);
		else
			return nil;
	return trees[i];
}

static Fsdev*
getdev(Tree *t, int i, int mustexist)
{
	dprint("getdev %d\n", i);
	if(i &lt; 0)
		panic("#k: bug: bad dev index %d in getdev", i);
	if(i &gt;= t-&gt;nadevs || t-&gt;devs[i] == nil)
		if(mustexist)
			error(Enonexist);
		else
			return nil;
	return t-&gt;devs[i];
}

static Fsdev*
path2dev(int q)
{
	Tree	*t;

	dprint("path2dev %ux\n", q);
	t = gettree(path2treeno(q), Mustexist);
	return getdev(t, path2devno(q) - Qfirst, Mustexist);
}

static Tree*
treealloc(char *name)
{
	int	i;
	Tree	*t;

	dprint("treealloc %s\n", name);
	for(i = 0; i &lt; nelem(trees); i++)
		if(trees[i] == nil)
			break;
	if(i == nelem(trees))
		return nil;
	t = trees[i] = mallocz(sizeof(Tree), 1);
	if(t == nil)
		return nil;
	if(i == ntrees)
		ntrees++;
	kstrdup(&amp;t-&gt;name, name);
	return t;
}

static Tree*
lookuptree(char *name)
{
	int i;

	dprint("lookuptree %s\n", name);
	for(i = 0; i &lt; ntrees; i++)
		if(trees[i] != nil &amp;&amp; strcmp(trees[i]-&gt;name, name) == 0)
			return trees[i];
	return nil;
}

static Fsdev*
devalloc(Tree *t, char *name)
{
	int	i, ndevs;
	Fsdev	*mp, **devs;

	dprint("devalloc %s %s\n", t-&gt;name, name);
	mp = mallocz(sizeof(Fsdev), 1);
	if(mp == nil)
		return nil;
	for(i = 0; i &lt; t-&gt;nadevs; i++)
		if(t-&gt;devs[i] == nil)
			break;
	if(i &gt;= t-&gt;nadevs){
		if(t-&gt;nadevs % Incr == 0){
			ndevs = t-&gt;nadevs + Incr;
			devs = realloc(t-&gt;devs, ndevs * sizeof(Fsdev*));
			if(devs == nil){
				free(mp);
				return nil;
			}
			t-&gt;devs = devs;
		}
		t-&gt;devs[t-&gt;nadevs] = nil;
		t-&gt;nadevs++;
	}
	kstrdup(&amp;mp-&gt;name, name);
	mp-&gt;vers = ++qidvers;
	mp-&gt;tree = t;
	t-&gt;devs[i] = mp;
	t-&gt;ndevs++;
	return mp;
}

static void
deltree(Tree *t)
{
	int i;

	dprint("deltree %s\n", t-&gt;name);
	for(i = 0; i &lt; ntrees; i++)
		if(trees[i] == t){
			if(i &gt; 0){		/* "fs" never goes away */
				free(t-&gt;name);
				free(t-&gt;devs);
				free(t);
				trees[i] = nil;
			}
			return;
		}
	panic("#k: deltree: bug: tree not found");
}

/*
 * A device is gone and we know that all its users are gone.
 * A tree is gone when all its devices are gone ("fs" is never gone).
 * Must close devices outside locks, so we could nest our own devices.
 */
static void
mdeldev(Fsdev *mp)
{
	int	i;
	Inner	*in;
	Tree	*t;

	dprint("deldev %s gone %d ref %uld\n", mp-&gt;name, mp-&gt;gone, mp-&gt;ref);

	mp-&gt;gone = 1;
	mp-&gt;vers = ++qidvers;

	wlock(&amp;lck);
	t = mp-&gt;tree;
	for(i = 0; i &lt; t-&gt;nadevs; i++)
		if(t-&gt;devs[i] == mp){
			t-&gt;devs[i] = nil;
			t-&gt;ndevs--;
			if(t-&gt;ndevs == 0)
				deltree(t);
			break;
		}
	wunlock(&amp;lck);

	free(mp-&gt;name);
	for(i = 0; i &lt; mp-&gt;ndevs; i++){
		in = mp-&gt;inner[i];
		if(in-&gt;idev != nil)
			cclose(in-&gt;idev);
		free(in-&gt;iname);
		free(in);
	}
	if(debug)
		memset(mp, 9, sizeof *mp);	/* poison */
	free(mp);
}

/*
 * Delete one or all devices in one or all trees.
 */
static void
mdelctl(char *tname, char *dname)
{
	int i, alldevs, alltrees, some;
	Fsdev *mp;
	Tree *t;

	dprint("delctl %s\n", dname);
	alldevs = strcmp(dname, "*") == 0;
	alltrees = strcmp(tname, "*") == 0;
	some = 0;
Again:
	wlock(&amp;lck);
	for(i = 0; i &lt; ntrees; i++){
		t = trees[i];
		if(t == nil)
			continue;
		if(alltrees == 0 &amp;&amp; strcmp(t-&gt;name, tname) != 0)
			continue;
		for(i = 0; i &lt; t-&gt;nadevs; i++){
			mp = t-&gt;devs[i];
			if(t-&gt;devs[i] == nil)
				continue;
			if(alldevs == 0 &amp;&amp; strcmp(mp-&gt;name, dname) != 0)
				continue;
			/*
			 * Careful: must close outside locks and that
			 * may change the file tree we are looking at.
			 */
			some++;
			mp-&gt;gone = 1;
			if(mp-&gt;ref == 0){
				incref(mp);	/* keep it there */
				wunlock(&amp;lck);
				mdeldev(mp);
				goto Again;	/* tree can change */
			}
		}
	}
	wunlock(&amp;lck);
	if(some == 0 &amp;&amp; alltrees == 0)
		error(Enonexist);
}

static void
setdsize(Fsdev* mp, vlong *ilen)
{
	int	i;
	vlong	inlen;
	Inner	*in;

	dprint("setdsize %s\n", mp-&gt;name);
	for (i = 0; i &lt; mp-&gt;ndevs; i++){
		in = mp-&gt;inner[i];
		in-&gt;isize = ilen[i];
		inlen = in-&gt;isize;
		switch(mp-&gt;type){
		case Finter:
			/* truncate to multiple of Blksize */
			inlen &amp;= ~(Blksize-1);
			in-&gt;isize = inlen;
			/* fall through */
		case Fmirror:
			/* use size of smallest inner device */
			if (mp-&gt;size == 0 || mp-&gt;size &gt; inlen)
				mp-&gt;size = inlen;
			break;
		case Fcat:
			mp-&gt;size += inlen;
			break;
		case Fpart:
			if(mp-&gt;start &gt; inlen)
				error("partition starts after device end");
			if(inlen &lt; mp-&gt;start + mp-&gt;size){
				print("#k: %s: partition truncated from "
					"%lld to %lld bytes\n", mp-&gt;name,
					mp-&gt;size, inlen - mp-&gt;start);
				mp-&gt;size = inlen - mp-&gt;start;
			}
			break;
		}
	}
	if(mp-&gt;type == Finter)
		mp-&gt;size *= mp-&gt;ndevs;
}

static void
validdevname(Tree *t, char *dname)
{
	int i;

	for(i = 0; i &lt; t-&gt;nadevs; i++)
		if(t-&gt;devs[i] != nil &amp;&amp; strcmp(t-&gt;devs[i]-&gt;name, dname) == 0)
			error(Eexist);
}

static void
parseconfig(char *a, long n, Cmdbuf **cbp, Cmdtab **ctp)
{
	Cmdbuf	*cb;
	Cmdtab	*ct;

	*cbp = cb = parsecmd(a, n);
	*ctp = ct = lookupcmd(cb, configs, nelem(configs));

	cb-&gt;f++;			/* skip command */
	cb-&gt;nf--;
	switch(ct-&gt;index){
	case Fmirror:
	case Fcat:
	case Finter:
		if(cb-&gt;nf &lt; 2)
			error("too few arguments for ctl");
		if(cb-&gt;nf - 1 &gt; Ndevs)
			error("too many devices in ctl");
		break;
	case Fdisk:
		if(cb-&gt;nf &lt; 1 || cb-&gt;nf &gt; 3)
			error("ctl usage: disk name [sz dev]");
		break;
	case Fpart:
		if(cb-&gt;nf != 4 &amp;&amp; (cb-&gt;nf != 3 || source == nil))
			error("ctl usage: part new [file] off len");
		break;
	}
}

static void
parsename(char *name, char *disk, char **tree, char **dev)
{
	char *slash;

	slash = strchr(name, '/');
	if(slash == nil){
		if(disk != nil)
			*tree = disk;
		else
			*tree = "fs";
		*dev = name;
	}else{
		*tree = name;
		*slash++ = 0;
		*dev = slash;
	}
	validname(*tree, 0);
	validname(*dev, 0);
}

static int
getattrs(Chan *c, vlong *lenp, int *permp)
{
	uchar	buf[128];	/* old DIRLEN plus a little should be plenty */
	Dir	d;
	long	l;

	*lenp = 0;
	*permp = 0;
	l = devtab[c-&gt;type]-&gt;stat(c, buf, sizeof buf);
	if (l &gt;= 0 &amp;&amp; convM2D(buf, l, &amp;d, nil) &gt; 0) {
		*lenp = d.length;
		*permp = d.mode &amp; 0777;
	}
	return l;
}

/*
 * Process a single line of configuration,
 * often of the form "cmd newname idev0 idev1".
 * locking is tricky, because we need a write lock to
 * add/remove devices yet adding/removing them may lead
 * to calls to this driver that require a read lock (when
 * inner devices are also provided by us).
 */
static void
mconfig(char* a, long n)
{
	int	i;
	int	*iperm;
	vlong	size, start;
	vlong	*ilen;
	char	*tname, *dname, *fakef[4];
	Chan	**idev;
	Cmdbuf	*cb;
	Cmdtab	*ct;
	Fsdev	*mp;
	Inner	*inprv;
	Tree	*t;

	/* ignore comments &amp; empty lines */
	if (*a == '\0' || *a == '#' || *a == '\n')
		return;

	dprint("mconfig\n");
	size = 0;
	start = 0;
	mp = nil;
	cb = nil;
	idev = nil;
	ilen = nil;
	iperm = nil;

	if(waserror()){
		free(cb);
		nexterror();
	}

	parseconfig(a, n, &amp;cb, &amp;ct);
	switch (ct-&gt;index) {
	case Fdisk:
		kstrdup(&amp;disk, cb-&gt;f[0]);
		if(cb-&gt;nf &gt;= 2)
			sectorsz = strtoul(cb-&gt;f[1], 0, 0);
		else
			sectorsz = Sectorsz;
		if(cb-&gt;nf == 3)
			kstrdup(&amp;source, cb-&gt;f[2]);
		else{
			free(source);
			source = nil;
		}
		poperror();
		free(cb);
		return;
	case Fclear:
		poperror();
		free(cb);
		mdelctl("*", "*");		/* del everything */
		return;
	case Fpart:
		if(cb-&gt;nf == 3){
			/*
			 * got a request in the format of sd(3),
			 * pretend we got one in our format.
			 * later we change end to be len.
			 */
			fakef[0] = cb-&gt;f[0];
			fakef[1] = source;
			fakef[2] = cb-&gt;f[1];
			fakef[3] = cb-&gt;f[2];
			cb-&gt;f = fakef;
			cb-&gt;nf = 4;
		}
		start = strtoll(cb-&gt;f[2], nil, 10);
		size =  strtoll(cb-&gt;f[3], nil, 10);
		if(cb-&gt;f == fakef)
			size -= start;		/* it was end */
		cb-&gt;nf -= 2;
		break;
	}
	parsename(cb-&gt;f[0], disk, &amp;tname, &amp;dname);
	for(i = 1; i &lt; cb-&gt;nf; i++)
		validname(cb-&gt;f[i], 1);

	if(ct-&gt;index == Fdel){
		mdelctl(tname, dname);
		poperror();
		free(cb);
		return;
	}

	/*
	 * Open all inner devices while we have only a read lock.
	 */
	poperror();
	rlock(&amp;lck);
	if(waserror()){
		runlock(&amp;lck);
Fail:
		for(i = 1; i &lt; cb-&gt;nf; i++)
			if(idev != nil &amp;&amp; idev[i-1] != nil)
				cclose(idev[i]);
		if(mp != nil)
			mdeldev(mp);
		free(idev);
		free(ilen);
		free(iperm);
		free(cb);
		nexterror();
	}
	/* record names, lengths and perms of all named files */
	idev = smalloc(sizeof(Chan*) * Ndevs);
	ilen = smalloc(sizeof(vlong) * Ndevs);
	iperm = smalloc(sizeof(int) * Ndevs);
	for(i = 1; i &lt; cb-&gt;nf; i++){
		idev[i-1] = namec(cb-&gt;f[i], Aopen, ORDWR, 0);
		getattrs(idev[i-1], &amp;ilen[i-1], &amp;iperm[i-1]);
	}
	poperror();
	runlock(&amp;lck);

	/*
	 * Get a write lock and add the device if we can.
	 */
	wlock(&amp;lck);
	if(waserror()){
		wunlock(&amp;lck);
		goto Fail;
	}

	t = lookuptree(tname);
	if(t != nil)
		validdevname(t, dname);
	else
		t = treealloc(tname);
	if(t == nil)
		error("no more trees");
	mp = devalloc(t, dname);
	if(mp == nil){
		if(t-&gt;ndevs == 0)	/* it was created for us */
			deltree(t);	/* but we will not mdeldev() */
		error(Enomem);
	}

	/* construct mp from iname, idev and iperm arrays */
	mp-&gt;type = ct-&gt;index;
	if(mp-&gt;type == Fpart){
		mp-&gt;start = start * sectorsz;
		mp-&gt;size = size * sectorsz;
	}
	mp-&gt;perm = 0666;
	for(i = 1; i &lt; cb-&gt;nf; i++){
		inprv = mp-&gt;inner[i-1] = mallocz(sizeof(Inner), 1);
		if(inprv == nil)
			error(Enomem);
		mp-&gt;ndevs++;
		kstrdup(&amp;inprv-&gt;iname, cb-&gt;f[i]);
		inprv-&gt;idev = idev[i-1];
		idev[i-1] = nil;
		/* use the most restrictive of the inner permissions */
		mp-&gt;perm &amp;= iperm[i-1];
	}
	setdsize(mp, ilen);

	poperror();
	wunlock(&amp;lck);
	free(idev);
	free(ilen);
	free(iperm);
	free(cb);
}

static void
rdconf(void)
{
	int mustrd;
	char *c, *e, *p, *s;
	Chan *cc;
	static int configed;

	/* only read config file once */
	if (configed)
		return;
	configed = 1;

	dprint("rdconf\n");
	/* add the std "fs" tree */
	trees[0] = &amp;fstree;
	ntrees++;
	fstree.name = "fs";

	/* identify the config file */
	s = getconf("fsconfig");
	if (s == nil){
		mustrd = 0;
		s = "/dev/sdC0/fscfg";
	} else
		mustrd = 1;

	/* read it */
	cc = nil;
	c = nil;
	if (waserror()){
		if (cc != nil)
			cclose(cc);
		if (c)
			free(c);
		if (!mustrd)
			return;
		nexterror();
	}
	cc = namec(s, Aopen, OREAD, 0);
	devtab[cc-&gt;type]-&gt;read(cc, confstr, sizeof confstr, 0);
	cclose(cc);
	cc = nil;

	/* validate, copy and erase config; mconfig will repopulate confstr */
	if (strncmp(confstr, cfgstr, sizeof cfgstr - 1) != 0)
		error("bad #k config, first line must be: 'fsdev:\\n'");
	kstrdup(&amp;c, confstr + sizeof cfgstr - 1);
	memset(confstr, 0, sizeof confstr);

	/* process config copy one line at a time */
	for (p = c; p != nil &amp;&amp; *p != '\0'; p = e){
		e = strchr(p, '\n');
		if (e == nil)
			e = p + strlen(p);
		else
			e++;
		mconfig(p, e - p);
	}
	USED(cc);		/* until now, can be used in waserror clause */
	poperror();
}

static int
mgen(Chan *c, char*, Dirtab*, int, int i, Dir *dp)
{
	int	treeno;
	Fsdev	*mp;
	Qid	qid;
	Tree	*t;

	dprint("mgen %#ullx %d\n", c-&gt;qid.path, i);
	qid.type = QTDIR;
	qid.vers = 0;
	if(c-&gt;qid.path == Qtop){
		if(i == DEVDOTDOT){
			devdir(c, tqid, "#k", 0, eve, DMDIR|0775, dp);
			return 1;
		}
		t = gettree(i, Optional);
		if(t == nil){
			dprint("no\n");
			return -1;
		}
		qid.path = mkpath(i, Qdir);
		devdir(c, qid, t-&gt;name, 0, eve, DMDIR|0775, dp);
		return 1;
	}

	treeno = path2treeno(c-&gt;qid.path);
	t = gettree(treeno, Optional);
	if(t == nil){
		dprint("no\n");
		return -1;
	}
	if((c-&gt;qid.type &amp; QTDIR) != 0){
		if(i == DEVDOTDOT){
			devdir(c, tqid, "#k", 0, eve, DMDIR|0775, dp);
			return 1;
		}
		if(treeno == 0){
			/* take care of #k/fs/ctl */
			if(i == 0){
				devdir(c, cqid, "ctl", 0, eve, 0664, dp);
				return 1;
			}
			i--;
		}
		mp = getdev(t, i, Optional);
		if(mp == nil){
			dprint("no\n");
			return -1;
		}
		qid.type = QTFILE;
		qid.vers = mp-&gt;vers;
		qid.path = mkpath(treeno, Qfirst+i);
		devdir(c, qid, mp-&gt;name, mp-&gt;size, eve, mp-&gt;perm, dp);
		return 1;
	}

	if(i == DEVDOTDOT){
		qid.path = mkpath(treeno, Qdir);
		devdir(c, qid, t-&gt;name, 0, eve, DMDIR|0775, dp);
		return 1;
	}
	dprint("no\n");
	return -1;
}

static Chan*
mattach(char *spec)
{
	dprint("mattach\n");
	return devattach(fsdevtab.dc, spec);
}

static Walkqid*
mwalk(Chan *c, Chan *nc, char **name, int nname)
{
	Walkqid *wq;

	rdconf();

	dprint("mwalk %llux\n", c-&gt;qid.path);
	rlock(&amp;lck);
	if(waserror()){
		runlock(&amp;lck);
		nexterror();
	}
	wq = devwalk(c, nc, name, nname, 0, 0, mgen);
	poperror();
	runlock(&amp;lck);
	return wq;
}

static int
mstat(Chan *c, uchar *db, int n)
{
	int	p;
	Dir	d;
	Fsdev	*mp;
	Qid	q;
	Tree	*t;

	dprint("mstat %llux\n", c-&gt;qid.path);
	rlock(&amp;lck);
	if(waserror()){
		runlock(&amp;lck);
		nexterror();
	}
	p = c-&gt;qid.path;
	memset(&amp;d, 0, sizeof d);
	switch(p){
	case Qtop:
		devdir(c, tqid, "#k", 0, eve, DMDIR|0775, &amp;d);
		break;
	case Qctl:
		devdir(c, cqid, "ctl", 0, eve, 0664, &amp;d);
		break;
	default:
		t = gettree(path2treeno(p), Mustexist);
		if(c-&gt;qid.type &amp; QTDIR)
			devdir(c, c-&gt;qid, t-&gt;name, 0, eve, DMDIR|0775, &amp;d);
		else{
			mp = getdev(t, path2devno(p) - Qfirst, Mustexist);
			q = c-&gt;qid;
			q.vers = mp-&gt;vers;
			devdir(c, q, mp-&gt;name, mp-&gt;size, eve, mp-&gt;perm, &amp;d);
		}
	}
	n = convD2M(&amp;d, db, n);
	if (n == 0)
		error(Ebadarg);
	poperror();
	runlock(&amp;lck);
	return n;
}

static Chan*
mopen(Chan *c, int omode)
{
	int	q;
	Fsdev	*mp;

	dprint("mopen %llux\n", c-&gt;qid.path);
	if((c-&gt;qid.type &amp; QTDIR) &amp;&amp; omode != OREAD)
		error(Eperm);
	if(c-&gt;qid.path != Qctl &amp;&amp; (c-&gt;qid.type&amp;QTDIR) == 0){
		rlock(&amp;lck);
		if(waserror()){
			runlock(&amp;lck);
			nexterror();
		}
		q = c-&gt;qid.path;
		mp = path2dev(q);
		if(mp-&gt;gone)
			error(Egone);
		devpermcheck(eve, mp-&gt;perm, omode);
		incref(mp);
		poperror();
		runlock(&amp;lck);
	}
	/*
	 * Our mgen does not return the info for the qid
	 * but only for its children. Don't use devopen here.
	 */
	c-&gt;offset = 0;
	c-&gt;mode = openmode(omode &amp; ~OTRUNC);
	c-&gt;flag |= COPEN;
	return c;
}

static void
mclose(Chan *c)
{
	int	mustdel, q;
	Fsdev	*mp;

	dprint("mclose %llux\n", c-&gt;qid.path);
	if(c-&gt;qid.type &amp; QTDIR || !(c-&gt;flag &amp; COPEN))
		return;
	rlock(&amp;lck);
	if(waserror()){
		runlock(&amp;lck);
		nexterror();
	}
	mustdel = 0;
	mp = nil;
	q = c-&gt;qid.path;
	if(q == Qctl){
		free(disk);
		disk = nil;	/* restore defaults */
		free(source);
		source = nil;
		sectorsz = Sectorsz;
	}else{
		mp = path2dev(q);
		if(mp-&gt;gone != 0 &amp;&amp; mp-&gt;ref == 1)
			mustdel = 1;
		else
			decref(mp);
	}
	poperror();
	runlock(&amp;lck);
	if(mustdel)
		mdeldev(mp);
}

static long
io(Fsdev *mp, Inner *in, int isread, void *a, long l, vlong off)
{
	long wl;
	Chan	*mc;

	mc = in-&gt;idev;
	if(mc == nil)
		error(Egone);
	if (waserror()) {
		print("#k: %s: byte %,lld count %ld (of #k/%s): %s error: %s\n",
			in-&gt;iname, off, l, mp-&gt;name, (isread? "read": "write"),
			(up &amp;&amp; up-&gt;errstr? up-&gt;errstr: ""));
		nexterror();
	}
	if (isread)
		wl = devtab[mc-&gt;type]-&gt;read(mc, a, l, off);
	else
		wl = devtab[mc-&gt;type]-&gt;write(mc, a, l, off);
	poperror();
	return wl;
}

/* NB: a transfer could span multiple inner devices */
static long
catio(Fsdev *mp, int isread, void *a, long n, vlong off)
{
	int	i;
	long	l, res;
	Inner	*in;

	if(debug)
		print("catio %d %p %ld %lld\n", isread, a, n, off);
	res = n;
	for (i = 0; n &gt; 0 &amp;&amp; i &lt; mp-&gt;ndevs; i++){
		in = mp-&gt;inner[i];
		if (off &gt;= in-&gt;isize){
			off -= in-&gt;isize;
			continue;		/* not there yet */
		}
		if (off + n &gt; in-&gt;isize)
			l = in-&gt;isize - off;
		else
			l = n;
		if(debug)
			print("\tdev %d %p %ld %lld\n", i, a, l, off);

		if (io(mp, in, isread, a, l, off) != l)
			error(Eio);

		a = (char*)a + l;
		off = 0;
		n -= l;
	}
	if(debug)
		print("\tres %ld\n", res - n);
	return res - n;
}

static long
interio(Fsdev *mp, int isread, void *a, long n, vlong off)
{
	int	i;
	long	boff, res, l, wl, wsz;
	vlong	woff, blk, mblk;

	blk  = off / Blksize;
	boff = off % Blksize;
	wsz  = Blksize - boff;
	res = n;
	while(n &gt; 0){
		mblk = blk / mp-&gt;ndevs;
		i    = blk % mp-&gt;ndevs;
		woff = mblk*Blksize + boff;
		if (n &gt; wsz)
			l = wsz;
		else
			l = n;

		wl = io(mp, mp-&gt;inner[i], isread, a, l, woff);
		if (wl != l)
			error(Eio);

		blk++;
		boff = 0;
		wsz = Blksize;
		a = (char*)a + l;
		n -= l;
	}
	return res;
}

static char*
seprintconf(char *s, char *e)
{
	int	i, j;
	Tree	*t;

	*s = 0;
	for(i = 0; i &lt; ntrees; i++){
		t = trees[i];
		if(t != nil)
			for(j = 0; j &lt; t-&gt;nadevs; j++)
				if(t-&gt;devs[j] != nil)
					s = seprintdev(s, e, t-&gt;devs[j]);
	}
	return s;
}

static long
mread(Chan *c, void *a, long n, vlong off)
{
	int	i, retry;
	long	l, res;
	Fsdev	*mp;
	Tree	*t;

	dprint("mread %llux\n", c-&gt;qid.path);
	rlock(&amp;lck);
	if(waserror()){
		runlock(&amp;lck);
		nexterror();
	}
	res = -1;
	if(c-&gt;qid.type &amp; QTDIR){
		res = devdirread(c, a, n, 0, 0, mgen);
		goto Done;
	}
	if(c-&gt;qid.path == Qctl){
		seprintconf(confstr, confstr + sizeof(confstr));
		res = readstr((long)off, a, n, confstr);
		goto Done;
	}

	t = gettree(path2treeno(c-&gt;qid.path), Mustexist);
	mp = getdev(t, path2devno(c-&gt;qid.path) - Qfirst, Mustexist);

	if(off &gt;= mp-&gt;size){
		res = 0;
		goto Done;
	}
	if(off + n &gt; mp-&gt;size)
		n = mp-&gt;size - off;
	if(n == 0){
		res = 0;
		goto Done;
	}

	switch(mp-&gt;type){
	case Fcat:
		res = catio(mp, Isread, a, n, off);
		break;
	case Finter:
		res = interio(mp, Isread, a, n, off);
		break;
	case Fpart:
		res = io(mp, mp-&gt;inner[0], Isread, a, n, mp-&gt;start + off);
		break;
	case Fmirror:
		retry = 0;
		do {
			if (retry &gt; 0) {
				print("#k/%s: retry %d read for byte %,lld "
					"count %ld: %s\n", mp-&gt;name, retry, off,
					n, (up &amp;&amp; up-&gt;errstr? up-&gt;errstr: ""));
				/*
				 * pause before retrying in case it's due to
				 * a transient bus or controller problem.
				 */
				tsleep(&amp;up-&gt;sleep, return0, 0, Retrypause);
			}
			for (i = 0; i &lt; mp-&gt;ndevs; i++){
				if (waserror())
					continue;
				l = io(mp, mp-&gt;inner[i], Isread, a, n, off);
				poperror();
				if (l &gt;= 0){
					res = l;
					break;		/* read a good copy */
				}
			}
		} while (i == mp-&gt;ndevs &amp;&amp; ++retry &lt;= Maxretries);
		if (retry &gt; Maxretries) {
			/* no mirror had a good copy of the block */
			print("#k/%s: byte %,lld count %ld: CAN'T READ "
				"from mirror: %s\n", mp-&gt;name, off, n,
				(up &amp;&amp; up-&gt;errstr? up-&gt;errstr: ""));
			error(Eio);
		} else if (retry &gt; 0)
			print("#k/%s: byte %,lld count %ld: retry read OK "
				"from mirror: %s\n", mp-&gt;name, off, n,
				(up &amp;&amp; up-&gt;errstr? up-&gt;errstr: ""));
		break;
	}
Done:
	poperror();
	runlock(&amp;lck);
	return res;
}

static long
mwrite(Chan *c, void *a, long n, vlong off)
{
	int	i, allbad, anybad, retry;
	long	l, res;
	Fsdev	*mp;
	Tree	*t;

	dprint("mwrite %llux\n", c-&gt;qid.path);
	if (c-&gt;qid.type &amp; QTDIR)
		error(Eisdir);
	if (c-&gt;qid.path == Qctl){
		mconfig(a, n);
		return n;
	}

	rlock(&amp;lck);
	if(waserror()){
		runlock(&amp;lck);
		nexterror();
	}

	t = gettree(path2treeno(c-&gt;qid.path), Mustexist);
	mp = getdev(t, path2devno(c-&gt;qid.path) - Qfirst, Mustexist);

	if(off &gt;= mp-&gt;size){
		res = 0;
		goto Done;
	}
	if(off + n &gt; mp-&gt;size)
		n = mp-&gt;size - off;
	if(n == 0){
		res = 0;
		goto Done;
	}
	res = n;
	switch(mp-&gt;type){
	case Fcat:
		res = catio(mp, Iswrite, a, n, off);
		break;
	case Finter:
		res = interio(mp, Iswrite, a, n, off);
		break;
	case Fpart:
		res = io(mp, mp-&gt;inner[0], Iswrite, a, n, mp-&gt;start + off);
		if (res != n)
			error(Eio);
		break;
	case Fmirror:
		retry = 0;
		do {
			if (retry &gt; 0) {
				print("#k/%s: retry %d write for byte %,lld "
					"count %ld: %s\n", mp-&gt;name, retry, off,
					n, (up &amp;&amp; up-&gt;errstr? up-&gt;errstr: ""));
				/*
				 * pause before retrying in case it's due to
				 * a transient bus or controller problem.
				 */
				tsleep(&amp;up-&gt;sleep, return0, 0, Retrypause);
			}
			allbad = 1;
			anybad = 0;
			for (i = mp-&gt;ndevs - 1; i &gt;= 0; i--){
				if (waserror()) {
					anybad = 1;
					continue;
				}
				l = io(mp, mp-&gt;inner[i], Iswrite, a, n, off);
				poperror();
				if (l == n)
					allbad = 0;	/* wrote a good copy */
				else
					anybad = 1;
			}
		} while (anybad &amp;&amp; ++retry &lt;= Maxretries);
		if (allbad) {
			/* no mirror took a good copy of the block */
			print("#k/%s: byte %,lld count %ld: CAN'T WRITE "
				"to mirror: %s\n", mp-&gt;name, off, n,
				(up &amp;&amp; up-&gt;errstr? up-&gt;errstr: ""));
			error(Eio);
		} else if (retry &gt; 0)
			print("#k/%s: byte %,lld count %ld: retry wrote OK "
				"to mirror: %s\n", mp-&gt;name, off, n,
				(up &amp;&amp; up-&gt;errstr? up-&gt;errstr: ""));

		break;
	}
Done:
	poperror();
	runlock(&amp;lck);
	return res;
}

Dev fsdevtab = {
	'k',
	"fs",

	devreset,
	devinit,
	devshutdown,
	mattach,
	mwalk,
	mstat,
	mopen,
	devcreate,
	mclose,
	mread,
	devbread,
	mwrite,
	devbwrite,
	devremove,
	devwstat,
	devpower,
	devconfig,
};
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

