<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/9660srv/iobuf.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/9660srv/iobuf.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;auth.h&gt;
#include &lt;fcall.h&gt;
#include "dat.h"
#include "fns.h"

/*
 * We used to use 100 i/o buffers of size 2kb (Sectorsize).
 * Unfortunately, reading 2kb at a time often hopping around
 * the disk doesn't let us get near the disk bandwidth.
 *
 * Based on a trace of iobuf address accesses taken while
 * tarring up a Plan 9 distribution CD, we now use 16 128kb
 * buffers.  This works for ISO9660 because data is required
 * to be laid out contiguously; effectively we're doing agressive
 * readahead.  Because the buffers are so big and the typical 
 * disk accesses so concentrated, it's okay that we have so few
 * of them.
 *
 * If this is used to access multiple discs at once, it's not clear
 * how gracefully the scheme degrades, but I'm not convinced
 * it's worth worrying about.		-rsc
 */

/* trying a larger value to get greater throughput - geoff */
#define	BUFPERCLUST	256 /* sectors/cluster; was 64, 64*Sectorsize = 128kb */
#define	NCLUST		16

int nclust = NCLUST;

static Ioclust*	iohead;
static Ioclust*	iotail;

static Ioclust*	getclust(Xdata*, vlong);
static void	putclust(Ioclust*);
static void	xread(Ioclust*);

void
iobuf_init(void)
{
	int i, j, n;
	Ioclust *c;
	Iobuf *b;
	uchar *mem;

	n = nclust*sizeof(Ioclust) +
		nclust*BUFPERCLUST*(sizeof(Iobuf)+Sectorsize);
	mem = sbrk(n);
	if(mem == (void*)-1)
		panic(0, "iobuf_init");
	memset(mem, 0, n);

	for(i=0; i&lt;nclust; i++){
		c = (Ioclust*)mem;
		mem += sizeof(Ioclust);
		c-&gt;addr = -1;
		c-&gt;prev = iotail;
		if(iotail)
			iotail-&gt;next = c;
		iotail = c;
		if(iohead == nil)
			iohead = c;

		c-&gt;buf = (Iobuf*)mem;
		mem += BUFPERCLUST*sizeof(Iobuf);
		c-&gt;iobuf = mem;
		mem += BUFPERCLUST*Sectorsize;
		for(j=0; j&lt;BUFPERCLUST; j++){
			b = &amp;c-&gt;buf[j];
			b-&gt;clust = c;
			b-&gt;addr = -1;
			b-&gt;iobuf = c-&gt;iobuf+j*Sectorsize;
		}
	}
}

void
purgebuf(Xdata *dev)
{
	Ioclust *p;

	for(p=iohead; p!=nil; p=p-&gt;next)
		if(p-&gt;dev == dev){
			p-&gt;addr = -1;
			p-&gt;busy = 0;
		}
}

static Ioclust*
getclust(Xdata *dev, vlong addr)
{
	Ioclust *c, *f;

	f = nil;
	for(c=iohead; c; c=c-&gt;next){
		if(!c-&gt;busy)
			f = c;
		if(c-&gt;addr == addr &amp;&amp; c-&gt;dev == dev){
			c-&gt;busy++;
			return c;
		}
	}

	if(f == nil)
		panic(0, "out of buffers");

	f-&gt;addr = addr;
	f-&gt;dev = dev;
	f-&gt;busy++;
	if(waserror()){
		f-&gt;addr = -1;	/* stop caching */
		putclust(f);
		nexterror();
	}
	xread(f);
	poperror();
	return f;
}

static void
putclust(Ioclust *c)
{
	if(c-&gt;busy &lt;= 0)
		panic(0, "putbuf");
	c-&gt;busy--;

	/* Link onto head for LRU */
	if(c == iohead)
		return;
	c-&gt;prev-&gt;next = c-&gt;next;

	if(c-&gt;next)
		c-&gt;next-&gt;prev = c-&gt;prev;
	else
		iotail = c-&gt;prev;

	c-&gt;prev = nil;
	c-&gt;next = iohead;
	iohead-&gt;prev = c;
	iohead = c;
}

Iobuf*
getbuf(Xdata *dev, uvlong addr)
{
	int off;
	Ioclust *c;

	off = addr%BUFPERCLUST;
	c = getclust(dev, addr - off);
	if(c-&gt;nbuf &lt; off){
		c-&gt;busy--;
		error("short read or I/O error");
	}
	return &amp;c-&gt;buf[off];
}

void
putbuf(Iobuf *b)
{
	putclust(b-&gt;clust);
}

static void
xread(Ioclust *c)
{
	int n;
	Xdata *dev;

	dev = c-&gt;dev;
	seek(dev-&gt;dev, (vlong)c-&gt;addr * Sectorsize, 0);
	n = readn(dev-&gt;dev, c-&gt;iobuf, BUFPERCLUST*Sectorsize);
	if(n &lt; Sectorsize)
		error("short read or I/O error");
	c-&gt;nbuf = n/Sectorsize;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

