<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/auth/secureidcheck.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/auth/secureidcheck.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * This code uses RADIUS as a portable way to validate tokens such as SecurID.
 * It is relatively simple to send a UDP packet and get a response, but various
 * things can go wrong.  Speaking the proprietary ACE protocol would allow
 * handling "next token code" and other error messages.  More importantly, the
 * timeout threshold is inherently hard to pick.  We observe responses taking
 * longer than 10 seconds in normal times.  That is a long time to wait before
 * retrying on a second server.  Moreover, if the UDP response is lost, retrying
 * on a second server will also fail because the valid token code may be
 * presented only once.  This whole approach is flawed, but best we can do.
 */
/* RFC2138 */
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;ip.h&gt;
#include &lt;ctype.h&gt;
#include &lt;mp.h&gt;
#include &lt;libsec.h&gt;
#include &lt;bio.h&gt;
#include &lt;ndb.h&gt;

#define AUTHLOG "auth"

enum{
	R_AccessRequest	=1,	/* Packet code */
	R_AccessAccept	=2,
	R_AccessReject	=3,
	R_AccessChallenge=11,
	R_UserName	=1,
	R_UserPassword	=2,
	R_NASIPAddress	=4,
	R_ReplyMessage	=18,
	R_State		=24,
	R_NASIdentifier	=32,
};

typedef struct Secret{
	uchar	*s;
	int	len;
} Secret;

typedef struct Attribute{
	struct Attribute *next;
	uchar	type;
	uchar	len;		/* number of bytes in value */
	uchar	val[256];
} Attribute;

typedef struct Packet{
	uchar	code, ID;
	uchar	authenticator[16];
	Attribute first;
} Packet;

/* assumes pass is at most 16 chars */
void
hide(Secret *shared, uchar *auth, Secret *pass, uchar *x)
{
	DigestState *M;
	int i, n = pass-&gt;len;

	M = md5(shared-&gt;s, shared-&gt;len, nil, nil);
	md5(auth, 16, x, M);
	if(n &gt; 16)
		n = 16;
	for(i = 0; i &lt; n; i++)
		x[i] ^= pass-&gt;s[i];
}

int
authcmp(Secret *shared, uchar *buf, int m, uchar *auth)
{
	DigestState *M;
	uchar x[16];

	M = md5(buf, 4, nil, nil);	/* Code+ID+Length */
	M = md5(auth, 16, nil, M);	/* RequestAuth */
	M = md5(buf+20, m-20, nil, M);	/* Attributes */
	md5(shared-&gt;s, shared-&gt;len, x, M);
	return memcmp(x, buf+4, 16);
}

Packet*
newRequest(uchar *auth)
{
	static uchar ID = 0;
	Packet *p;

	p = (Packet*)malloc(sizeof(*p));
	if(p == nil)
		return nil;
	p-&gt;code = R_AccessRequest;
	p-&gt;ID = ++ID;
	memmove(p-&gt;authenticator, auth, 16);
	p-&gt;first.next = nil;
	p-&gt;first.type = 0;
	return p;
}

void
freePacket(Packet *p)
{
	Attribute *a, *x;

	if(!p)
		return;
	a = p-&gt;first.next;
	while(a){
		x = a;
		a = a-&gt;next;
		free(x);
	}
	free(p);
}

int
ding(void*, char *msg)
{
	syslog(0, AUTHLOG, "ding %s", msg);
	if(strstr(msg, "alarm"))
		return 1;
	return 0;
}

Packet *
rpc(char *dest, Secret *shared, Packet *req)
{
	uchar buf[4096], buf2[4096], *b, *e;
	Packet *resp;
	Attribute *a;
	int m, n, fd, try;

	/* marshal request */
	e = buf + sizeof buf;
	buf[0] = req-&gt;code;
	buf[1] = req-&gt;ID;
	memmove(buf+4, req-&gt;authenticator, 16);
	b = buf+20;
	for(a = &amp;req-&gt;first; a; a = a-&gt;next){
		if(b + 2 + a-&gt;len &gt; e)
			return nil;
		*b++ = a-&gt;type;
		*b++ = 2 + a-&gt;len;
		memmove(b, a-&gt;val, a-&gt;len);
		b += a-&gt;len;
	}
	n = b-buf;
	buf[2] = n&gt;&gt;8;
	buf[3] = n;

	/* send request, wait for reply */
	fd = dial(dest, 0, 0, 0);
	if(fd &lt; 0){
		syslog(0, AUTHLOG, "%s: rpc can't get udp channel", dest);
		return nil;
	}
	atnotify(ding, 1);
	m = -1;
	for(try = 0; try &lt; 2; try++){
		/*
		 * increased timeout from 4sec to 15sec because
		 * corporate server really takes that long.
		 */
		alarm(15000);
		m = write(fd, buf, n);
		if(m != n){
			syslog(0, AUTHLOG, "%s: rpc write err %d %d: %r",
				dest, m, n);
			m = -1;
			break;
		}
		m = read(fd, buf2, sizeof buf2);
		alarm(0);
		if(m &lt; 0){
			syslog(0, AUTHLOG, "%s rpc read err %d: %r", dest, m);
			break;			/* failure */
		}
		if(m == 0 || buf2[1] != buf[1]){	/* need matching ID */
			syslog(0, AUTHLOG, "%s unmatched reply %d", dest, m);
			continue;
		}
		if(authcmp(shared, buf2, m, buf+4) == 0)
			break;
		syslog(0, AUTHLOG, "%s bad rpc chksum", dest);
	}
	close(fd);
	if(m &lt;= 0)
		return nil;

	/* unmarshal reply */
	b = buf2;
	e = buf2+m;
	resp = (Packet*)malloc(sizeof(*resp));
	if(resp == nil)
		return nil;
	resp-&gt;code = *b++;
	resp-&gt;ID = *b++;
	n = *b++;
	n = (n&lt;&lt;8) | *b++;
	if(m != n){
		syslog(0, AUTHLOG, "rpc got %d bytes, length said %d", m, n);
		if(m &gt; n)
			e = buf2+n;
	}
	memmove(resp-&gt;authenticator, b, 16);
	b += 16;
	a = &amp;resp-&gt;first;
	a-&gt;type = 0;
	for(;;){
		if(b &gt;= e){
			a-&gt;next = nil;
			break;		/* exit loop */
		}
		a-&gt;type = *b++;
		a-&gt;len = (*b++) - 2;
		if(b + a-&gt;len &gt; e){	/* corrupt packet */
			a-&gt;next = nil;
			freePacket(resp);
			return nil;
		}
		memmove(a-&gt;val, b, a-&gt;len);
		b += a-&gt;len;
		if(b &lt; e){		/* any more attributes? */
			a-&gt;next = (Attribute*)malloc(sizeof(*a));
			if(a-&gt;next == nil){
				free(req);
				return nil;
			}
			a = a-&gt;next;
		}
	}
	return resp;
}

int
setAttribute(Packet *p, uchar type, uchar *s, int n)
{
	Attribute *a;

	a = &amp;p-&gt;first;
	if(a-&gt;type != 0){
		a = (Attribute*)malloc(sizeof(*a));
		if(a == nil)
			return -1;
		a-&gt;next = p-&gt;first.next;
		p-&gt;first.next = a;
	}
	a-&gt;type = type;
	a-&gt;len = n;
	if(a-&gt;len &gt; 253)	/* RFC2138, section 5 */
		a-&gt;len = 253;
	memmove(a-&gt;val, s, a-&gt;len);
	return 0;
}

/* return a reply message attribute string */
char*
replymsg(Packet *p)
{
	Attribute *a;
	static char buf[255];

	for(a = &amp;p-&gt;first; a; a = a-&gt;next)
		if(a-&gt;type == R_ReplyMessage){
			if(a-&gt;len &gt;= sizeof buf)
				a-&gt;len = sizeof(buf)-1;
			memmove(buf, a-&gt;val, a-&gt;len);
			buf[a-&gt;len] = 0;
		}
	return buf;
}

/* for convenience while debugging */
char *replymess;
Attribute *stateattr;

void
logPacket(Packet *p)
{
	int i;
	char *np, *e;
	char buf[255], pbuf[4*1024];
	uchar *au = p-&gt;authenticator;
	Attribute *a;

	e = pbuf + sizeof(pbuf);

	np = seprint(pbuf, e, "Packet ID=%d auth=%x %x %x... ",
		p-&gt;ID, au[0], au[1], au[2]);
	switch(p-&gt;code){
	case R_AccessRequest:
		np = seprint(np, e, "request\n");
		break;
	case R_AccessAccept:
		np = seprint(np, e, "accept\n");
		break;
	case R_AccessReject:
		np = seprint(np, e, "reject\n");
		break;
	case R_AccessChallenge:
		np = seprint(np, e, "challenge\n");
		break;
	default:
		np = seprint(np, e, "code=%d\n", p-&gt;code);
		break;
	}
	replymess = "0000000";
	for(a = &amp;p-&gt;first; a; a = a-&gt;next){
		if(a-&gt;len &gt; 253 )
			a-&gt;len = 253;
		memmove(buf, a-&gt;val, a-&gt;len);
		np = seprint(np, e, " [%d]", a-&gt;type);
		for(i = 0; i &lt; a-&gt;len; i++)
			if(isprint(a-&gt;val[i]))
				np = seprint(np, e, "%c", a-&gt;val[i]);
			else
				np = seprint(np, e, "\\%o", a-&gt;val[i]);
		np = seprint(np, e, "\n");
		buf[a-&gt;len] = 0;
		if(a-&gt;type == R_ReplyMessage)
			replymess = strdup(buf);
		else if(a-&gt;type == R_State)
			stateattr = a;
	}

	syslog(0, AUTHLOG, "%s", pbuf);
}

static uchar*
getipv4addr(void)
{
	Ipifc *nifc;
	Iplifc *lifc;
	static Ipifc *ifc;

	ifc = readipifc("/net", ifc, -1);
	for(nifc = ifc; nifc; nifc = nifc-&gt;next)
		for(lifc = nifc-&gt;lifc; lifc; lifc = lifc-&gt;next)
			if (ipcmp(lifc-&gt;ip, IPnoaddr) != 0 &amp;&amp;
			    ipcmp(lifc-&gt;ip, v4prefix) != 0)
				return lifc-&gt;ip;
	return nil;
}

extern Ndb *db;

/* returns 0 on success, error message on failure */
char*
secureidcheck(char *user, char *response)
{
	char *radiussecret = nil;
	char *rv = "authentication failed";
	char dest[3*IPaddrlen+20], ruser[64];
	uchar *ip;
	uchar x[16];
	ulong u[4];
	Ndbs s;
	Ndbtuple *t = nil, *nt, *tt;
	Packet *req = nil, *resp = nil;
	Secret shared, pass;
	static Ndb *netdb;

	if(netdb == nil)
		netdb = ndbopen(0);

	/* bad responses make them disable the fob, avoid silly checks */
	if(strlen(response) &lt; 4 || strpbrk(response,"abcdefABCDEF") != nil)
		goto out;

	/* get radius secret */
	radiussecret = ndbgetvalue(db, &amp;s, "radius", "lra-radius", "secret", &amp;t);
	if(radiussecret == nil){
		syslog(0, AUTHLOG, "secureidcheck: nil radius secret: %r");
		goto out;
	}

	/* translate user name if we have to */
	strcpy(ruser, user);
	for(nt = t; nt; nt = nt-&gt;entry)
		if(strcmp(nt-&gt;attr, "uid") == 0 &amp;&amp; strcmp(nt-&gt;val, user) == 0)
			for(tt = nt-&gt;line; tt != nt; tt = tt-&gt;line)
				if(strcmp(tt-&gt;attr, "rid") == 0){
					strcpy(ruser, tt-&gt;val);
					break;
				}
	ndbfree(t);
	t = nil;

	u[0] = fastrand();
	u[1] = fastrand();
	u[2] = fastrand();
	u[3] = fastrand();
	req = newRequest((uchar*)u);
	if(req == nil)
		goto out;
	shared.s = (uchar*)radiussecret;
	shared.len = strlen(radiussecret);
	ip = getipv4addr();
	if(ip == nil){
		syslog(0, AUTHLOG, "no interfaces: %r\n");
		goto out;
	}
	if(setAttribute(req, R_NASIPAddress, ip + IPv4off, 4) &lt; 0)
		goto out;

	if(setAttribute(req, R_UserName, (uchar*)ruser, strlen(ruser)) &lt; 0)
		goto out;
	pass.s = (uchar*)response;
	pass.len = strlen(response);
	hide(&amp;shared, req-&gt;authenticator, &amp;pass, x);
	if(setAttribute(req, R_UserPassword, x, 16) &lt; 0)
		goto out;

	t = ndbsearch(netdb, &amp;s, "sys", "lra-radius");
	if(t == nil){
		syslog(0, AUTHLOG, "secureidcheck: nil radius sys search: %r\n");
		goto out;
	}
	for(nt = t; nt; nt = nt-&gt;entry){
		if(strcmp(nt-&gt;attr, "ip") != 0)
			continue;

		snprint(dest, sizeof dest, "udp!%s!radius", nt-&gt;val);
		resp = rpc(dest, &amp;shared, req);
		if(resp == nil){
			syslog(0, AUTHLOG, "%s nil response", dest);
			continue;
		}
		if(resp-&gt;ID != req-&gt;ID){
			syslog(0, AUTHLOG, "%s mismatched ID  req=%d resp=%d",
				dest, req-&gt;ID, resp-&gt;ID);
			freePacket(resp);
			resp = nil;
			continue;
		}

		switch(resp-&gt;code){
		case R_AccessAccept:
			syslog(0, AUTHLOG, "%s accepted ruser=%s", dest, ruser);
			rv = nil;
			break;
		case R_AccessReject:
			syslog(0, AUTHLOG, "%s rejected ruser=%s %s",
				dest, ruser, replymsg(resp));
			rv = "secureid failed";
			break;
		case R_AccessChallenge:
			syslog(0, AUTHLOG, "%s challenge ruser=%s %s",
				dest, ruser, replymsg(resp));
			rv = "secureid out of sync";
			break;
		default:
			syslog(0, AUTHLOG, "%s code=%d ruser=%s %s",
				dest, resp-&gt;code, ruser, replymsg(resp));
			break;
		}
		break;	/* we have a proper reply, no need to ask again */
	}
out:
	if (t)
		ndbfree(t);
	free(radiussecret);
	freePacket(req);
	freePacket(resp);
	return rv;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

