<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/cdfs/scsi.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/cdfs/scsi.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * Now thread-safe.
 *
 * The codeqlock guarantees that once codes != nil, that pointer will never 
 * change nor become invalid.
 *
 * The QLock in the Scsi structure moderates access to the raw device.
 * We should probably export some of the already-locked routines, but
 * there hasn't been a need.
 */

#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;disk.h&gt;

enum {
	/* commands */
	Testrdy		= 0x00,
	Reqsense	= 0x03,
	Write10		= 0x2a,
	Writever10	= 0x2e,
	Readtoc		= 0x43,

	/* sense[2] (key) sense codes */
	Sensenone	= 0,
	Sensenotrdy	= 2,
	Sensebadreq	= 5,

	/* sense[12] (asc) sense codes */
	Lunnotrdy	= 0x04,
	Recovnoecc	= 0x17,
	Recovecc	= 0x18,
	Badcdb		= 0x24,
	Newmedium	= 0x28,
	Nomedium	= 0x3a,
};

int scsiverbose;

#define codefile "/sys/lib/scsicodes"

static char *codes;
static QLock codeqlock;

static void
getcodes(void)
{
	Dir *d;
	int n, fd;

	if(codes != nil)
		return;

	qlock(&amp;codeqlock);
	if(codes != nil) {
		qunlock(&amp;codeqlock);
		return;
	}

	if((d = dirstat(codefile)) == nil || (fd = open(codefile, OREAD)) &lt; 0) {
		qunlock(&amp;codeqlock);
		return;
	}

	codes = malloc(1+d-&gt;length+1);
	if(codes == nil) {
		close(fd);
		qunlock(&amp;codeqlock);
		free(d);
		return;
	}

	codes[0] = '\n';	/* for searches */
	n = readn(fd, codes+1, d-&gt;length);
	close(fd);
	free(d);

	if(n &lt; 0) {
		free(codes);
		codes = nil;
		qunlock(&amp;codeqlock);
		return;
	}
	codes[n] = '\0';
	qunlock(&amp;codeqlock);
}
	
char*
scsierror(int asc, int ascq)
{
	char *p, *q;
	static char search[32];
	static char buf[128];

	getcodes();

	if(codes) {
		snprint(search, sizeof search, "\n%.2ux%.2ux ", asc, ascq);
		if(p = strstr(codes, search)) {
			p += 6;
			if((q = strchr(p, '\n')) == nil)
				q = p+strlen(p);
			snprint(buf, sizeof buf, "%.*s", (int)(q-p), p);
			return buf;
		}

		snprint(search, sizeof search, "\n%.2ux00", asc);
		if(p = strstr(codes, search)) {
			p += 6;
			if((q = strchr(p, '\n')) == nil)
				q = p+strlen(p);
			snprint(buf, sizeof buf, "(ascq #%.2ux) %.*s", ascq, (int)(q-p), p);
			return buf;
		}
	}

	snprint(buf, sizeof buf, "scsi #%.2ux %.2ux", asc, ascq);
	return buf;
}


static int
_scsicmd(Scsi *s, uchar *cmd, int ccount, void *data, int dcount, int io, int dolock)
{
	uchar resp[16];
	int n;
	long status;

	if(dolock)
		qlock(s);
	if(write(s-&gt;rawfd, cmd, ccount) != ccount) {
		werrstr("cmd write: %r");
		if(dolock)
			qunlock(s);
		return -1;
	}

	switch(io){
	case Sread:
		n = read(s-&gt;rawfd, data, dcount);
		/* read toc errors are frequent and not very interesting */
		if(n &lt; 0 &amp;&amp; (scsiverbose == 1 ||
		    scsiverbose == 2 &amp;&amp; cmd[0] != Readtoc))
			fprint(2, "dat read: %r: cmd 0x%2.2uX\n", cmd[0]);
		break;
	case Swrite:
		n = write(s-&gt;rawfd, data, dcount);
		if(n != dcount &amp;&amp; scsiverbose)
			fprint(2, "dat write: %r: cmd 0x%2.2uX\n", cmd[0]);
		break;
	default:
	case Snone:
		n = write(s-&gt;rawfd, resp, 0);
		if(n != 0 &amp;&amp; scsiverbose)
			fprint(2, "none write: %r: cmd 0x%2.2uX\n", cmd[0]);
		break;
	}

	memset(resp, 0, sizeof(resp));
	if(read(s-&gt;rawfd, resp, sizeof(resp)) &lt; 0) {
		werrstr("resp read: %r\n");
		if(dolock)
			qunlock(s);
		return -1;
	}
	if(dolock)
		qunlock(s);

	resp[sizeof(resp)-1] = '\0';
	status = atoi((char*)resp);
	if(status == 0)
		return n;

	werrstr("cmd %2.2uX: status %luX dcount %d n %d", cmd[0], status, dcount, n);
	return -1;
}

int
scsicmd(Scsi *s, uchar *cmd, int ccount, void *data, int dcount, int io)
{
	return _scsicmd(s, cmd, ccount, data, dcount, io, 1);
}

static int
_scsiready(Scsi *s, int dolock)
{
	char err[ERRMAX];
	uchar cmd[6], resp[16];
	int status, i;

	if(dolock)
		qlock(s);
	werrstr("");
	for(i=0; i&lt;3; i++) {
		memset(cmd, 0, sizeof(cmd));
		cmd[0] = Testrdy;	/* unit ready */
		if(write(s-&gt;rawfd, cmd, sizeof(cmd)) != sizeof(cmd)) {
			if(scsiverbose)
				fprint(2, "ur cmd write: %r\n");
			werrstr("short unit-ready raw write");
			continue;
		}
		write(s-&gt;rawfd, resp, 0);
		if(read(s-&gt;rawfd, resp, sizeof(resp)) &lt; 0) {
			if(scsiverbose)
				fprint(2, "ur resp read: %r\n");
			continue;
		}
		resp[sizeof(resp)-1] = '\0';
		status = atoi((char*)resp);
		if(status == 0 || status == 0x02) {
			if(dolock)
				qunlock(s);
			return 0;
		}
		if(scsiverbose)
			fprint(2, "target: bad status: %x\n", status);
	}
	rerrstr(err, sizeof err);
	if(err[0] == '\0')
		werrstr("unit did not become ready");
	if(dolock)
		qunlock(s);
	return -1;
}

int
scsiready(Scsi *s)
{
	return _scsiready(s, 1);
}

int
scsi(Scsi *s, uchar *cmd, int ccount, void *v, int dcount, int io)
{
	uchar req[6], sense[255], *data;
	int tries, code, key, n;
	char *p;

	data = v;
	SET(key, code);
	qlock(s);
	for(tries=0; tries&lt;2; tries++) {
		n = _scsicmd(s, cmd, ccount, data, dcount, io, 0);
		if(n &gt;= 0) {
			qunlock(s);
			return n;
		}

		/*
		 * request sense
		 */
		memset(req, 0, sizeof(req));
		req[0] = Reqsense;
		req[4] = sizeof(sense);
		memset(sense, 0xFF, sizeof(sense));
		if((n=_scsicmd(s, req, sizeof(req), sense, sizeof(sense), Sread, 0)) &lt; 14)
			if(scsiverbose)
				fprint(2, "reqsense scsicmd %d: %r\n", n);
	
		if(_scsiready(s, 0) &lt; 0)
			if(scsiverbose)
				fprint(2, "unit not ready\n");
	
		key = sense[2] &amp; 0xf;
		code = sense[12];			/* asc */
		if(code == Recovnoecc || code == Recovecc) { /* recovered errors */
			qunlock(s);
			return dcount;
		}

		/* retry various odd cases */
		if(code == Newmedium &amp;&amp; cmd[0] == Readtoc) {
			/* read toc and media changed */
			s-&gt;nchange++;
			s-&gt;changetime = time(0);
		} else if((cmd[0] == Write10 || cmd[0] == Writever10) &amp;&amp;
		    key == Sensenotrdy &amp;&amp;
		    code == Lunnotrdy &amp;&amp; sense[13] == 0x08) {
			/* long write in progress, per mmc-6 */
			tries = 0;
			sleep(1);
		} else if(cmd[0] == Write10 || cmd[0] == Writever10)
			break;		/* don't retry worm writes */
	}

	/* drive not ready, or medium not present */
	if(cmd[0] == Readtoc &amp;&amp; key == Sensenotrdy &amp;&amp;
	    (code == Nomedium || code == Lunnotrdy)) {
		s-&gt;changetime = 0;
		qunlock(s);
		return -1;
	}
	qunlock(s);

	if(cmd[0] == Readtoc &amp;&amp; key == Sensebadreq &amp;&amp; code == Badcdb)
		return -1;			/* blank media */

	p = scsierror(code, sense[13]);

	werrstr("cmd #%.2ux: %s", cmd[0], p);

	if(scsiverbose)
		fprint(2, "scsi cmd #%.2ux: %.2ux %.2ux %.2ux: %s\n",
			cmd[0], key, code, sense[13], p);

//	if(key == Sensenone)
//		return dcount;
	return -1;
}

Scsi*
openscsi(char *dev)
{
	Scsi *s;
	int rawfd, ctlfd, l, n;
	char *name, *p, buf[512];

	l = strlen(dev)+1+3+1;
	name = malloc(l);
	if(name == nil)
		return nil;

	snprint(name, l, "%s/raw", dev);
	if((rawfd = open(name, ORDWR)) &lt; 0) {
		free(name);
		return nil;
	}

	snprint(name, l, "%s/ctl", dev);
	if((ctlfd = open(name, ORDWR)) &lt; 0) {
	Error:
		free(name);
		close(rawfd);
		return nil;
	}

	n = readn(ctlfd, buf, sizeof buf);
	close(ctlfd);
	if(n &lt;= 0) {
		if(n == 0)
			werrstr("eof on %s", name);
		goto Error;
	}

	if(strncmp(buf, "inquiry ", 8) != 0 || (p = strchr(buf, '\n')) == nil) {
		werrstr("inquiry mal-formatted in %s", name);
		goto Error;
	}
	*p = '\0';
	free(name);
	name = nil;

	if((p = strdup(buf+8)) == nil)
		goto Error;

	s = mallocz(sizeof(*s), 1);
	if(s == nil) {
	Error1:
		free(p);
		goto Error;
	}

	s-&gt;rawfd = rawfd;
	s-&gt;inquire = p;
	s-&gt;changetime = time(0);
	
	if(scsiready(s) &lt; 0)
		goto Error1;

	return s;
}

void
closescsi(Scsi *s)
{
	close(s-&gt;rawfd);
	free(s-&gt;inquire);
	free(s);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

