<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/cfs/disk.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/cfs/disk.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include "cformat.h"
#include "lru.h"
#include "bcache.h"
#include "disk.h"

int	icformat(Disk*, ulong);

/*
 *  read in the disk structures,  return -1 if the format
 *  is inconsistent.
 */
int
dinit(Disk *d, int f, int psize, char *expname)
{
	ulong	i;
	uvlong	length;
	char	buf[1024];
	Bbuf	*b;
	Dalloc	*ba;
	Dir	*dir;

	/*
	 *  get disk size
	 */
	dir = dirfstat(f);
	if(dir == nil){
		perror("dinit: stat");
		return -1;
	}
	length = dir-&gt;length;
	free(dir);

	/*
	 *  read first physical block to get logical block size, # of inodes,
	 *  and # of allocation blocks
	 */
	if(seek(f, 0, 0) &lt; 0){
		perror("dinit: seek");
		return -1;
	}
	if(read(f, buf, sizeof(buf)) != sizeof(buf)){
		perror("dinit: read");
		return -1;
	}
	ba = (Dalloc*)buf;
	if(ba-&gt;bsize &lt;= 0){
		fprint(2, "dinit: bsize 0x%lux&lt;= 0\n", ba-&gt;bsize);
		return -1;
	}
	if((ba-&gt;bsize % psize) != 0){
		fprint(2, "dinit: logical bsize (%lud) not multiple of physical (%ud)\n",
			ba-&gt;bsize, psize);
		return -1;
	}
	d-&gt;bsize = ba-&gt;bsize;
	d-&gt;nb = length/d-&gt;bsize;
	d-&gt;b2b = (d-&gt;bsize - sizeof(Dahdr))*8;
	d-&gt;nab = (d-&gt;nb+d-&gt;b2b-1)/d-&gt;b2b;
	d-&gt;p2b = d-&gt;bsize/sizeof(Dptr);
	strncpy(d-&gt;name, ba-&gt;name, sizeof d-&gt;name);

	if (expname != nil &amp;&amp; strncmp(d-&gt;name, expname, sizeof d-&gt;name) != 0) {
		/* Mismatch with recorded name; fail here to force a format */
		fprint(2, "cfs: name mismatch\n");
		return -1;
	}

	/*
	 *  check allocation blocks for consistency
	 */
	if(bcinit(d, f, d-&gt;bsize) &lt; 0){
		fprint(2, "dinit: couldn't init block cache\n");
		return -1;
	}
	for(i = 0; i &lt; d-&gt;nab; i++){
		b = bcread(d, i);
		if(b == 0){
			perror("dinit: read");
			return -1;
		}
		ba = (Dalloc*)b-&gt;data;
		if(ba-&gt;magic != Amagic){
			fprint(2, "dinit: bad magic in alloc block %uld\n", i);
			return -1;
		}
		if(d-&gt;bsize != ba-&gt;bsize){
			fprint(2, "dinit: bad bsize in alloc block %uld\n", i);
			return -1;
		}
		if(d-&gt;nab != ba-&gt;nab){
			fprint(2, "dinit: bad nab in alloc block %uld\n", i);
			return -1;
		}
		if(strncmp(d-&gt;name, ba-&gt;name, sizeof(d-&gt;name))){
			fprint(2, "dinit: bad name in alloc block %uld\n", i);
			return -1;
		}
	}
	return 0;
}

/*
 *  format the disk as a cache
 */
int
dformat(Disk *d, int f, char *name, ulong bsize, ulong psize)
{
	int	i;
	uvlong	length;
	Bbuf	*b;
	Dalloc	*ba;
	Dir	*dir;
	Dptr	dptr;

	fprint(2, "formatting disk\n");

	/*
	 *  calculate basic numbers
	 */
	dir = dirfstat(f);
	if(dir == nil)
		return -1;
	length = dir-&gt;length;
	d-&gt;bsize = bsize;
	if((d-&gt;bsize % psize) != 0){
		fprint(2, "cfs: logical bsize not multiple of physical\n");
		return -1;
	}
	d-&gt;nb = length/d-&gt;bsize;
	d-&gt;b2b = (d-&gt;bsize - sizeof(Dahdr))*8;
	d-&gt;nab = (d-&gt;nb+d-&gt;b2b-1)/d-&gt;b2b;
	d-&gt;p2b = d-&gt;bsize/sizeof(Dptr);

	/*
	 *  init allocation blocks
	 */
	if(bcinit(d, f, d-&gt;bsize) &lt; 0)
		return -1;
	for(i = 0; i &lt; d-&gt;nab; i++){
		b = bcalloc(d, i);
		if(b == 0){
			perror("cfs: bcalloc");
			return -1;
		}
		memset(b-&gt;data, 0, d-&gt;bsize);
		ba = (Dalloc*)b-&gt;data;
		ba-&gt;magic = Amagic;
		ba-&gt;bsize = d-&gt;bsize;
		ba-&gt;nab = d-&gt;nab;
		strncpy(ba-&gt;name, name, sizeof(ba-&gt;name));
		bcmark(d, b);
	}

	/*
	 *  allocate allocation blocks
	 */
	for(i = 0; i &lt; d-&gt;nab; i++)
		if(dalloc(d, &amp;dptr) == Notabno){
			fprint(2, "can't allocate allocation blocks\n");
			return -1;
		}

	return bcsync(d);
}

/*
 *  allocate a block from a bit vector page
 *
 *  a return value of Notabno means no blocks left
 */
static ulong
_balloc(Dalloc *ba, ulong max)
{
	int len;	/* number of valid words */
	ulong i;	/* bit position in long */
	ulong m;	/* 1&lt;&lt;i */
	ulong v;	/* old value of long */
	ulong *p, *e;

	/*
	 *  find a word with a 0 bit
	 */
	len = (max+BtoUL-1)/BtoUL;
	for(p = ba-&gt;bits, e = p + len; p &lt; e; p++)
		if(*p != 0xFFFFFFFF)
			break;
	if(p == e)
		return Notabno;

	/*
	 *  find the first 0 bit
	 */
	v = *p;
	for(m = 1, i = 0; i &lt; BtoUL; i++, m &lt;&lt;= 1)
		if((m|v) != v)
			break;

	/*
	 *  calculate block number
	 */
	i += (p - ba-&gt;bits)*BtoUL;
	if(i &gt;= max)
		return Notabno;

	/*
	 *  set bit to 1
	 */
	*p = v | m;
	return i;
}

/*
 *  allocate a block
 *
 *  return Notabno if none left
 */
ulong
dalloc(Disk *d, Dptr *p)
{
	ulong	bno, max, rv;
	Bbuf	*b;
	Dalloc	*ba;

	max = d-&gt;nb;
	for(bno = 0; bno &lt; d-&gt;nab; bno++){
		b = bcread(d, bno);
		ba = (Dalloc*)b-&gt;data;
		rv = _balloc(ba, max &gt; d-&gt;b2b ? d-&gt;b2b : max);
		if(rv != Notabno){
			rv = bno*d-&gt;b2b + rv;
			if(p){
				p-&gt;start = p-&gt;end = 0;
				p-&gt;bno = rv;
			}
			bcmark(d, b);
			return rv;
		}
		max -= d-&gt;b2b;
	}
	if(p)
		p-&gt;bno = Notabno;
	return Notabno;
}

/*
 *  allocate a block of pointers
 */
ulong
dpalloc(Disk *d, Dptr *p)
{
	Bbuf *b;
	Dptr *sp, *ep;

	if(dalloc(d, p) == Notabno)
		return Notabno;

	/*
	 *  allocate the page and invalidate all the
	 *  pointers
	 */
	b = bcalloc(d, p-&gt;bno);
	if(b == 0)
		return -1;
	sp = (Dptr*)b-&gt;data;
	for(ep = sp + d-&gt;p2b; sp &lt; ep; sp++){
		sp-&gt;bno = Notabno;
		sp-&gt;start = sp-&gt;end = 0;
	}
	p-&gt;bno |= Indbno;
	p-&gt;start = 0;
	p-&gt;end = d-&gt;bsize;

	/*
	 *  mark the page as dirty
	 */
	bcmark(d, b);
	return 0;
}

/*
 *  free a block
 */
int
_bfree(Disk *d, ulong i)
{
	ulong bno, m;
	ulong *p;
	Bbuf *b;
	Dalloc *ba;

	/*
	 *  get correct allocation block
	 */
	bno = i/d-&gt;b2b;
	if(bno &gt;= d-&gt;nab)
		return -1;
	b = bcread(d, bno);
	if(b == 0)
		return -1;
	ba = (Dalloc*)b-&gt;data;

	/*
	 *  change bit
	 */
	i -= bno*d-&gt;b2b;
	p = ba-&gt;bits + (i/BtoUL);
	m = 1&lt;&lt;(i%BtoUL);
	*p &amp;= ~m;
	bcmark(d, b);

	return 0;
}

/*
 *  free a block (or blocks)
 */
int
dfree(Disk *d, Dptr *dp)
{
	ulong bno;
	Dptr *sp, *ep;
	Bbuf *b;

	bno = dp-&gt;bno;
	dp-&gt;bno = Notabno;

	/*
	 *  nothing to free
	 */
	if(bno == Notabno)
		return 0;

	/*
	 *  direct pointer
	 */
	if((bno &amp; Indbno) == 0)
		return _bfree(d, bno);

	/*
	 *  first indirect page
	 */
	bno &amp;= ~Indbno;
	_bfree(d, bno);

	/*
	 *  then all the pages it points to
	 *
	 *	DANGER: this algorithm may fail if there are more
	 *		allocation blocks than block buffers
	 */
	b = bcread(d, bno);
	if(b == 0)
		return -1;
	sp = (Dptr*)b-&gt;data;
	for(ep = sp + d-&gt;p2b; sp &lt; ep; sp++)
		if(dfree(d, sp) &lt; 0)
			return -1;
	return 0;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

