<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/cfs/inode.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/cfs/inode.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include "cformat.h"
#include "lru.h"
#include "bcache.h"
#include "disk.h"
#include "inode.h"
#include "stats.h"

/*
 *  read the inode blocks and make sure they
 *  haven't been trashed.
 *
 *  make the in-core table of qid to inode mappings.
 *	N.B. this is just an array. we need a linear search to find
 *	     a particular inode. this could be done faster.
 *
 *  nab is the first inode block.
 */
int
iinit(Icache *ic, int f, int psize, char* name)
{
	Ibuf *b;
	Imap *m;
	ulong ino;
	Bbuf *bb;
	Dinode *bi;

	/*
	 *  get basic sizes and allocation info from disk
	 */
	if(dinit(ic, f, psize, name) &lt; 0)
		return -1;

	/*
	 *  read first inode block to get number of inodes
	 */
	bb = bcread(ic, ic-&gt;nab);
	if(bb == 0){
		fprint(2, "iinit: can't read disk\n");
		return -1;
	}
	bi = (Dinode*)bb-&gt;data;
	if(bi-&gt;nino==0 || bi-&gt;nino&gt;2048){
		fprint(2, "iinit: bad nino\n");
		return -1;
	}
	ic-&gt;nino = bi-&gt;nino;

	/*
	 *  set up sizing constants
	 */
	ic-&gt;i2b = (ic-&gt;bsize - sizeof(Dihdr))/sizeof(Inode);
	ic-&gt;nib = (ic-&gt;nino + ic-&gt;i2b - 1)/ic-&gt;i2b;

	/*
	 *  allocate the in-core qid/inode map, build it's lru
	 */
	if(ic-&gt;map)
		free(ic-&gt;map);
	ic-&gt;map = malloc(sizeof(Imap)*ic-&gt;nino);
	if(ic-&gt;map == 0){
		fprint(2, "iinit: can't alloc map\n");
		return -1;
	}
	lruinit(&amp;ic-&gt;mlru);
	for(m = ic-&gt;map; m &lt; &amp;ic-&gt;map[ic-&gt;nino]; m++){
		m-&gt;inuse = 0;
		m-&gt;b = 0;
		lruadd(&amp;ic-&gt;mlru, m);
	}

	/*
	 *  mark all cache buffers as empty, put them on the lru list
	 */
	lruinit(&amp;ic-&gt;blru);
	for(b = ic-&gt;ib; b &lt; &amp;ic-&gt;ib[Nicache]; b++){
		b-&gt;inuse = 0;
		lruadd(&amp;ic-&gt;blru, b);
	}

	/*
	 *  Read all inodes and
	 *  build the in-core qid/inode map
	 */
	for(ino = 0; ino &lt; ic-&gt;nino; ino++){
		b = iread(ic, ino);
		if(b == 0){
			fprint(2, "iinit: can't read inode %ld\n", ino);
			return -1;
		}
		if(b-&gt;inode.inuse){
			m = &amp;ic-&gt;map[ino];
			m-&gt;inuse = 1;
			m-&gt;qid = b-&gt;inode.qid;
			lruref(&amp;ic-&gt;mlru, m);
		}
	}
	return 0;
}

/*
 *  format the inode blocks
 */
int
iformat(Icache *ic, int f, ulong nino, char *name, int bsize, int psize)
{
	int nib;
	ulong bno, i2b, i;
	Bbuf *bb;
	Dinode *bi;

	/*
	 *  first format disk allocation
	 */
	if(dformat(ic, f, name, bsize, psize) &lt; 0)
		return -1;

	fprint(2, "formatting inodes\n");

	i2b = (bsize - sizeof(Dihdr))/sizeof(Inode);
	nib = (nino + i2b - 1)/i2b;

	for(bno = ic-&gt;nab; bno &lt; ic-&gt;nab + nib; bno++){
		if(dalloc(ic, 0) == Notabno){
			fprint(2, "iformat: balloc failed\n");
			return -1;
		}
		bb = bcalloc(ic, bno);
		if(bb == 0){
			fprint(2, "iformat: bcalloc failed\n");
			return -1;
		}
		bi = (Dinode*)bb-&gt;data;
		bi-&gt;magic = Imagic;
		bi-&gt;nino = nino;
		for(i = 0; i &lt; i2b; i++)
			bi-&gt;inode[i].inuse = 0;
		bcmark(ic, bb);
	}

	bcsync(ic);

	return iinit(ic, f, psize, name);
}

/*
 *  allocate a cache buffer, use least recently used
 */
Ibuf*
ialloc(Icache *ic, ulong ino)
{
	Imap *m;
	Ibuf *b;

	b = (Ibuf*)ic-&gt;blru.lnext;
	if(b-&gt;inuse)
		ic-&gt;map[b-&gt;ino].b = 0;
	b-&gt;ino = ino;
	b-&gt;inuse = 1;
	m = &amp;ic-&gt;map[ino];
	m-&gt;b = b;
	return b;
}

/*
 *  free a cache buffer
 */
void
ifree(Icache *ic, Ibuf *b)
{
	b-&gt;inuse = 0;
	if(b-&gt;inuse)
		ic-&gt;map[b-&gt;ino].b = 0;
	lruderef(&amp;ic-&gt;blru, b);
}

/*
 *  get an inode into the cache.  if no inode exists for this qid, create one
 *  from an unused qid/inode map.
 */
Ibuf *
iget(Icache *ic, Qid qid)
{
	Imap *m, *me;
	Ibuf *b;

	/*
	 *  find map entry with same qid.path
	 */
	for(m = ic-&gt;map, me = &amp;ic-&gt;map[ic-&gt;nino]; m &lt; me; m++)
		if(m-&gt;inuse &amp;&amp; m-&gt;qid.path==qid.path){
			if(m-&gt;qid.vers != qid.vers){
				/*
				 *  our info is old, forget it
				 */
				DPRINT(2, "updating old file %llud.%lud\n",
					qid.path, qid.vers);
				m-&gt;qid = qid;
				iupdate(ic, m - ic-&gt;map, qid);
			}
			break;
		}

	/*
 	 *  if an already existing inode, just get it
	 */
	if(m != me)
		return iread(ic, m - ic-&gt;map);

	/*
	 *  create a new inode, throw out the least recently used inode
	 *  if necessary
	 */
	m = (Imap*)ic-&gt;mlru.lnext;
	if(m-&gt;inuse){
		DPRINT(2, "superceding file %llud.%ld by %llud.%ld\n",
			m-&gt;qid.path, m-&gt;qid.vers, qid.path, qid.vers);
		if(iremove(ic, m - ic-&gt;map) &lt; 0)
			return 0;
	}

	if(statson)
		cfsstat.ninsert++;
	/*
	 *  init inode and write to disk
	 */
	DPRINT(2, "new file %llud.%ld ino %ld\n",
		qid.path, qid.vers, m - ic-&gt;map);
	b = ialloc(ic, m - ic-&gt;map);
	b-&gt;inode.inuse = m-&gt;inuse = 1;
	b-&gt;inode.qid = qid;
	b-&gt;inode.length = 0x7fffffffffffffffLL;
	m-&gt;qid = qid;
	b-&gt;inode.ptr.bno = Notabno;
	iwrite(ic, b);
	return b;
}

/*
 *  read an inode into the cache
 *
 *  ASSUMPTION: the inode is valid
 */
Ibuf*
iread(Icache *ic, ulong ino)
{
	Ibuf *b;
	Imap *m;
	ulong bno;
	Bbuf *bb;
	Dinode *bi;

	/*
	 *  first see if we already have it in a cache entry
	 */
	m = &amp;ic-&gt;map[ino];
	if(m-&gt;inuse &amp;&amp; m-&gt;b){
		b = m-&gt;b;
		goto out;
	}

	/*
	 *  read it
	 */
	b = ialloc(ic, ino);
	bno = ic-&gt;nab + ino/ic-&gt;i2b;
	bb = bcread(ic, bno);
	if(bb == 0){
		ifree(ic, b);
		return 0;
	}
	bi = (Dinode*)bb-&gt;data;
	b-&gt;inode = bi-&gt;inode[ino % ic-&gt;i2b];

	/*
	 *  consistency check
	 */
	if(bi-&gt;nino!=ic-&gt;nino || bi-&gt;magic!=Imagic){
		fprint(2, "iread: inconsistent inode block\n");
		ifree(ic, b);
		return 0;
	}
out:
	b-&gt;inuse = 1;
	m-&gt;b = b;
	if(b-&gt;inode.inuse)
		lruref(&amp;ic-&gt;mlru, m);
	lruref(&amp;ic-&gt;blru, b);
	return b;
}

/*
 *  write an inode back to disk
 */
int
iwrite(Icache *ic, Ibuf *b)
{
	ulong bno;
	Bbuf *bb;
	Dinode *bi;

	bno = ic-&gt;nab + b-&gt;ino/ic-&gt;i2b;
	bb = bcread(ic, bno);
	if(bb == 0)
		return 0;
	bi = (Dinode*)bb-&gt;data;
	bi-&gt;inode[b-&gt;ino % ic-&gt;i2b] = b-&gt;inode;
	bcmark(ic, bb);
	lruref(&amp;ic-&gt;mlru, &amp;ic-&gt;map[b-&gt;ino]);
	lruref(&amp;ic-&gt;blru, b);
	return 0;
}

/*
 *  Forget what we know about an inode without removing it
 *
 *	N.B: ordering of iwrite and dfree is important
 */
int
iupdate(Icache *ic, ulong ino, Qid qid)
{
	Ibuf *b;
	Imap *m;
	Dptr d;

	if(statson)
		cfsstat.nupdate++;
	b = iread(ic, ino);
	if(b == 0)
		return -1;

	/*
	 *  update inode and map
	 */
	b-&gt;inode.qid = qid;
	b-&gt;inode.length = 0x7fffffffffffffffLL;	/* Set to maximum */
	m = &amp;ic-&gt;map[ino];
	m-&gt;qid = qid;

	/*
	 *  the free is not done if the write fails!
	 *  this is important
	 */
	d = b-&gt;inode.ptr;
	b-&gt;inode.ptr.bno = Notabno;
	if(iwrite(ic, b) &lt; 0)
		return -1;
	dfree(ic, &amp;d);
	return 0;
}

/*
 *  remove an inode
 *
 *	N.B: ordering of iwrite and dfree is important
 */
int
iremove(Icache *ic, ulong ino)
{
	Ibuf *b;
	Imap *m;

	if(statson)
		cfsstat.ndelete++;
	m = &amp;ic-&gt;map[ino];

	/*
	 *  read in inode
	 */
	b = iread(ic, ino);
	if(b == 0)
		return -1;

	/*
	 *  mark it unused on disk
	 */
	b-&gt;inode.inuse = 0;
	if(iwrite(ic, b) &lt; 0)
		return -1;

	/*
	 *  throw out it's data pages
	 */
	dfree(ic, &amp;b-&gt;inode.ptr);

	/*
	 *  free the inode buffer
	 */
	ifree(ic, b);

	/*
	 *  make map entry least recently used
	 */
	lruderef(&amp;ic-&gt;mlru, m);
	return 0;
}

/*
 *  increment our version number
 */
void
iinc(Icache *ic, Ibuf *b)
{
	b-&gt;inode.qid.vers++;
	ic-&gt;map[b-&gt;ino].qid = b-&gt;inode.qid;
	iwrite(ic, b);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

