<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/fossil/archive.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/fossil/archive.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * Archiver.  In charge of sending blocks to Venti.
 */

#include "stdinc.h"
#include "dat.h"
#include "fns.h"
#include "error.h"

#include "9.h"	/* for consPrint */

#define DEBUG 0

static void archThread(void*);

struct Arch
{
	int ref;
	uint blockSize;
	uint diskSize;
	Cache *c;
	Fs *fs;
	VtSession *z;

	VtLock *lk;
	VtRendez *starve;
	VtRendez *die;
};

Arch *
archInit(Cache *c, Disk *disk, Fs *fs, VtSession *z)
{
	Arch *a;

	a = vtMemAllocZ(sizeof(Arch));

	a-&gt;c = c;
	a-&gt;z = z;
	a-&gt;fs = fs;
	a-&gt;blockSize = diskBlockSize(disk);
	a-&gt;lk = vtLockAlloc();
	a-&gt;starve = vtRendezAlloc(a-&gt;lk);

	a-&gt;ref = 2;
	vtThread(archThread, a);

	return a;
}

void
archFree(Arch *a)
{
	/* kill slave */
	vtLock(a-&gt;lk);
	a-&gt;die = vtRendezAlloc(a-&gt;lk);
	vtWakeup(a-&gt;starve);
	while(a-&gt;ref &gt; 1)
		vtSleep(a-&gt;die);
	vtUnlock(a-&gt;lk);
	vtRendezFree(a-&gt;starve);
	vtRendezFree(a-&gt;die);
	vtLockFree(a-&gt;lk);
	vtMemFree(a);
}

static int
ventiSend(Arch *a, Block *b, uchar *data)
{
	uint n;
	uchar score[VtScoreSize];

	if(DEBUG &gt; 1)
		fprint(2, "ventiSend: sending %#ux %L to venti\n", b-&gt;addr, &amp;b-&gt;l);
	n = vtZeroTruncate(vtType[b-&gt;l.type], data, a-&gt;blockSize);
	if(DEBUG &gt; 1)
		fprint(2, "ventiSend: truncate %d to %d\n", a-&gt;blockSize, n);
	if(!vtWrite(a-&gt;z, score, vtType[b-&gt;l.type], data, n)){
		fprint(2, "ventiSend: vtWrite block %#ux failed: %R\n", b-&gt;addr);
		return 0;
	}
	if(!vtSha1Check(score, data, n)){
		uchar score2[VtScoreSize];
		vtSha1(score2, data, n);
		fprint(2, "ventiSend: vtWrite block %#ux failed vtSha1Check %V %V\n",
			b-&gt;addr, score, score2);
		return 0;
	}
	if(!vtSync(a-&gt;z))
		return 0;
	return 1;
}

/*
 * parameters for recursion; there are so many,
 * and some only change occasionally.  this is
 * easier than spelling things out at each call.
 */
typedef struct Param Param;
struct Param
{
	/* these never change */
	uint snapEpoch;	/* epoch for snapshot being archived */
	uint blockSize;
	Cache *c;
	Arch *a;

	/* changes on every call */
	uint depth;

	/* statistics */
	uint nfixed;
	uint nsend;
	uint nvisit;
	uint nfailsend;
	uint maxdepth;
	uint nreclaim;
	uint nfake;
	uint nreal;

	/* these occasionally change (must save old values and put back) */
	uint dsize;
	uint psize;

	/* return value; avoids using stack space */
	Label l;
	uchar score[VtScoreSize];
};

static void
shaBlock(uchar score[VtScoreSize], Block *b, uchar *data, uint bsize)
{
	vtSha1(score, data, vtZeroTruncate(vtType[b-&gt;l.type], data, bsize));
}

static uint
etype(Entry *e)
{
	uint t;

	if(e-&gt;flags&amp;VtEntryDir)
		t = BtDir;
	else
		t = BtData;
	return t+e-&gt;depth;
}

static uchar*
copyBlock(Block *b, u32int blockSize)
{
	uchar *data;

	data = vtMemAlloc(blockSize);
	if(data == nil)
		return nil;
	memmove(data, b-&gt;data, blockSize);
	return data;
}

/*
 * Walk over the block tree, archiving it to Venti.
 *
 * We don't archive the snapshots. Instead we zero the
 * entries in a temporary copy of the block and archive that.
 *
 * Return value is:
 *
 *	ArchFailure	some error occurred
 *	ArchSuccess	block and all children archived
 * 	ArchFaked	success, but block or children got copied
 */
enum
{
	ArchFailure,
	ArchSuccess,
	ArchFaked,
};
static int
archWalk(Param *p, u32int addr, uchar type, u32int tag)
{
	int ret, i, x, psize, dsize;
	uchar *data, score[VtScoreSize];
	Block *b;
	Label l;
	Entry *e;
	WalkPtr w;

	p-&gt;nvisit++;

	b = cacheLocalData(p-&gt;c, addr, type, tag, OReadWrite,0);
	if(b == nil){
		fprint(2, "archive(%ud, %#ux): cannot find block: %R\n", p-&gt;snapEpoch, addr);
		if(strcmp(vtGetError(), ELabelMismatch) == 0){
			/* might as well plod on so we write _something_ to Venti */
			memmove(p-&gt;score, vtZeroScore, VtScoreSize);
			return ArchFaked;
		}
		return ArchFailure;
	}

	if(DEBUG) fprint(2, "%*sarchive(%ud, %#ux): block label %L\n",
		p-&gt;depth*2, "",  p-&gt;snapEpoch, b-&gt;addr, &amp;b-&gt;l);
	p-&gt;depth++;
	if(p-&gt;depth &gt; p-&gt;maxdepth)
		p-&gt;maxdepth = p-&gt;depth;

	data = b-&gt;data;
	if((b-&gt;l.state&amp;BsVenti) == 0){
		initWalk(&amp;w, b, b-&gt;l.type==BtDir ? p-&gt;dsize : p-&gt;psize);
		for(i=0; nextWalk(&amp;w, score, &amp;type, &amp;tag, &amp;e); i++){
			if(e){
				if(!(e-&gt;flags&amp;VtEntryActive))
					continue;
				if((e-&gt;snap &amp;&amp; !e-&gt;archive)
				|| (e-&gt;flags&amp;VtEntryNoArchive)){
					if(0) fprint(2, "snap; faking %#ux\n", b-&gt;addr);
					if(data == b-&gt;data){
						data = copyBlock(b, p-&gt;blockSize);
						if(data == nil){
							ret = ArchFailure;
							goto Out;
						}
						w.data = data;
					}
					memmove(e-&gt;score, vtZeroScore, VtScoreSize);
					e-&gt;depth = 0;
					e-&gt;size = 0;
					e-&gt;tag = 0;
					e-&gt;flags &amp;= ~VtEntryLocal;
					entryPack(e, data, w.n-1);
					continue;
				}
			}
			addr = globalToLocal(score);
			if(addr == NilBlock)
				continue;
			dsize = p-&gt;dsize;
			psize = p-&gt;psize;
			if(e){
				p-&gt;dsize= e-&gt;dsize;
				p-&gt;psize = e-&gt;psize;
			}
			vtUnlock(b-&gt;lk);
			x = archWalk(p, addr, type, tag);
			vtLock(b-&gt;lk);
			if(e){
				p-&gt;dsize = dsize;
				p-&gt;psize = psize;
			}
			while(b-&gt;iostate != BioClean &amp;&amp; b-&gt;iostate != BioDirty)
				vtSleep(b-&gt;ioready);
			switch(x){
			case ArchFailure:
				fprint(2, "archWalk %#ux failed; ptr is in %#ux offset %d\n",
					addr, b-&gt;addr, i);
				ret = ArchFailure;
				goto Out;
			case ArchFaked:
				/*
				 * When we're writing the entry for an archive directory
				 * (like /archive/2003/1215) then even if we've faked
				 * any data, record the score unconditionally.
				 * This way, we will always record the Venti score here.
				 * Otherwise, temporary data or corrupted file system
				 * would cause us to keep holding onto the on-disk
				 * copy of the archive.
				 */
				if(e==nil || !e-&gt;archive)
				if(data == b-&gt;data){
if(0) fprint(2, "faked %#ux, faking %#ux (%V)\n", addr, b-&gt;addr, p-&gt;score);
					data = copyBlock(b, p-&gt;blockSize);
					if(data == nil){
						ret = ArchFailure;
						goto Out;
					}
					w.data = data;
				}
				/* fall through */
if(0) fprint(2, "falling\n");
			case ArchSuccess:
				if(e){
					memmove(e-&gt;score, p-&gt;score, VtScoreSize);
					e-&gt;flags &amp;= ~VtEntryLocal;
					entryPack(e, data, w.n-1);
				}else
					memmove(data+(w.n-1)*VtScoreSize, p-&gt;score, VtScoreSize);
				if(data == b-&gt;data){
					blockDirty(b);
					/*
					 * If b is in the active tree, then we need to note that we've
					 * just removed addr from the active tree (replacing it with the 
					 * copy we just stored to Venti).  If addr is in other snapshots,
					 * this will close addr but not free it, since it has a non-empty
					 * epoch range.
					 *
					 * If b is in the active tree but has been copied (this can happen
					 * if we get killed at just the right moment), then we will
					 * mistakenly leak its kids.  
					 *
					 * The children of an archive directory (e.g., /archive/2004/0604)
					 * are not treated as in the active tree.
					 */
					if((b-&gt;l.state&amp;BsCopied)==0 &amp;&amp; (e==nil || e-&gt;snap==0))
						blockRemoveLink(b, addr, p-&gt;l.type, p-&gt;l.tag, 0);
				}
				break;
			}
		}

		if(!ventiSend(p-&gt;a, b, data)){
			p-&gt;nfailsend++;
			ret = ArchFailure;
			goto Out;
		}
		p-&gt;nsend++;
		if(data != b-&gt;data)
			p-&gt;nfake++;
		if(data == b-&gt;data){	/* not faking it, so update state */
			p-&gt;nreal++;
			l = b-&gt;l;
			l.state |= BsVenti;
			if(!blockSetLabel(b, &amp;l, 0)){
				ret = ArchFailure;
				goto Out;
			}
		}
	}

	shaBlock(p-&gt;score, b, data, p-&gt;blockSize);
if(0) fprint(2, "ventisend %V %p %p %p\n", p-&gt;score, data, b-&gt;data, w.data);
	ret = data!=b-&gt;data ? ArchFaked : ArchSuccess;
	p-&gt;l = b-&gt;l;
Out:
	if(data != b-&gt;data)
		vtMemFree(data);
	p-&gt;depth--;
	blockPut(b);
	return ret;
}

static void
archThread(void *v)
{
	Arch *a = v;
	Block *b;
	Param p;
	Super super;
	int ret;
	u32int addr;
	uchar rbuf[VtRootSize];
	VtRoot root;

	vtThreadSetName("arch");

	for(;;){
		/* look for work */
		vtLock(a-&gt;fs-&gt;elk);
		b = superGet(a-&gt;c, &amp;super);
		if(b == nil){
			vtUnlock(a-&gt;fs-&gt;elk);
			fprint(2, "archThread: superGet: %R\n");
			sleep(60*1000);
			continue;
		}
		addr = super.next;
		if(addr != NilBlock &amp;&amp; super.current == NilBlock){
			super.current = addr;
			super.next = NilBlock;
			superPack(&amp;super, b-&gt;data);
			blockDirty(b);
		}else
			addr = super.current;
		blockPut(b);
		vtUnlock(a-&gt;fs-&gt;elk);

		if(addr == NilBlock){
			/* wait for work */
			vtLock(a-&gt;lk);
			vtSleep(a-&gt;starve);
			if(a-&gt;die != nil)
				goto Done;
			vtUnlock(a-&gt;lk);
			continue;
		}

sleep(10*1000);	/* window of opportunity to provoke races */

		/* do work */
		memset(&amp;p, 0, sizeof p);
		p.blockSize = a-&gt;blockSize;
		p.dsize = 3*VtEntrySize;	/* root has three Entries */
		p.c = a-&gt;c;
		p.a = a;

		ret = archWalk(&amp;p, addr, BtDir, RootTag);
		switch(ret){
		default:
			abort();
		case ArchFailure:
			fprint(2, "archiveBlock %#ux: %R\n", addr);
			sleep(60*1000);
			continue;
		case ArchSuccess:
		case ArchFaked:
			break;
		}

		if(0) fprint(2, "archiveSnapshot 0x%#ux: maxdepth %ud nfixed %ud"
			" send %ud nfailsend %ud nvisit %ud"
			" nreclaim %ud nfake %ud nreal %ud\n",
			addr, p.maxdepth, p.nfixed,
			p.nsend, p.nfailsend, p.nvisit,
			p.nreclaim, p.nfake, p.nreal);
		if(0) fprint(2, "archiveBlock %V (%ud)\n", p.score, p.blockSize);

		/* tie up vac root */
		memset(&amp;root, 0, sizeof root);
		root.version = VtRootVersion;
		strecpy(root.type, root.type+sizeof root.type, "vac");
		strecpy(root.name, root.name+sizeof root.name, "fossil");
		memmove(root.score, p.score, VtScoreSize);
		memmove(root.prev, super.last, VtScoreSize);
		root.blockSize = a-&gt;blockSize;
		vtRootPack(&amp;root, rbuf);
		if(!vtWrite(a-&gt;z, p.score, VtRootType, rbuf, VtRootSize)
		|| !vtSha1Check(p.score, rbuf, VtRootSize)){
			fprint(2, "vtWriteBlock %#ux: %R\n", addr);
			sleep(60*1000);
			continue;
		}

		/* record success */
		vtLock(a-&gt;fs-&gt;elk);
		b = superGet(a-&gt;c, &amp;super);
		if(b == nil){
			vtUnlock(a-&gt;fs-&gt;elk);
			fprint(2, "archThread: superGet: %R\n");
			sleep(60*1000);
			continue;
		}
		super.current = NilBlock;
		memmove(super.last, p.score, VtScoreSize);
		superPack(&amp;super, b-&gt;data);
		blockDirty(b);
		blockPut(b);
		vtUnlock(a-&gt;fs-&gt;elk);

		consPrint("archive vac:%V\n", p.score);
	}

Done:
	a-&gt;ref--;
	vtWakeup(a-&gt;die);
	vtUnlock(a-&gt;lk);
}

void
archKick(Arch *a)
{
	if(a == nil){
		fprint(2, "warning: archKick nil\n");
		return;
	}
	vtLock(a-&gt;lk);
	vtWakeup(a-&gt;starve);
	vtUnlock(a-&gt;lk);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

