<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/fossil/disk.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/fossil/disk.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include "stdinc.h"
#include "dat.h"
#include "fns.h"
#include "error.h"

static void diskThread(void *a);

enum {
	/*
	 * disable measurement since it gets alignment faults on BG
	 * and the guts used to be commented out.
	 */
	Timing	= 0,			/* flag */
	QueueSize = 100,		/* maximum block to queue */
};

struct Disk {
	VtLock *lk;
	int ref;

	int fd;
	Header h;

	VtRendez *flow;
	VtRendez *starve;
	VtRendez *flush;
	VtRendez *die;

	int nqueue;

	Block *cur;		/* block to do on current scan */
	Block *next;		/* blocks to do next scan */
};

/* keep in sync with Part* enum in dat.h */
static char *partname[] = {
	[PartError]	"error",
	[PartSuper]	"super",
	[PartLabel]	"label",
	[PartData]	"data",
	[PartVenti]	"venti",
};

Disk *
diskAlloc(int fd)
{
	u8int buf[HeaderSize];
	Header h;
	Disk *disk;

	if(pread(fd, buf, HeaderSize, HeaderOffset) &lt; HeaderSize){
		vtSetError("short read: %r");
		vtOSError();
		return nil;
	}

	if(!headerUnpack(&amp;h, buf)){
		vtSetError("bad disk header");
		return nil;
	}
	disk = vtMemAllocZ(sizeof(Disk));
	disk-&gt;lk = vtLockAlloc();
	disk-&gt;starve = vtRendezAlloc(disk-&gt;lk);
	disk-&gt;flow = vtRendezAlloc(disk-&gt;lk);
	disk-&gt;flush = vtRendezAlloc(disk-&gt;lk);
	disk-&gt;fd = fd;
	disk-&gt;h = h;

	disk-&gt;ref = 2;
	vtThread(diskThread, disk);

	return disk;
}

void
diskFree(Disk *disk)
{
	diskFlush(disk);

	/* kill slave */
	vtLock(disk-&gt;lk);
	disk-&gt;die = vtRendezAlloc(disk-&gt;lk);
	vtWakeup(disk-&gt;starve);
	while(disk-&gt;ref &gt; 1)
		vtSleep(disk-&gt;die);
	vtUnlock(disk-&gt;lk);
	vtRendezFree(disk-&gt;flow);
	vtRendezFree(disk-&gt;starve);
	vtRendezFree(disk-&gt;die);
	vtLockFree(disk-&gt;lk);
	close(disk-&gt;fd);
	vtMemFree(disk);
}

static u32int
partStart(Disk *disk, int part)
{
	switch(part){
	default:
		assert(0);
	case PartSuper:
		return disk-&gt;h.super;
	case PartLabel:
		return disk-&gt;h.label;
	case PartData:
		return disk-&gt;h.data;
	}
}


static u32int
partEnd(Disk *disk, int part)
{
	switch(part){
	default:
		assert(0);
	case PartSuper:
		return disk-&gt;h.super+1;
	case PartLabel:
		return disk-&gt;h.data;
	case PartData:
		return disk-&gt;h.end;
	}
}

int
diskReadRaw(Disk *disk, int part, u32int addr, uchar *buf)
{
	ulong start, end;
	u64int offset;
	int n, nn;

	start = partStart(disk, part);
	end = partEnd(disk, part);

	if(addr &gt;= end-start){
		vtSetError(EBadAddr);
		return 0;
	}

	offset = ((u64int)(addr + start))*disk-&gt;h.blockSize;
	n = disk-&gt;h.blockSize;
	while(n &gt; 0){
		nn = pread(disk-&gt;fd, buf, n, offset);
		if(nn &lt; 0){
			vtOSError();
			return 0;
		}
		if(nn == 0){
			vtSetError("eof reading disk");
			return 0;
		}
		n -= nn;
		offset += nn;
		buf += nn;
	}
	return 1;
}

int
diskWriteRaw(Disk *disk, int part, u32int addr, uchar *buf)
{
	ulong start, end;
	u64int offset;
	int n;

	start = partStart(disk, part);
	end = partEnd(disk, part);

	if(addr &gt;= end - start){
		vtSetError(EBadAddr);
		return 0;
	}

	offset = ((u64int)(addr + start))*disk-&gt;h.blockSize;
	n = pwrite(disk-&gt;fd, buf, disk-&gt;h.blockSize, offset);
	if(n &lt; 0){
		vtOSError();
		return 0;
	}
	if(n &lt; disk-&gt;h.blockSize) {
		vtSetError("short write");
		return 0;
	}

	return 1;
}

static void
diskQueue(Disk *disk, Block *b)
{
	Block **bp, *bb;

	vtLock(disk-&gt;lk);
	while(disk-&gt;nqueue &gt;= QueueSize)
		vtSleep(disk-&gt;flow);
	if(disk-&gt;cur == nil || b-&gt;addr &gt; disk-&gt;cur-&gt;addr)
		bp = &amp;disk-&gt;cur;
	else
		bp = &amp;disk-&gt;next;

	for(bb=*bp; bb; bb=*bp){
		if(b-&gt;addr &lt; bb-&gt;addr)
			break;
		bp = &amp;bb-&gt;ionext;
	}
	b-&gt;ionext = bb;
	*bp = b;
	if(disk-&gt;nqueue == 0)
		vtWakeup(disk-&gt;starve);
	disk-&gt;nqueue++;
	vtUnlock(disk-&gt;lk);
}


void
diskRead(Disk *disk, Block *b)
{
	assert(b-&gt;iostate == BioEmpty || b-&gt;iostate == BioLabel);
	blockSetIOState(b, BioReading);
	diskQueue(disk, b);
}

void
diskWrite(Disk *disk, Block *b)
{
	assert(b-&gt;nlock == 1);
	assert(b-&gt;iostate == BioDirty);
	blockSetIOState(b, BioWriting);
	diskQueue(disk, b);
}

void
diskWriteAndWait(Disk *disk, Block *b)
{
	int nlock;

	/*
	 * If b-&gt;nlock &gt; 1, the block is aliased within
	 * a single thread.  That thread is us.
	 * DiskWrite does some funny stuff with VtLock
	 * and blockPut that basically assumes b-&gt;nlock==1.
	 * We humor diskWrite by temporarily setting
	 * nlock to 1.  This needs to be revisited.
	 */
	nlock = b-&gt;nlock;
	if(nlock &gt; 1)
		b-&gt;nlock = 1;
	diskWrite(disk, b);
	while(b-&gt;iostate != BioClean)
		vtSleep(b-&gt;ioready);
	b-&gt;nlock = nlock;
}

int
diskBlockSize(Disk *disk)
{
	return disk-&gt;h.blockSize;	/* immuttable */
}

int
diskFlush(Disk *disk)
{
	Dir dir;

	vtLock(disk-&gt;lk);
	while(disk-&gt;nqueue &gt; 0)
		vtSleep(disk-&gt;flush);
	vtUnlock(disk-&gt;lk);

	/* there really should be a cleaner interface to flush an fd */
	nulldir(&amp;dir);
	if(dirfwstat(disk-&gt;fd, &amp;dir) &lt; 0){
		vtOSError();
		return 0;
	}
	return 1;
}

u32int
diskSize(Disk *disk, int part)
{
	return partEnd(disk, part) - partStart(disk, part);
}

static uintptr
mypc(int x)
{
	return getcallerpc(&amp;x);
}

static char *
disk2file(Disk *disk)
{
	static char buf[256];

	if (fd2path(disk-&gt;fd, buf, sizeof buf) &lt; 0)
		strncpy(buf, "GOK", sizeof buf);
	return buf;
}

static void
diskThread(void *a)
{
	Disk *disk = a;
	Block *b;
	uchar *buf, *p;
	double t;
	int nio;

	vtThreadSetName("disk");

//fprint(2, "diskThread %d\n", getpid());

	buf = vtMemAlloc(disk-&gt;h.blockSize);

	vtLock(disk-&gt;lk);
	if (Timing) {
		nio = 0;
		t = -nsec();
	}
	for(;;){
		while(disk-&gt;nqueue == 0){
			if (Timing) {
				t += nsec();
				if(nio &gt;= 10000){
					fprint(2, "disk: io=%d at %.3fms\n",
						nio, t*1e-6/nio);
					nio = 0;
					t = 0;
				}
			}
			if(disk-&gt;die != nil)
				goto Done;
			vtSleep(disk-&gt;starve);
			if (Timing)
				t -= nsec();
		}
		assert(disk-&gt;cur != nil || disk-&gt;next != nil);

		if(disk-&gt;cur == nil){
			disk-&gt;cur = disk-&gt;next;
			disk-&gt;next = nil;
		}
		b = disk-&gt;cur;
		disk-&gt;cur = b-&gt;ionext;
		vtUnlock(disk-&gt;lk);

		/*
		 * no one should hold onto blocking in the
		 * reading or writing state, so this lock should
		 * not cause deadlock.
		 */
if(0)fprint(2, "fossil: diskThread: %d:%d %x\n", getpid(), b-&gt;part, b-&gt;addr);
		bwatchLock(b);
		vtLock(b-&gt;lk);
		b-&gt;pc = mypc(0);
		assert(b-&gt;nlock == 1);
		switch(b-&gt;iostate){
		default:
			abort();
		case BioReading:
			if(!diskReadRaw(disk, b-&gt;part, b-&gt;addr, b-&gt;data)){
				fprint(2, "fossil: diskReadRaw failed: %s: "
					"score %V: part=%s block %ud: %r\n",
					disk2file(disk), b-&gt;score,
					partname[b-&gt;part], b-&gt;addr);
				blockSetIOState(b, BioReadError);
			}else
				blockSetIOState(b, BioClean);
			break;
		case BioWriting:
			p = blockRollback(b, buf);
			/* NB: ctime result ends with a newline */
			if(!diskWriteRaw(disk, b-&gt;part, b-&gt;addr, p)){
				fprint(2, "fossil: diskWriteRaw failed: %s: "
				    "score %V: date %s part=%s block %ud: %r\n",
					disk2file(disk), b-&gt;score,
					ctime(time(0)),
					partname[b-&gt;part], b-&gt;addr);
				break;
			}
			if(p != buf)
				blockSetIOState(b, BioClean);
			else
				blockSetIOState(b, BioDirty);
			break;
		}

		blockPut(b);		/* remove extra reference, unlock */
		vtLock(disk-&gt;lk);
		disk-&gt;nqueue--;
		if(disk-&gt;nqueue == QueueSize-1)
			vtWakeup(disk-&gt;flow);
		if(disk-&gt;nqueue == 0)
			vtWakeup(disk-&gt;flush);
		if(Timing)
			nio++;
	}
Done:
//fprint(2, "diskThread done\n");
	disk-&gt;ref--;
	vtWakeup(disk-&gt;die);
	vtUnlock(disk-&gt;lk);
	vtMemFree(buf);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

