<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/gs/jpeg/rdrle.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/gs/jpeg/rdrle.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * rdrle.c
 *
 * Copyright (C) 1991-1996, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains routines to read input images in Utah RLE format.
 * The Utah Raster Toolkit library is required (version 3.1 or later).
 *
 * These routines may need modification for non-Unix environments or
 * specialized applications.  As they stand, they assume input from
 * an ordinary stdio stream.  They further assume that reading begins
 * at the start of the file; start_input may need work if the
 * user interface has already read some data (e.g., to determine that
 * the file is indeed RLE format).
 *
 * Based on code contributed by Mike Lijewski,
 * with updates from Robert Hutchinson.
 */

#include "cdjpeg.h"		/* Common decls for cjpeg/djpeg applications */

#ifdef RLE_SUPPORTED

/* rle.h is provided by the Utah Raster Toolkit. */

#include &lt;rle.h&gt;

/*
 * We assume that JSAMPLE has the same representation as rle_pixel,
 * to wit, "unsigned char".  Hence we can't cope with 12- or 16-bit samples.
 */

#if BITS_IN_JSAMPLE != 8
  Sorry, this code only copes with 8-bit JSAMPLEs. /* deliberate syntax err */
#endif

/*
 * We support the following types of RLE files:
 *   
 *   GRAYSCALE   - 8 bits, no colormap
 *   MAPPEDGRAY  - 8 bits, 1 channel colomap
 *   PSEUDOCOLOR - 8 bits, 3 channel colormap
 *   TRUECOLOR   - 24 bits, 3 channel colormap
 *   DIRECTCOLOR - 24 bits, no colormap
 *
 * For now, we ignore any alpha channel in the image.
 */

typedef enum
  { GRAYSCALE, MAPPEDGRAY, PSEUDOCOLOR, TRUECOLOR, DIRECTCOLOR } rle_kind;


/*
 * Since RLE stores scanlines bottom-to-top, we have to invert the image
 * to conform to JPEG's top-to-bottom order.  To do this, we read the
 * incoming image into a virtual array on the first get_pixel_rows call,
 * then fetch the required row from the virtual array on subsequent calls.
 */

typedef struct _rle_source_struct * rle_source_ptr;

typedef struct _rle_source_struct {
  struct cjpeg_source_struct pub; /* public fields */

  rle_kind visual;              /* actual type of input file */
  jvirt_sarray_ptr image;       /* virtual array to hold the image */
  JDIMENSION row;		/* current row # in the virtual array */
  rle_hdr header;               /* Input file information */
  rle_pixel** rle_row;          /* holds a row returned by rle_getrow() */

} rle_source_struct;


/*
 * Read the file header; return image size and component count.
 */

METHODDEF(void)
start_input_rle (j_compress_ptr cinfo, cjpeg_source_ptr sinfo)
{
  rle_source_ptr source = (rle_source_ptr) sinfo;
  JDIMENSION width, height;
#ifdef PROGRESS_REPORT
  cd_progress_ptr progress = (cd_progress_ptr) cinfo-&gt;progress;
#endif

  /* Use RLE library routine to get the header info */
  source-&gt;header = *rle_hdr_init(NULL);
  source-&gt;header.rle_file = source-&gt;pub.input_file;
  switch (rle_get_setup(&amp;(source-&gt;header))) {
  case RLE_SUCCESS:
    /* A-OK */
    break;
  case RLE_NOT_RLE:
    ERREXIT(cinfo, JERR_RLE_NOT);
    break;
  case RLE_NO_SPACE:
    ERREXIT(cinfo, JERR_RLE_MEM);
    break;
  case RLE_EMPTY:
    ERREXIT(cinfo, JERR_RLE_EMPTY);
    break;
  case RLE_EOF:
    ERREXIT(cinfo, JERR_RLE_EOF);
    break;
  default:
    ERREXIT(cinfo, JERR_RLE_BADERROR);
    break;
  }

  /* Figure out what we have, set private vars and return values accordingly */
  
  width  = source-&gt;header.xmax - source-&gt;header.xmin + 1;
  height = source-&gt;header.ymax - source-&gt;header.ymin + 1;
  source-&gt;header.xmin = 0;		/* realign horizontally */
  source-&gt;header.xmax = width-1;

  cinfo-&gt;image_width      = width;
  cinfo-&gt;image_height     = height;
  cinfo-&gt;data_precision   = 8;  /* we can only handle 8 bit data */

  if (source-&gt;header.ncolors == 1 &amp;&amp; source-&gt;header.ncmap == 0) {
    source-&gt;visual     = GRAYSCALE;
    TRACEMS2(cinfo, 1, JTRC_RLE_GRAY, width, height);
  } else if (source-&gt;header.ncolors == 1 &amp;&amp; source-&gt;header.ncmap == 1) {
    source-&gt;visual     = MAPPEDGRAY;
    TRACEMS3(cinfo, 1, JTRC_RLE_MAPGRAY, width, height,
             1 &lt;&lt; source-&gt;header.cmaplen);
  } else if (source-&gt;header.ncolors == 1 &amp;&amp; source-&gt;header.ncmap == 3) {
    source-&gt;visual     = PSEUDOCOLOR;
    TRACEMS3(cinfo, 1, JTRC_RLE_MAPPED, width, height,
	     1 &lt;&lt; source-&gt;header.cmaplen);
  } else if (source-&gt;header.ncolors == 3 &amp;&amp; source-&gt;header.ncmap == 3) {
    source-&gt;visual     = TRUECOLOR;
    TRACEMS3(cinfo, 1, JTRC_RLE_FULLMAP, width, height,
	     1 &lt;&lt; source-&gt;header.cmaplen);
  } else if (source-&gt;header.ncolors == 3 &amp;&amp; source-&gt;header.ncmap == 0) {
    source-&gt;visual     = DIRECTCOLOR;
    TRACEMS2(cinfo, 1, JTRC_RLE, width, height);
  } else
    ERREXIT(cinfo, JERR_RLE_UNSUPPORTED);
  
  if (source-&gt;visual == GRAYSCALE || source-&gt;visual == MAPPEDGRAY) {
    cinfo-&gt;in_color_space   = JCS_GRAYSCALE;
    cinfo-&gt;input_components = 1;
  } else {
    cinfo-&gt;in_color_space   = JCS_RGB;
    cinfo-&gt;input_components = 3;
  }

  /*
   * A place to hold each scanline while it's converted.
   * (GRAYSCALE scanlines don't need converting)
   */
  if (source-&gt;visual != GRAYSCALE) {
    source-&gt;rle_row = (rle_pixel**) (*cinfo-&gt;mem-&gt;alloc_sarray)
      ((j_common_ptr) cinfo, JPOOL_IMAGE,
       (JDIMENSION) width, (JDIMENSION) cinfo-&gt;input_components);
  }

  /* request a virtual array to hold the image */
  source-&gt;image = (*cinfo-&gt;mem-&gt;request_virt_sarray)
    ((j_common_ptr) cinfo, JPOOL_IMAGE, FALSE,
     (JDIMENSION) (width * source-&gt;header.ncolors),
     (JDIMENSION) height, (JDIMENSION) 1);

#ifdef PROGRESS_REPORT
  if (progress != NULL) {
    /* count file input as separate pass */
    progress-&gt;total_extra_passes++;
  }
#endif

  source-&gt;pub.buffer_height = 1;
}


/*
 * Read one row of pixels.
 * Called only after load_image has read the image into the virtual array.
 * Used for GRAYSCALE, MAPPEDGRAY, TRUECOLOR, and DIRECTCOLOR images.
 */

METHODDEF(JDIMENSION)
get_rle_row (j_compress_ptr cinfo, cjpeg_source_ptr sinfo)
{
  rle_source_ptr source = (rle_source_ptr) sinfo;

  source-&gt;row--;
  source-&gt;pub.buffer = (*cinfo-&gt;mem-&gt;access_virt_sarray)
    ((j_common_ptr) cinfo, source-&gt;image, source-&gt;row, (JDIMENSION) 1, FALSE);

  return 1;
}

/*
 * Read one row of pixels.
 * Called only after load_image has read the image into the virtual array.
 * Used for PSEUDOCOLOR images.
 */

METHODDEF(JDIMENSION)
get_pseudocolor_row (j_compress_ptr cinfo, cjpeg_source_ptr sinfo)
{
  rle_source_ptr source = (rle_source_ptr) sinfo;
  JSAMPROW src_row, dest_row;
  JDIMENSION col;
  rle_map *colormap;
  int val;

  colormap = source-&gt;header.cmap;
  dest_row = source-&gt;pub.buffer[0];
  source-&gt;row--;
  src_row = * (*cinfo-&gt;mem-&gt;access_virt_sarray)
    ((j_common_ptr) cinfo, source-&gt;image, source-&gt;row, (JDIMENSION) 1, FALSE);

  for (col = cinfo-&gt;image_width; col &gt; 0; col--) {
    val = GETJSAMPLE(*src_row++);
    *dest_row++ = (JSAMPLE) (colormap[val      ] &gt;&gt; 8);
    *dest_row++ = (JSAMPLE) (colormap[val + 256] &gt;&gt; 8);
    *dest_row++ = (JSAMPLE) (colormap[val + 512] &gt;&gt; 8);
  }

  return 1;
}


/*
 * Load the image into a virtual array.  We have to do this because RLE
 * files start at the lower left while the JPEG standard has them starting
 * in the upper left.  This is called the first time we want to get a row
 * of input.  What we do is load the RLE data into the array and then call
 * the appropriate routine to read one row from the array.  Before returning,
 * we set source-&gt;pub.get_pixel_rows so that subsequent calls go straight to
 * the appropriate row-reading routine.
 */

METHODDEF(JDIMENSION)
load_image (j_compress_ptr cinfo, cjpeg_source_ptr sinfo)
{
  rle_source_ptr source = (rle_source_ptr) sinfo;
  JDIMENSION row, col;
  JSAMPROW  scanline, red_ptr, green_ptr, blue_ptr;
  rle_pixel **rle_row;
  rle_map *colormap;
  char channel;
#ifdef PROGRESS_REPORT
  cd_progress_ptr progress = (cd_progress_ptr) cinfo-&gt;progress;
#endif

  colormap = source-&gt;header.cmap;
  rle_row = source-&gt;rle_row;

  /* Read the RLE data into our virtual array.
   * We assume here that (a) rle_pixel is represented the same as JSAMPLE,
   * and (b) we are not on a machine where FAR pointers differ from regular.
   */
  RLE_CLR_BIT(source-&gt;header, RLE_ALPHA); /* don't read the alpha channel */

#ifdef PROGRESS_REPORT
  if (progress != NULL) {
    progress-&gt;pub.pass_limit = cinfo-&gt;image_height;
    progress-&gt;pub.pass_counter = 0;
    (*progress-&gt;pub.progress_monitor) ((j_common_ptr) cinfo);
  }
#endif

  switch (source-&gt;visual) {

  case GRAYSCALE:
  case PSEUDOCOLOR:
    for (row = 0; row &lt; cinfo-&gt;image_height; row++) {
      rle_row = (rle_pixel **) (*cinfo-&gt;mem-&gt;access_virt_sarray)
         ((j_common_ptr) cinfo, source-&gt;image, row, (JDIMENSION) 1, TRUE);
      rle_getrow(&amp;source-&gt;header, rle_row);
#ifdef PROGRESS_REPORT
      if (progress != NULL) {
        progress-&gt;pub.pass_counter++;
        (*progress-&gt;pub.progress_monitor) ((j_common_ptr) cinfo);
      }
#endif
    }
    break;

  case MAPPEDGRAY:
  case TRUECOLOR:
    for (row = 0; row &lt; cinfo-&gt;image_height; row++) {
      scanline = * (*cinfo-&gt;mem-&gt;access_virt_sarray)
        ((j_common_ptr) cinfo, source-&gt;image, row, (JDIMENSION) 1, TRUE);
      rle_row = source-&gt;rle_row;
      rle_getrow(&amp;source-&gt;header, rle_row);

      for (col = 0; col &lt; cinfo-&gt;image_width; col++) {
        for (channel = 0; channel &lt; source-&gt;header.ncolors; channel++) {
          *scanline++ = (JSAMPLE)
            (colormap[GETJSAMPLE(rle_row[channel][col]) + 256 * channel] &gt;&gt; 8);
        }
      }

#ifdef PROGRESS_REPORT
      if (progress != NULL) {
        progress-&gt;pub.pass_counter++;
        (*progress-&gt;pub.progress_monitor) ((j_common_ptr) cinfo);
      }
#endif
    }
    break;

  case DIRECTCOLOR:
    for (row = 0; row &lt; cinfo-&gt;image_height; row++) {
      scanline = * (*cinfo-&gt;mem-&gt;access_virt_sarray)
        ((j_common_ptr) cinfo, source-&gt;image, row, (JDIMENSION) 1, TRUE);
      rle_getrow(&amp;source-&gt;header, rle_row);

      red_ptr   = rle_row[0];
      green_ptr = rle_row[1];
      blue_ptr  = rle_row[2];

      for (col = cinfo-&gt;image_width; col &gt; 0; col--) {
        *scanline++ = *red_ptr++;
        *scanline++ = *green_ptr++;
        *scanline++ = *blue_ptr++;
      }

#ifdef PROGRESS_REPORT
      if (progress != NULL) {
        progress-&gt;pub.pass_counter++;
        (*progress-&gt;pub.progress_monitor) ((j_common_ptr) cinfo);
      }
#endif
    }
  }

#ifdef PROGRESS_REPORT
  if (progress != NULL)
    progress-&gt;completed_extra_passes++;
#endif

  /* Set up to call proper row-extraction routine in future */
  if (source-&gt;visual == PSEUDOCOLOR) {
    source-&gt;pub.buffer = source-&gt;rle_row;
    source-&gt;pub.get_pixel_rows = get_pseudocolor_row;
  } else {
    source-&gt;pub.get_pixel_rows = get_rle_row;
  }
  source-&gt;row = cinfo-&gt;image_height;

  /* And fetch the topmost (bottommost) row */
  return (*source-&gt;pub.get_pixel_rows) (cinfo, sinfo);   
}


/*
 * Finish up at the end of the file.
 */

METHODDEF(void)
finish_input_rle (j_compress_ptr cinfo, cjpeg_source_ptr sinfo)
{
  /* no work */
}


/*
 * The module selection routine for RLE format input.
 */

GLOBAL(cjpeg_source_ptr)
jinit_read_rle (j_compress_ptr cinfo)
{
  rle_source_ptr source;

  /* Create module interface object */
  source = (rle_source_ptr)
      (*cinfo-&gt;mem-&gt;alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
                                  SIZEOF(rle_source_struct));
  /* Fill in method ptrs */
  source-&gt;pub.start_input = start_input_rle;
  source-&gt;pub.finish_input = finish_input_rle;
  source-&gt;pub.get_pixel_rows = load_image;

  return (cjpeg_source_ptr) source;
}

#endif /* RLE_SUPPORTED */
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

