<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/ip/ftpfs/file.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/ip/ftpfs/file.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;String.h&gt;
#include "ftpfs.h"

enum
{
	Chunk=		1024,		/* chunk size for buffered data */
	Nfile=		128,		/* maximum number of cached files */
};

/* a file (with cached data) */
struct File
{
	char	*mem;		/* part of file cached in memory */
	ulong	len;		/* length of cached data */
	long	off;		/* current offset into tpath */
	short	fd;		/* fd to cache file */
	char	inuse;
	char	dirty;
	ulong	atime;		/* time of last access */
	Node	*node;
	char 	*template;
};

static File	files[Nfile];
static ulong	now;
static int	ntmp;

/*
 *  lookup a file, create one if not found.  if there are no
 *  free files, free the last oldest clean one.
 */
static File*
fileget(Node *node)
{
	File *fp;
	File *oldest;

	fp = node-&gt;fp;
	if(fp)
		return fp;

	oldest = 0;
	for(fp = files; fp &lt; &amp;files[Nfile]; fp++){
		if(fp-&gt;inuse == 0)
			break;
		if(fp-&gt;dirty == 0 &amp;&amp; (oldest == 0 || oldest-&gt;atime &gt; fp-&gt;atime))
			oldest = fp;
	}

	if(fp == &amp;files[Nfile]){
		uncache(oldest-&gt;node);
		fp = oldest;
	}
	node-&gt;fp = fp;
	fp-&gt;node = node;
	fp-&gt;atime = now++;
	fp-&gt;inuse = 1;
	fp-&gt;fd = -1;
	if(fp-&gt;mem){
		free(fp-&gt;mem);
		fp-&gt;mem = nil;
	}
	return fp;
}

/*
 *  free a cached file
 */
void
filefree(Node *node)
{
	File *fp;

	fp = node-&gt;fp;
	if(fp == 0)
		return;

	if(fp-&gt;fd &gt; 0){
		ntmp--;
		close(fp-&gt;fd);
		remove(fp-&gt;template);
		free(fp-&gt;template);
		fp-&gt;template = 0;
	}
	fp-&gt;fd = -1;
	if(fp-&gt;mem){
		free(fp-&gt;mem);
		fp-&gt;mem = nil;
	}
	fp-&gt;len = 0;
	fp-&gt;inuse = 0;
	fp-&gt;dirty = 0;

	node-&gt;fp = 0;
}

/*
 *  satisfy read first from in memory chunk and then from temporary
 *  file.  It's up to the caller to make sure that the file is valid.
 */
int
fileread(Node *node, char *a, long off, int n)
{
	int sofar;
	int i;
	File *fp;

	fp = node-&gt;fp;
	if(fp == 0)
		fatal("fileread");

	if(off + n &gt; fp-&gt;len)
		n = fp-&gt;len - off;

	for(sofar = 0; sofar &lt; n; sofar += i, off += i, a += i){
		if(off &gt;= fp-&gt;len)
			return sofar;
		if(off &lt; Chunk){
			i = n;
			if(off + i &gt; Chunk)
				i = Chunk - off;
			memmove(a, fp-&gt;mem + off, i);
			continue;
		}
		if(fp-&gt;off != off)
			if(seek(fp-&gt;fd, off, 0) &lt; 0){
				fp-&gt;off = -1;
				return -1;
			}
		i = read(fp-&gt;fd, a, n-sofar);
		if(i &lt; 0){
			fp-&gt;off = -1;
			return -1;
		}
		if(i == 0)
			break;
		fp-&gt;off = off + i;
	}
	return sofar;
}

void
uncachedir(Node *parent, Node *child)
{
	Node *sp;

	if(parent == 0 || parent == child)
		return;
	for(sp = parent-&gt;children; sp; sp = sp-&gt;sibs)
		if(sp-&gt;opens == 0)
		if(sp != child)
		if(sp-&gt;fp != nil)
		if(sp-&gt;fp-&gt;dirty == 0)
		if(sp-&gt;fp-&gt;fd &gt;= 0){
			filefree(sp);
			UNCACHED(sp);
		}
}

static int
createtmp(File *fp)
{
	char template[32];

	strcpy(template, "/tmp/ftpXXXXXXXXXXX");
	mktemp(template);
	if(strcmp(template, "/") == 0){
		fprint(2, "ftpfs can't create tmp file %s: %r\n", template);
		return -1;
	}
	if(ntmp &gt;= 16)
		uncachedir(fp-&gt;node-&gt;parent, fp-&gt;node);

	fp-&gt;fd = create(template, ORDWR|ORCLOSE, 0600);
	fp-&gt;template = strdup(template);
	fp-&gt;off = 0;
	ntmp++;
	return fp-&gt;fd;
}

/*
 *  write cached data (first Chunk bytes stay in memory)
 */
int
filewrite(Node *node, char *a, long off, int n)
{
	int i, sofar;
	File *fp;

	fp = fileget(node);

	if(fp-&gt;mem == nil){
		fp-&gt;mem = malloc(Chunk);
		if(fp-&gt;mem == nil)
			return seterr("out of memory");
	}

	for(sofar = 0; sofar &lt; n; sofar += i, off += i, a += i){
		if(off &lt; Chunk){
			i = n;
			if(off + i &gt; Chunk)
				i = Chunk - off;
			memmove(fp-&gt;mem + off, a, i);
			continue;
		}
		if(fp-&gt;fd &lt; 0)
			if(createtmp(fp) &lt; 0)
				return seterr("can't create temp file");
		if(fp-&gt;off != off)
			if(seek(fp-&gt;fd, off, 0) &lt; 0){
				fp-&gt;off = -1;
				return seterr("can't seek temp file");
			}
		i = write(fp-&gt;fd, a, n-sofar);
		if(i &lt;= 0){
			fp-&gt;off = -1;
			return seterr("can't write temp file");
		}
		fp-&gt;off = off + i;
	}

	if(off &gt; fp-&gt;len)
		fp-&gt;len = off;
	if(off &gt; node-&gt;d-&gt;length)
		node-&gt;d-&gt;length = off;
	return sofar;
}

/*
 *  mark a file as dirty
 */
void
filedirty(Node *node)
{
	File *fp;

	fp = fileget(node);
	fp-&gt;dirty = 1;
}

/*
 *  mark a file as clean
 */
void
fileclean(Node *node)
{
	if(node-&gt;fp)
		node-&gt;fp-&gt;dirty = 0;
}

int
fileisdirty(Node *node)
{
	return node-&gt;fp &amp;&amp; node-&gt;fp-&gt;dirty;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

