<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/scat/qtree.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/scat/qtree.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include	&lt;u.h&gt;
#include	&lt;libc.h&gt;
#include	&lt;bio.h&gt;
#include	"sky.h"

static void	qtree_expand(Biobuf*, uchar*, int, int, uchar*);
static void	qtree_copy(uchar*, int, int, uchar*, int);
static void	qtree_bitins(uchar*, int, int, Pix*, int, int);
static void	read_bdirect(Biobuf*, Pix*, int, int, int, uchar*, int);

void
qtree_decode(Biobuf *infile, Pix *a, int n, int nqx, int nqy, int nbitplanes)
{
	int log2n, k, bit, b, nqmax;
	int nx,ny,nfx,nfy,c;
	int nqx2, nqy2;
	unsigned char *scratch;

	/*
	 * log2n is log2 of max(nqx,nqy) rounded up to next power of 2
	 */
	nqmax = nqy;
	if(nqx &gt; nqmax)
		nqmax = nqx;
	log2n = log(nqmax)/LN2+0.5;
	if (nqmax &gt; (1&lt;&lt;log2n))
		log2n++;

	/*
	 * allocate scratch array for working space
	 */
	nqx2 = (nqx+1)/2;
	nqy2 = (nqy+1)/2;
	scratch = (uchar*)malloc(nqx2*nqy2);
	if(scratch == nil) {
		fprint(2, "qtree_decode: insufficient memory\n");
		exits("memory");
	}

	/*
	 * now decode each bit plane, starting at the top
	 * A is assumed to be initialized to zero
	 */
	for(bit = nbitplanes-1; bit &gt;= 0; bit--) {
		/*
		 * Was bitplane was quadtree-coded or written directly?
		 */
		b = input_nybble(infile);
		if(b == 0) {
			/*
			 * bit map was written directly
			 */
			read_bdirect(infile, a, n, nqx, nqy, scratch, bit);
		} else
		if(b != 0xf) {
			fprint(2, "qtree_decode: bad format code %x\n",b);
			exits("format");
		} else {
			/*
			 * bitmap was quadtree-coded, do log2n expansions
			 * read first code
			 */

			scratch[0] = input_huffman(infile);

			/*
			 * do log2n expansions, reading codes from file as necessary
			 */
			nx = 1;
			ny = 1;
			nfx = nqx;
			nfy = nqy;
			c = 1&lt;&lt;log2n;
			for(k = 1; k&lt;log2n; k++) {
				/*
				 * this somewhat cryptic code generates the sequence
				 * n[k-1] = (n[k]+1)/2 where n[log2n]=nqx or nqy
				 */
				c = c&gt;&gt;1;
				nx = nx&lt;&lt;1;
				ny = ny&lt;&lt;1;
				if(nfx &lt;= c)
					nx--;
				else
					nfx -= c;
				if(nfy &lt;= c)
					ny--;
				else
					nfy -= c;
				qtree_expand(infile, scratch, nx, ny, scratch);
			}

			/*
			 * copy last set of 4-bit codes to bitplane bit of array a
			 */
			qtree_bitins(scratch, nqx, nqy, a, n, bit);
		}
	}
	free(scratch);
}

/*
 * do one quadtree expansion step on array a[(nqx+1)/2,(nqy+1)/2]
 * results put into b[nqx,nqy] (which may be the same as a)
 */
static
void
qtree_expand(Biobuf *infile, uchar *a, int nx, int ny, uchar *b)
{
	uchar *b1;

	/*
	 * first copy a to b, expanding each 4-bit value
	 */
	qtree_copy(a, nx, ny, b, ny);

	/*
	 * now read new 4-bit values into b for each non-zero element
	 */
	b1 = &amp;b[nx*ny];
	while(b1 &gt; b) {
		b1--;
		if(*b1 != 0)
			*b1 = input_huffman(infile);
	}
}

/*
 * copy 4-bit values from a[(nx+1)/2,(ny+1)/2] to b[nx,ny], expanding
 * each value to 2x2 pixels
 * a,b may be same array
 */
static
void
qtree_copy(uchar *a, int nx, int ny, uchar *b, int n)
{
	int i, j, k, nx2, ny2;
	int s00, s10;

	/*
	 * first copy 4-bit values to b
	 * start at end in case a,b are same array
	 */
	nx2 = (nx+1)/2;
	ny2 = (ny+1)/2;
	k = ny2*(nx2-1) + ny2-1;	/* k   is index of a[i,j] */
	for (i = nx2-1; i &gt;= 0; i--) {
		s00 = 2*(n*i+ny2-1);	/* s00 is index of b[2*i,2*j] */
		for (j = ny2-1; j &gt;= 0; j--) {
			b[s00] = a[k];
			k -= 1;
			s00 -= 2;
		}
	}

	/*
	 * now expand each 2x2 block
	 */
	for(i = 0; i&lt;nx-1; i += 2) {
		s00 = n*i;				/* s00 is index of b[i,j] */
		s10 = s00+n;				/* s10 is index of b[i+1,j] */
		for(j = 0; j&lt;ny-1; j += 2) {
			b[s10+1] =  b[s00]     &amp; 1;
			b[s10  ] = (b[s00]&gt;&gt;1) &amp; 1;
			b[s00+1] = (b[s00]&gt;&gt;2) &amp; 1;
			b[s00  ] = (b[s00]&gt;&gt;3) &amp; 1;
			s00 += 2;
			s10 += 2;
		}
		if(j &lt; ny) {
			/*
			 * row size is odd, do last element in row
			 * s00+1, s10+1 are off edge
			 */
			b[s10  ] = (b[s00]&gt;&gt;1) &amp; 1;
			b[s00  ] = (b[s00]&gt;&gt;3) &amp; 1;
		}
	}
	if(i &lt; nx) {
		/*
		 * column size is odd, do last row
		 * s10, s10+1 are off edge
		 */
		s00 = n*i;
		for (j = 0; j&lt;ny-1; j += 2) {
			b[s00+1] = (b[s00]&gt;&gt;2) &amp; 1;
			b[s00  ] = (b[s00]&gt;&gt;3) &amp; 1;
			s00 += 2;
		}
		if(j &lt; ny) {
			/*
			 * both row and column size are odd, do corner element
			 * s00+1, s10, s10+1 are off edge
			 */
			b[s00  ] = (b[s00]&gt;&gt;3) &amp; 1;
		}
	}
}

/*
 * Copy 4-bit values from a[(nx+1)/2,(ny+1)/2] to b[nx,ny], expanding
 * each value to 2x2 pixels and inserting into bitplane BIT of B.
 * A,B may NOT be same array (it wouldn't make sense to be inserting
 * bits into the same array anyway.)
 */
static
void
qtree_bitins(uchar *a, int nx, int ny, Pix *b, int n, int bit)
{
	int i, j;
	Pix *s00, *s10;
	Pix px;

	/*
	 * expand each 2x2 block
	 */
	for(i=0; i&lt;nx-1; i+=2) {
		s00 = &amp;b[n*i];			/* s00 is index of b[i,j] */
		s10 = s00+n;			/* s10 is index of b[i+1,j] */
		for(j=0; j&lt;ny-1; j+=2) {
			px = *a++;
			s10[1] |= ( px     &amp; 1) &lt;&lt; bit;
			s10[0] |= ((px&gt;&gt;1) &amp; 1) &lt;&lt; bit;
			s00[1] |= ((px&gt;&gt;2) &amp; 1) &lt;&lt; bit;
			s00[0] |= ((px&gt;&gt;3) &amp; 1) &lt;&lt; bit;
			s00 += 2;
			s10 += 2;
		}
		if(j &lt; ny) {
			/*
			 * row size is odd, do last element in row
			 * s00+1, s10+1 are off edge
			 */
			px = *a++;
			s10[0] |= ((px&gt;&gt;1) &amp; 1) &lt;&lt; bit;
			s00[0] |= ((px&gt;&gt;3) &amp; 1) &lt;&lt; bit;
		}
	}
	if(i &lt; nx) {
		/*
		 * column size is odd, do last row
		 * s10, s10+1 are off edge
		 */
		s00 = &amp;b[n*i];
		for(j=0; j&lt;ny-1; j+=2) {
			px = *a++;
			s00[1] |= ((px&gt;&gt;2) &amp; 1) &lt;&lt; bit;
			s00[0] |= ((px&gt;&gt;3) &amp; 1) &lt;&lt; bit;
			s00 += 2;
		}
		if(j &lt; ny) {
			/*
			 * both row and column size are odd, do corner element
			 * s00+1, s10, s10+1 are off edge
			 */
			s00[0] |= ((*a&gt;&gt;3) &amp; 1) &lt;&lt; bit;
		}
	}
}

static
void
read_bdirect(Biobuf *infile, Pix *a, int n, int nqx, int nqy, uchar *scratch, int bit)
{
	int i;

	/*
	 * read bit image packed 4 pixels/nybble
	 */
	for(i = 0; i &lt; ((nqx+1)/2) * ((nqy+1)/2); i++) {
		scratch[i] = input_nybble(infile);
	}

	/*
	 * insert in bitplane BIT of image A
	 */
	qtree_bitins(scratch, nqx, nqy, a, n, bit);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

