<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/unix/drawterm/kern/qio.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/unix/drawterm/kern/qio.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include	"u.h"
#include	"lib.h"
#include	"dat.h"
#include	"fns.h"
#include	"error.h"

static ulong padblockcnt;
static ulong concatblockcnt;
static ulong pullupblockcnt;
static ulong copyblockcnt;
static ulong consumecnt;
static ulong producecnt;
static ulong qcopycnt;

static int debugging;

#define QDEBUG	if(0)

/*
 *  IO queues
 */
struct Queue
{
	Lock lk;

	Block*	bfirst;		/* buffer */
	Block*	blast;

	int	len;		/* bytes allocated to queue */
	int	dlen;		/* data bytes in queue */
	int	limit;		/* max bytes in queue */
	int	inilim;		/* initial limit */
	int	state;
	int	noblock;	/* true if writes return immediately when q full */
	int	eof;		/* number of eofs read by user */

	void	(*kick)(void*);	/* restart output */
	void	(*bypass)(void*, Block*);	/* bypass queue altogether */
	void*	arg;		/* argument to kick */

	QLock	rlock;		/* mutex for reading processes */
	Rendez	rr;		/* process waiting to read */
	QLock	wlock;		/* mutex for writing processes */
	Rendez	wr;		/* process waiting to write */

	char	err[ERRMAX];
};

enum
{
	Maxatomic	= 64*1024,
};

uint	qiomaxatomic = Maxatomic;

void
ixsummary(void)
{
	debugging ^= 1;
	iallocsummary();
	print("pad %lud, concat %lud, pullup %lud, copy %lud\n",
		padblockcnt, concatblockcnt, pullupblockcnt, copyblockcnt);
	print("consume %lud, produce %lud, qcopy %lud\n",
		consumecnt, producecnt, qcopycnt);
}

/*
 *  free a list of blocks
 */
void
freeblist(Block *b)
{
	Block *next;

	for(; b != 0; b = next){
		next = b-&gt;next;
		b-&gt;next = 0;
		freeb(b);
	}
}

/*
 *  pad a block to the front (or the back if size is negative)
 */
Block*
padblock(Block *bp, int size)
{
	int n;
	Block *nbp;

	QDEBUG checkb(bp, "padblock 1");
	if(size &gt;= 0){
		if(bp-&gt;rp - bp-&gt;base &gt;= size){
			bp-&gt;rp -= size;
			return bp;
		}

		if(bp-&gt;next)
			panic("padblock 0x%p", getcallerpc(&amp;bp));
		n = BLEN(bp);
		padblockcnt++;
		nbp = allocb(size+n);
		nbp-&gt;rp += size;
		nbp-&gt;wp = nbp-&gt;rp;
		memmove(nbp-&gt;wp, bp-&gt;rp, n);
		nbp-&gt;wp += n;
		freeb(bp);
		nbp-&gt;rp -= size;
	} else {
		size = -size;

		if(bp-&gt;next)
			panic("padblock 0x%p", getcallerpc(&amp;bp));

		if(bp-&gt;lim - bp-&gt;wp &gt;= size)
			return bp;

		n = BLEN(bp);
		padblockcnt++;
		nbp = allocb(size+n);
		memmove(nbp-&gt;wp, bp-&gt;rp, n);
		nbp-&gt;wp += n;
		freeb(bp);
	}
	QDEBUG checkb(nbp, "padblock 1");
	return nbp;
}

/*
 *  return count of bytes in a string of blocks
 */
int
blocklen(Block *bp)
{
	int len;

	len = 0;
	while(bp) {
		len += BLEN(bp);
		bp = bp-&gt;next;
	}
	return len;
}

/*
 * return count of space in blocks
 */
int
blockalloclen(Block *bp)
{
	int len;

	len = 0;
	while(bp) {
		len += BALLOC(bp);
		bp = bp-&gt;next;
	}
	return len;
}

/*
 *  copy the  string of blocks into
 *  a single block and free the string
 */
Block*
concatblock(Block *bp)
{
	int len;
	Block *nb, *f;

	if(bp-&gt;next == 0)
		return bp;

	nb = allocb(blocklen(bp));
	for(f = bp; f; f = f-&gt;next) {
		len = BLEN(f);
		memmove(nb-&gt;wp, f-&gt;rp, len);
		nb-&gt;wp += len;
	}
	concatblockcnt += BLEN(nb);
	freeblist(bp);
	QDEBUG checkb(nb, "concatblock 1");
	return nb;
}

/*
 *  make sure the first block has at least n bytes
 */
Block*
pullupblock(Block *bp, int n)
{
	int i;
	Block *nbp;

	/*
	 *  this should almost always be true, it's
	 *  just to avoid every caller checking.
	 */
	if(BLEN(bp) &gt;= n)
		return bp;

	/*
	 *  if not enough room in the first block,
	 *  add another to the front of the list.
	 */
	if(bp-&gt;lim - bp-&gt;rp &lt; n){
		nbp = allocb(n);
		nbp-&gt;next = bp;
		bp = nbp;
	}

	/*
	 *  copy bytes from the trailing blocks into the first
	 */
	n -= BLEN(bp);
	while((nbp = bp-&gt;next)){
		i = BLEN(nbp);
		if(i &gt; n) {
			memmove(bp-&gt;wp, nbp-&gt;rp, n);
			pullupblockcnt++;
			bp-&gt;wp += n;
			nbp-&gt;rp += n;
			QDEBUG checkb(bp, "pullupblock 1");
			return bp;
		} else {
			/* shouldn't happen but why crash if it does */
			if(i &lt; 0){
				print("pullup negative length packet\n");
				i = 0;
			}
			memmove(bp-&gt;wp, nbp-&gt;rp, i);
			pullupblockcnt++;
			bp-&gt;wp += i;
			bp-&gt;next = nbp-&gt;next;
			nbp-&gt;next = 0;
			freeb(nbp);
			n -= i;
			if(n == 0){
				QDEBUG checkb(bp, "pullupblock 2");
				return bp;
			}
		}
	}
	freeb(bp);
	return 0;
}

/*
 *  make sure the first block has at least n bytes
 */
Block*
pullupqueue(Queue *q, int n)
{
	Block *b;

	if(BLEN(q-&gt;bfirst) &gt;= n)
		return q-&gt;bfirst;
	q-&gt;bfirst = pullupblock(q-&gt;bfirst, n);
	for(b = q-&gt;bfirst; b != nil &amp;&amp; b-&gt;next != nil; b = b-&gt;next)
		;
	q-&gt;blast = b;
	return q-&gt;bfirst;
}

/*
 *  trim to len bytes starting at offset
 */
Block *
trimblock(Block *bp, int offset, int len)
{
	ulong l;
	Block *nb, *startb;

	QDEBUG checkb(bp, "trimblock 1");
	if(blocklen(bp) &lt; offset+len) {
		freeblist(bp);
		return nil;
	}

	while((l = BLEN(bp)) &lt; offset) {
		offset -= l;
		nb = bp-&gt;next;
		bp-&gt;next = nil;
		freeb(bp);
		bp = nb;
	}

	startb = bp;
	bp-&gt;rp += offset;

	while((l = BLEN(bp)) &lt; len) {
		len -= l;
		bp = bp-&gt;next;
	}

	bp-&gt;wp -= (BLEN(bp) - len);

	if(bp-&gt;next) {
		freeblist(bp-&gt;next);
		bp-&gt;next = nil;
	}

	return startb;
}

/*
 *  copy 'count' bytes into a new block
 */
Block*
copyblock(Block *bp, int count)
{
	int l;
	Block *nbp;

	QDEBUG checkb(bp, "copyblock 0");
	nbp = allocb(count);
	for(; count &gt; 0 &amp;&amp; bp != 0; bp = bp-&gt;next){
		l = BLEN(bp);
		if(l &gt; count)
			l = count;
		memmove(nbp-&gt;wp, bp-&gt;rp, l);
		nbp-&gt;wp += l;
		count -= l;
	}
	if(count &gt; 0){
		memset(nbp-&gt;wp, 0, count);
		nbp-&gt;wp += count;
	}
	copyblockcnt++;
	QDEBUG checkb(nbp, "copyblock 1");

	return nbp;
}

Block*
adjustblock(Block* bp, int len)
{
	int n;
	Block *nbp;

	if(len &lt; 0){
		freeb(bp);
		return nil;
	}

	if(bp-&gt;rp+len &gt; bp-&gt;lim){
		nbp = copyblock(bp, len);
		freeblist(bp);
		QDEBUG checkb(nbp, "adjustblock 1");

		return nbp;
	}

	n = BLEN(bp);
	if(len &gt; n)
		memset(bp-&gt;wp, 0, len-n);
	bp-&gt;wp = bp-&gt;rp+len;
	QDEBUG checkb(bp, "adjustblock 2");

	return bp;
}


/*
 *  throw away up to count bytes from a
 *  list of blocks.  Return count of bytes
 *  thrown away.
 */
int
pullblock(Block **bph, int count)
{
	Block *bp;
	int n, bytes;

	bytes = 0;
	if(bph == nil)
		return 0;

	while(*bph != nil &amp;&amp; count != 0) {
		bp = *bph;
		n = BLEN(bp);
		if(count &lt; n)
			n = count;
		bytes += n;
		count -= n;
		bp-&gt;rp += n;
		QDEBUG checkb(bp, "pullblock ");
		if(BLEN(bp) == 0) {
			*bph = bp-&gt;next;
			bp-&gt;next = nil;
			freeb(bp);
		}
	}
	return bytes;
}

/*
 *  get next block from a queue, return null if nothing there
 */
Block*
qget(Queue *q)
{
	int dowakeup;
	Block *b;

	/* sync with qwrite */
	ilock(&amp;q-&gt;lk);

	b = q-&gt;bfirst;
	if(b == nil){
		q-&gt;state |= Qstarve;
		iunlock(&amp;q-&gt;lk);
		return nil;
	}
	q-&gt;bfirst = b-&gt;next;
	b-&gt;next = 0;
	q-&gt;len -= BALLOC(b);
	q-&gt;dlen -= BLEN(b);
	QDEBUG checkb(b, "qget");

	/* if writer flow controlled, restart */
	if((q-&gt;state &amp; Qflow) &amp;&amp; q-&gt;len &lt; q-&gt;limit/2){
		q-&gt;state &amp;= ~Qflow;
		dowakeup = 1;
	} else
		dowakeup = 0;

	iunlock(&amp;q-&gt;lk);

	if(dowakeup)
		wakeup(&amp;q-&gt;wr);

	return b;
}

/*
 *  throw away the next 'len' bytes in the queue
 */
int
qdiscard(Queue *q, int len)
{
	Block *b;
	int dowakeup, n, sofar;

	ilock(&amp;q-&gt;lk);
	for(sofar = 0; sofar &lt; len; sofar += n){
		b = q-&gt;bfirst;
		if(b == nil)
			break;
		QDEBUG checkb(b, "qdiscard");
		n = BLEN(b);
		if(n &lt;= len - sofar){
			q-&gt;bfirst = b-&gt;next;
			b-&gt;next = 0;
			q-&gt;len -= BALLOC(b);
			q-&gt;dlen -= BLEN(b);
			freeb(b);
		} else {
			n = len - sofar;
			b-&gt;rp += n;
			q-&gt;dlen -= n;
		}
	}

	/*
	 *  if writer flow controlled, restart
	 *
	 *  This used to be
	 *	q-&gt;len &lt; q-&gt;limit/2
	 *  but it slows down tcp too much for certain write sizes.
	 *  I really don't understand it completely.  It may be
	 *  due to the queue draining so fast that the transmission
	 *  stalls waiting for the app to produce more data.  - presotto
	 */
	if((q-&gt;state &amp; Qflow) &amp;&amp; q-&gt;len &lt; q-&gt;limit){
		q-&gt;state &amp;= ~Qflow;
		dowakeup = 1;
	} else
		dowakeup = 0;

	iunlock(&amp;q-&gt;lk);

	if(dowakeup)
		wakeup(&amp;q-&gt;wr);

	return sofar;
}

/*
 *  Interrupt level copy out of a queue, return # bytes copied.
 */
int
qconsume(Queue *q, void *vp, int len)
{
	Block *b;
	int n, dowakeup;
	uchar *p = vp;
	Block *tofree = nil;

	/* sync with qwrite */
	ilock(&amp;q-&gt;lk);

	for(;;) {
		b = q-&gt;bfirst;
		if(b == 0){
			q-&gt;state |= Qstarve;
			iunlock(&amp;q-&gt;lk);
			return -1;
		}
		QDEBUG checkb(b, "qconsume 1");

		n = BLEN(b);
		if(n &gt; 0)
			break;
		q-&gt;bfirst = b-&gt;next;
		q-&gt;len -= BALLOC(b);

		/* remember to free this */
		b-&gt;next = tofree;
		tofree = b;
	};

	if(n &lt; len)
		len = n;
	memmove(p, b-&gt;rp, len);
	consumecnt += n;
	b-&gt;rp += len;
	q-&gt;dlen -= len;

	/* discard the block if we're done with it */
	if((q-&gt;state &amp; Qmsg) || len == n){
		q-&gt;bfirst = b-&gt;next;
		b-&gt;next = 0;
		q-&gt;len -= BALLOC(b);
		q-&gt;dlen -= BLEN(b);

		/* remember to free this */
		b-&gt;next = tofree;
		tofree = b;
	}

	/* if writer flow controlled, restart */
	if((q-&gt;state &amp; Qflow) &amp;&amp; q-&gt;len &lt; q-&gt;limit/2){
		q-&gt;state &amp;= ~Qflow;
		dowakeup = 1;
	} else
		dowakeup = 0;

	iunlock(&amp;q-&gt;lk);

	if(dowakeup)
		wakeup(&amp;q-&gt;wr);

	if(tofree != nil)
		freeblist(tofree);

	return len;
}

int
qpass(Queue *q, Block *b)
{
	int dlen, len, dowakeup;

	/* sync with qread */
	dowakeup = 0;
	ilock(&amp;q-&gt;lk);
	if(q-&gt;len &gt;= q-&gt;limit){
		freeblist(b);
		iunlock(&amp;q-&gt;lk);
		return -1;
	}
	if(q-&gt;state &amp; Qclosed){
		freeblist(b);
		iunlock(&amp;q-&gt;lk);
		return BALLOC(b);
	}

	/* add buffer to queue */
	if(q-&gt;bfirst)
		q-&gt;blast-&gt;next = b;
	else
		q-&gt;bfirst = b;
	len = BALLOC(b);
	dlen = BLEN(b);
	QDEBUG checkb(b, "qpass");
	while(b-&gt;next){
		b = b-&gt;next;
		QDEBUG checkb(b, "qpass");
		len += BALLOC(b);
		dlen += BLEN(b);
	}
	q-&gt;blast = b;
	q-&gt;len += len;
	q-&gt;dlen += dlen;

	if(q-&gt;len &gt;= q-&gt;limit/2)
		q-&gt;state |= Qflow;

	if(q-&gt;state &amp; Qstarve){
		q-&gt;state &amp;= ~Qstarve;
		dowakeup = 1;
	}
	iunlock(&amp;q-&gt;lk);

	if(dowakeup)
		wakeup(&amp;q-&gt;rr);

	return len;
}

int
qpassnolim(Queue *q, Block *b)
{
	int dlen, len, dowakeup;

	/* sync with qread */
	dowakeup = 0;
	ilock(&amp;q-&gt;lk);

	if(q-&gt;state &amp; Qclosed){
		freeblist(b);
		iunlock(&amp;q-&gt;lk);
		return BALLOC(b);
	}

	/* add buffer to queue */
	if(q-&gt;bfirst)
		q-&gt;blast-&gt;next = b;
	else
		q-&gt;bfirst = b;
	len = BALLOC(b);
	dlen = BLEN(b);
	QDEBUG checkb(b, "qpass");
	while(b-&gt;next){
		b = b-&gt;next;
		QDEBUG checkb(b, "qpass");
		len += BALLOC(b);
		dlen += BLEN(b);
	}
	q-&gt;blast = b;
	q-&gt;len += len;
	q-&gt;dlen += dlen;

	if(q-&gt;len &gt;= q-&gt;limit/2)
		q-&gt;state |= Qflow;

	if(q-&gt;state &amp; Qstarve){
		q-&gt;state &amp;= ~Qstarve;
		dowakeup = 1;
	}
	iunlock(&amp;q-&gt;lk);

	if(dowakeup)
		wakeup(&amp;q-&gt;rr);

	return len;
}

/*
 *  if the allocated space is way out of line with the used
 *  space, reallocate to a smaller block
 */
Block*
packblock(Block *bp)
{
	Block **l, *nbp;
	int n;

	for(l = &amp;bp; *l; l = &amp;(*l)-&gt;next){
		nbp = *l;
		n = BLEN(nbp);
		if((n&lt;&lt;2) &lt; BALLOC(nbp)){
			*l = allocb(n);
			memmove((*l)-&gt;wp, nbp-&gt;rp, n);
			(*l)-&gt;wp += n;
			(*l)-&gt;next = nbp-&gt;next;
			freeb(nbp);
		}
	}

	return bp;
}

int
qproduce(Queue *q, void *vp, int len)
{
	Block *b;
	int dowakeup;
	uchar *p = vp;

	/* sync with qread */
	dowakeup = 0;
	ilock(&amp;q-&gt;lk);

	/* no waiting receivers, room in buffer? */
	if(q-&gt;len &gt;= q-&gt;limit){
		q-&gt;state |= Qflow;
		iunlock(&amp;q-&gt;lk);
		return -1;
	}

	/* save in buffer */
	b = iallocb(len);
	if(b == 0){
		iunlock(&amp;q-&gt;lk);
		return 0;
	}
	memmove(b-&gt;wp, p, len);
	producecnt += len;
	b-&gt;wp += len;
	if(q-&gt;bfirst)
		q-&gt;blast-&gt;next = b;
	else
		q-&gt;bfirst = b;
	q-&gt;blast = b;
	/* b-&gt;next = 0; done by iallocb() */
	q-&gt;len += BALLOC(b);
	q-&gt;dlen += BLEN(b);
	QDEBUG checkb(b, "qproduce");

	if(q-&gt;state &amp; Qstarve){
		q-&gt;state &amp;= ~Qstarve;
		dowakeup = 1;
	}

	if(q-&gt;len &gt;= q-&gt;limit)
		q-&gt;state |= Qflow;
	iunlock(&amp;q-&gt;lk);

	if(dowakeup)
		wakeup(&amp;q-&gt;rr);

	return len;
}

/*
 *  copy from offset in the queue
 */
Block*
qcopy(Queue *q, int len, ulong offset)
{
	int sofar;
	int n;
	Block *b, *nb;
	uchar *p;

	nb = allocb(len);

	ilock(&amp;q-&gt;lk);

	/* go to offset */
	b = q-&gt;bfirst;
	for(sofar = 0; ; sofar += n){
		if(b == nil){
			iunlock(&amp;q-&gt;lk);
			return nb;
		}
		n = BLEN(b);
		if(sofar + n &gt; offset){
			p = b-&gt;rp + offset - sofar;
			n -= offset - sofar;
			break;
		}
		QDEBUG checkb(b, "qcopy");
		b = b-&gt;next;
	}

	/* copy bytes from there */
	for(sofar = 0; sofar &lt; len;){
		if(n &gt; len - sofar)
			n = len - sofar;
		memmove(nb-&gt;wp, p, n);
		qcopycnt += n;
		sofar += n;
		nb-&gt;wp += n;
		b = b-&gt;next;
		if(b == nil)
			break;
		n = BLEN(b);
		p = b-&gt;rp;
	}
	iunlock(&amp;q-&gt;lk);

	return nb;
}

/*
 *  called by non-interrupt code
 */
Queue*
qopen(int limit, int msg, void (*kick)(void*), void *arg)
{
	Queue *q;

	q = malloc(sizeof(Queue));
	if(q == 0)
		return 0;

	q-&gt;limit = q-&gt;inilim = limit;
	q-&gt;kick = kick;
	q-&gt;arg = arg;
	q-&gt;state = msg;
	
	q-&gt;state |= Qstarve;
	q-&gt;eof = 0;
	q-&gt;noblock = 0;

	return q;
}

/* open a queue to be bypassed */
Queue*
qbypass(void (*bypass)(void*, Block*), void *arg)
{
	Queue *q;

	q = malloc(sizeof(Queue));
	if(q == 0)
		return 0;

	q-&gt;limit = 0;
	q-&gt;arg = arg;
	q-&gt;bypass = bypass;
	q-&gt;state = 0;

	return q;
}

static int
notempty(void *a)
{
	Queue *q = a;

	return (q-&gt;state &amp; Qclosed) || q-&gt;bfirst != 0;
}

/*
 *  wait for the queue to be non-empty or closed.
 *  called with q ilocked.
 */
static int
qwait(Queue *q)
{
	/* wait for data */
	for(;;){
		if(q-&gt;bfirst != nil)
			break;

		if(q-&gt;state &amp; Qclosed){
			if(++q-&gt;eof &gt; 3)
				return -1;
			if(*q-&gt;err &amp;&amp; strcmp(q-&gt;err, Ehungup) != 0)
				return -1;
			return 0;
		}

		q-&gt;state |= Qstarve;	/* flag requesting producer to wake me */
		iunlock(&amp;q-&gt;lk);
		sleep(&amp;q-&gt;rr, notempty, q);
		ilock(&amp;q-&gt;lk);
	}
	return 1;
}

/*
 * add a block list to a queue
 */
void
qaddlist(Queue *q, Block *b)
{
	/* queue the block */
	if(q-&gt;bfirst)
		q-&gt;blast-&gt;next = b;
	else
		q-&gt;bfirst = b;
	q-&gt;len += blockalloclen(b);
	q-&gt;dlen += blocklen(b);
	while(b-&gt;next)
		b = b-&gt;next;
	q-&gt;blast = b;
}

/*
 *  called with q ilocked
 */
Block*
qremove(Queue *q)
{
	Block *b;

	b = q-&gt;bfirst;
	if(b == nil)
		return nil;
	q-&gt;bfirst = b-&gt;next;
	b-&gt;next = nil;
	q-&gt;dlen -= BLEN(b);
	q-&gt;len -= BALLOC(b);
	QDEBUG checkb(b, "qremove");
	return b;
}

/*
 *  copy the contents of a string of blocks into
 *  memory.  emptied blocks are freed.  return
 *  pointer to first unconsumed block.
 */
Block*
bl2mem(uchar *p, Block *b, int n)
{
	int i;
	Block *next;

	for(; b != nil; b = next){
		i = BLEN(b);
		if(i &gt; n){
			memmove(p, b-&gt;rp, n);
			b-&gt;rp += n;
			return b;
		}
		memmove(p, b-&gt;rp, i);
		n -= i;
		p += i;
		b-&gt;rp += i;
		next = b-&gt;next;
		freeb(b);
	}
	return nil;
}

/*
 *  copy the contents of memory into a string of blocks.
 *  return nil on error.
 */
Block*
mem2bl(uchar *p, int len)
{
	int n;
	Block *b, *first, **l;

	first = nil;
	l = &amp;first;
	if(waserror()){
		freeblist(first);
		nexterror();
	}
	do {
		n = len;
		if(n &gt; Maxatomic)
			n = Maxatomic;

		*l = b = allocb(n);
	/*	setmalloctag(b, (up-&gt;text[0]&lt;&lt;24)|(up-&gt;text[1]&lt;&lt;16)|(up-&gt;text[2]&lt;&lt;8)|up-&gt;text[3]); */
		memmove(b-&gt;wp, p, n);
		b-&gt;wp += n;
		p += n;
		len -= n;
		l = &amp;b-&gt;next;
	} while(len &gt; 0);
	poperror();

	return first;
}

/*
 *  put a block back to the front of the queue
 *  called with q ilocked
 */
void
qputback(Queue *q, Block *b)
{
	b-&gt;next = q-&gt;bfirst;
	if(q-&gt;bfirst == nil)
		q-&gt;blast = b;
	q-&gt;bfirst = b;
	q-&gt;len += BALLOC(b);
	q-&gt;dlen += BLEN(b);
}

/*
 *  flow control, get producer going again
 *  called with q ilocked
 */
static void
qwakeup_iunlock(Queue *q)
{
	int dowakeup = 0;

	/* if writer flow controlled, restart */
	if((q-&gt;state &amp; Qflow) &amp;&amp; q-&gt;len &lt; q-&gt;limit/2){
		q-&gt;state &amp;= ~Qflow;
		dowakeup = 1;
	}

	iunlock(&amp;q-&gt;lk);

	/* wakeup flow controlled writers */
	if(dowakeup){
		if(q-&gt;kick)
			q-&gt;kick(q-&gt;arg);
		wakeup(&amp;q-&gt;wr);
	}
}

/*
 *  get next block from a queue (up to a limit)
 */
Block*
qbread(Queue *q, int len)
{
	Block *b, *nb;
	int n;

	qlock(&amp;q-&gt;rlock);
	if(waserror()){
		qunlock(&amp;q-&gt;rlock);
		nexterror();
	}

	ilock(&amp;q-&gt;lk);
	switch(qwait(q)){
	case 0:
		/* queue closed */
		iunlock(&amp;q-&gt;lk);
		qunlock(&amp;q-&gt;rlock);
		poperror();
		return nil;
	case -1:
		/* multiple reads on a closed queue */
		iunlock(&amp;q-&gt;lk);
		error(q-&gt;err);
	}

	/* if we get here, there's at least one block in the queue */
	b = qremove(q);
	n = BLEN(b);

	/* split block if it's too big and this is not a message queue */
	nb = b;
	if(n &gt; len){
		if((q-&gt;state&amp;Qmsg) == 0){
			n -= len;
			b = allocb(n);
			memmove(b-&gt;wp, nb-&gt;rp+len, n);
			b-&gt;wp += n;
			qputback(q, b);
		}
		nb-&gt;wp = nb-&gt;rp + len;
	}

	/* restart producer */
	qwakeup_iunlock(q);

	poperror();
	qunlock(&amp;q-&gt;rlock);
	return nb;
}

/*
 *  read a queue.  if no data is queued, post a Block
 *  and wait on its Rendez.
 */
long
qread(Queue *q, void *vp, int len)
{
	Block *b, *first, **l;
	int m, n;

	qlock(&amp;q-&gt;rlock);
	if(waserror()){
		qunlock(&amp;q-&gt;rlock);
		nexterror();
	}

	ilock(&amp;q-&gt;lk);
again:
	switch(qwait(q)){
	case 0:
		/* queue closed */
		iunlock(&amp;q-&gt;lk);
		qunlock(&amp;q-&gt;rlock);
		poperror();
		return 0;
	case -1:
		/* multiple reads on a closed queue */
		iunlock(&amp;q-&gt;lk);
		error(q-&gt;err);
	}

	/* if we get here, there's at least one block in the queue */
	if(q-&gt;state &amp; Qcoalesce){
		/* when coalescing, 0 length blocks just go away */
		b = q-&gt;bfirst;
		if(BLEN(b) &lt;= 0){
			freeb(qremove(q));
			goto again;
		}

		/*  grab the first block plus as many
		 *  following blocks as will completely
		 *  fit in the read.
		 */
		n = 0;
		l = &amp;first;
		m = BLEN(b);
		for(;;) {
			*l = qremove(q);
			l = &amp;b-&gt;next;
			n += m;

			b = q-&gt;bfirst;
			if(b == nil)
				break;
			m = BLEN(b);
			if(n+m &gt; len)
				break;
		}
	} else {
		first = qremove(q);
		n = BLEN(first);
	}

	/* copy to user space outside of the ilock */
	iunlock(&amp;q-&gt;lk);
	b = bl2mem(vp, first, len);
	ilock(&amp;q-&gt;lk);

	/* take care of any left over partial block */
	if(b != nil){
		n -= BLEN(b);
		if(q-&gt;state &amp; Qmsg)
			freeb(b);
		else
			qputback(q, b);
	}

	/* restart producer */
	qwakeup_iunlock(q);

	poperror();
	qunlock(&amp;q-&gt;rlock);
	return n;
}

static int
qnotfull(void *a)
{
	Queue *q = a;

	return q-&gt;len &lt; q-&gt;limit || (q-&gt;state &amp; Qclosed);
}

ulong noblockcnt;

/*
 *  add a block to a queue obeying flow control
 */
long
qbwrite(Queue *q, Block *b)
{
	int n, dowakeup;

	n = BLEN(b);

	if(q-&gt;bypass){
		(*q-&gt;bypass)(q-&gt;arg, b);
		return n;
	}

	dowakeup = 0;
	qlock(&amp;q-&gt;wlock);
	if(waserror()){
		if(b != nil)
			freeb(b);
		qunlock(&amp;q-&gt;wlock);
		nexterror();
	}

	ilock(&amp;q-&gt;lk);

	/* give up if the queue is closed */
	if(q-&gt;state &amp; Qclosed){
		iunlock(&amp;q-&gt;lk);
		error(q-&gt;err);
	}

	/* if nonblocking, don't queue over the limit */
	if(q-&gt;len &gt;= q-&gt;limit){
		if(q-&gt;noblock){
			iunlock(&amp;q-&gt;lk);
			freeb(b);
			noblockcnt += n;
			qunlock(&amp;q-&gt;wlock);
			poperror();
			return n;
		}
	}

	/* queue the block */
	if(q-&gt;bfirst)
		q-&gt;blast-&gt;next = b;
	else
		q-&gt;bfirst = b;
	q-&gt;blast = b;
	b-&gt;next = 0;
	q-&gt;len += BALLOC(b);
	q-&gt;dlen += n;
	QDEBUG checkb(b, "qbwrite");
	b = nil;

	/* make sure other end gets awakened */
	if(q-&gt;state &amp; Qstarve){
		q-&gt;state &amp;= ~Qstarve;
		dowakeup = 1;
	}
	iunlock(&amp;q-&gt;lk);

	/*  get output going again */
	if(q-&gt;kick &amp;&amp; (dowakeup || (q-&gt;state&amp;Qkick)))
		q-&gt;kick(q-&gt;arg);

	/* wakeup anyone consuming at the other end */
	if(dowakeup){
		wakeup(&amp;q-&gt;rr);

		/* if we just wokeup a higher priority process, let it run */
	/*
		p = wakeup(&amp;q-&gt;rr);
		if(p != nil &amp;&amp; p-&gt;priority &gt; up-&gt;priority)
			sched();
	 */
	}

	/*
	 *  flow control, wait for queue to get below the limit
	 *  before allowing the process to continue and queue
	 *  more.  We do this here so that postnote can only
	 *  interrupt us after the data has been queued.  This
	 *  means that things like 9p flushes and ssl messages
	 *  will not be disrupted by software interrupts.
	 *
	 *  Note - this is moderately dangerous since a process
	 *  that keeps getting interrupted and rewriting will
	 *  queue infinite crud.
	 */
	for(;;){
		if(q-&gt;noblock || qnotfull(q))
			break;

		ilock(&amp;q-&gt;lk);
		q-&gt;state |= Qflow;
		iunlock(&amp;q-&gt;lk);
		sleep(&amp;q-&gt;wr, qnotfull, q);
	}
	USED(b);

	qunlock(&amp;q-&gt;wlock);
	poperror();
	return n;
}

/*
 *  write to a queue.  only Maxatomic bytes at a time is atomic.
 */
int
qwrite(Queue *q, void *vp, int len)
{
	int n, sofar;
	Block *b;
	uchar *p = vp;

	QDEBUG if(!islo())
		print("qwrite hi %p\n", getcallerpc(&amp;q));

	sofar = 0;
	do {
		n = len-sofar;
		if(n &gt; Maxatomic)
			n = Maxatomic;

		b = allocb(n);
	/*	setmalloctag(b, (up-&gt;text[0]&lt;&lt;24)|(up-&gt;text[1]&lt;&lt;16)|(up-&gt;text[2]&lt;&lt;8)|up-&gt;text[3]); */
		if(waserror()){
			freeb(b);
			nexterror();
		}
		memmove(b-&gt;wp, p+sofar, n);
		poperror();
		b-&gt;wp += n;

		qbwrite(q, b);

		sofar += n;
	} while(sofar &lt; len &amp;&amp; (q-&gt;state &amp; Qmsg) == 0);

	return len;
}

/*
 *  used by print() to write to a queue.  Since we may be splhi or not in
 *  a process, don't qlock.
 */
int
qiwrite(Queue *q, void *vp, int len)
{
	int n, sofar, dowakeup;
	Block *b;
	uchar *p = vp;

	dowakeup = 0;

	sofar = 0;
	do {
		n = len-sofar;
		if(n &gt; Maxatomic)
			n = Maxatomic;

		b = iallocb(n);
		if(b == nil)
			break;
		memmove(b-&gt;wp, p+sofar, n);
		b-&gt;wp += n;

		ilock(&amp;q-&gt;lk);

		QDEBUG checkb(b, "qiwrite");
		if(q-&gt;bfirst)
			q-&gt;blast-&gt;next = b;
		else
			q-&gt;bfirst = b;
		q-&gt;blast = b;
		q-&gt;len += BALLOC(b);
		q-&gt;dlen += n;

		if(q-&gt;state &amp; Qstarve){
			q-&gt;state &amp;= ~Qstarve;
			dowakeup = 1;
		}

		iunlock(&amp;q-&gt;lk);

		if(dowakeup){
			if(q-&gt;kick)
				q-&gt;kick(q-&gt;arg);
			wakeup(&amp;q-&gt;rr);
		}

		sofar += n;
	} while(sofar &lt; len &amp;&amp; (q-&gt;state &amp; Qmsg) == 0);

	return sofar;
}

/*
 *  be extremely careful when calling this,
 *  as there is no reference accounting
 */
void
qfree(Queue *q)
{
	qclose(q);
	free(q);
}

/*
 *  Mark a queue as closed.  No further IO is permitted.
 *  All blocks are released.
 */
void
qclose(Queue *q)
{
	Block *bfirst;

	if(q == nil)
		return;

	/* mark it */
	ilock(&amp;q-&gt;lk);
	q-&gt;state |= Qclosed;
	q-&gt;state &amp;= ~(Qflow|Qstarve);
	strcpy(q-&gt;err, Ehungup);
	bfirst = q-&gt;bfirst;
	q-&gt;bfirst = 0;
	q-&gt;len = 0;
	q-&gt;dlen = 0;
	q-&gt;noblock = 0;
	iunlock(&amp;q-&gt;lk);

	/* free queued blocks */
	freeblist(bfirst);

	/* wake up readers/writers */
	wakeup(&amp;q-&gt;rr);
	wakeup(&amp;q-&gt;wr);
}

/*
 *  Mark a queue as closed.  Wakeup any readers.  Don't remove queued
 *  blocks.
 */
void
qhangup(Queue *q, char *msg)
{
	/* mark it */
	ilock(&amp;q-&gt;lk);
	q-&gt;state |= Qclosed;
	if(msg == 0 || *msg == 0)
		strcpy(q-&gt;err, Ehungup);
	else
		strncpy(q-&gt;err, msg, ERRMAX-1);
	iunlock(&amp;q-&gt;lk);

	/* wake up readers/writers */
	wakeup(&amp;q-&gt;rr);
	wakeup(&amp;q-&gt;wr);
}

/*
 *  return non-zero if the q is hungup
 */
int
qisclosed(Queue *q)
{
	return q-&gt;state &amp; Qclosed;
}

/*
 *  mark a queue as no longer hung up
 */
void
qreopen(Queue *q)
{
	ilock(&amp;q-&gt;lk);
	q-&gt;state &amp;= ~Qclosed;
	q-&gt;state |= Qstarve;
	q-&gt;eof = 0;
	q-&gt;limit = q-&gt;inilim;
	iunlock(&amp;q-&gt;lk);
}

/*
 *  return bytes queued
 */
int
qlen(Queue *q)
{
	return q-&gt;dlen;
}

/*
 * return space remaining before flow control
 */
int
qwindow(Queue *q)
{
	int l;

	l = q-&gt;limit - q-&gt;len;
	if(l &lt; 0)
		l = 0;
	return l;
}

/*
 *  return true if we can read without blocking
 */
int
qcanread(Queue *q)
{
	return q-&gt;bfirst!=0;
}

/*
 *  change queue limit
 */
void
qsetlimit(Queue *q, int limit)
{
	q-&gt;limit = limit;
}

/*
 *  set blocking/nonblocking
 */
void
qnoblock(Queue *q, int onoff)
{
	q-&gt;noblock = onoff;
}

/*
 *  flush the output queue
 */
void
qflush(Queue *q)
{
	Block *bfirst;

	/* mark it */
	ilock(&amp;q-&gt;lk);
	bfirst = q-&gt;bfirst;
	q-&gt;bfirst = 0;
	q-&gt;len = 0;
	q-&gt;dlen = 0;
	iunlock(&amp;q-&gt;lk);

	/* free queued blocks */
	freeblist(bfirst);

	/* wake up readers/writers */
	wakeup(&amp;q-&gt;wr);
}

int
qfull(Queue *q)
{
	return q-&gt;state &amp; Qflow;
}

int
qstate(Queue *q)
{
	return q-&gt;state;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

