<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/vac/file.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/vac/file.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include "stdinc.h"
#include "vac.h"
#include "dat.h"
#include "fns.h"
#include "error.h"

#define debug 0

/*
 * Vac file system.  This is a simplified version of the same code in Fossil.
 * 
 * The locking order in the tree is upward: a thread can hold the lock
 * for a VacFile and then acquire the lock of f-&gt;up (the parent),
 * but not vice-versa.
 * 
 * A vac file is one or two venti files.  Plain data files are one venti file,
 * while directores are two: a venti data file containing traditional
 * directory entries, and a venti directory file containing venti 
 * directory entries.  The traditional directory entries in the data file
 * contain integers indexing into the venti directory entry file.
 * It's a little complicated, but it makes the data usable by standard
 * tools like venti/copy.
 *
 */
 
static int filemetaflush(VacFile*, char*);

struct VacFile
{
	VacFs	*fs;	/* immutable */

	/* meta data for file: protected by the lk in the parent */
	int		ref;	/* holds this data structure up */

	int		partial;	/* file was never really open */
	int		removed;	/* file has been removed */
	int		dirty;	/* dir is dirty with respect to meta data in block */
	u32int	boff;		/* block offset within msource for this file's metadata */
	VacDir	dir;		/* metadata for this file */
	VacFile	*up;		/* parent file */
	VacFile	*next;	/* sibling */

	RWLock	lk;		/* lock for the following */
	VtFile	*source;	/* actual data */
	VtFile	*msource;	/* metadata for children in a directory */
	VacFile	*down;	/* children */
	int		mode;
	
	uvlong	qidoffset;	/* qid offset */
};

static VacFile*
filealloc(VacFs *fs)
{
	VacFile *f;

	f = vtmallocz(sizeof(VacFile));
	f-&gt;ref = 1;
	f-&gt;fs = fs;
	f-&gt;boff = NilBlock;
	f-&gt;mode = fs-&gt;mode;
	return f;
}

static void
filefree(VacFile *f)
{
	vtfileclose(f-&gt;source);
	vtfileclose(f-&gt;msource);
	vdcleanup(&amp;f-&gt;dir);
	memset(f, ~0, sizeof *f);	/* paranoia */
	vtfree(f);
}

static int
chksource(VacFile *f)
{
	if(f-&gt;partial)
		return 0;

	if(f-&gt;source == nil
	|| ((f-&gt;dir.mode &amp; ModeDir) &amp;&amp; f-&gt;msource == nil)){
		werrstr(ERemoved);
		return -1;
	}
	return 0;
}

static int
filelock(VacFile *f)
{
	wlock(&amp;f-&gt;lk);
	if(chksource(f) &lt; 0){
		wunlock(&amp;f-&gt;lk);
		return -1;
	}
	return 0;
}

static void
fileunlock(VacFile *f)
{
	wunlock(&amp;f-&gt;lk);
}

static int
filerlock(VacFile *f)
{
	rlock(&amp;f-&gt;lk);
	if(chksource(f) &lt; 0){
		runlock(&amp;f-&gt;lk);
		return -1;
	}
	return 0;
}

static void
filerunlock(VacFile *f)
{
	runlock(&amp;f-&gt;lk);
}

/*
 * The file metadata, like f-&gt;dir and f-&gt;ref,
 * are synchronized via the parent's lock.
 * This is why locking order goes up.
 */
static void
filemetalock(VacFile *f)
{
	assert(f-&gt;up != nil);
	wlock(&amp;f-&gt;up-&gt;lk);
}

static void
filemetaunlock(VacFile *f)
{
	wunlock(&amp;f-&gt;up-&gt;lk);
}

uvlong
vacfilegetid(VacFile *f)
{
	/* immutable */
	return f-&gt;qidoffset + f-&gt;dir.qid;
}

uvlong
vacfilegetqidoffset(VacFile *f)
{
	return f-&gt;qidoffset;
}

ulong
vacfilegetmcount(VacFile *f)
{
	ulong mcount;

	filemetalock(f);
	mcount = f-&gt;dir.mcount;
	filemetaunlock(f);
	return mcount;
}

ulong
vacfilegetmode(VacFile *f)
{
	ulong mode;

	filemetalock(f);
	mode = f-&gt;dir.mode;
	filemetaunlock(f);
	return mode;
}

int
vacfileisdir(VacFile *f)
{
	/* immutable */
	return (f-&gt;dir.mode &amp; ModeDir) != 0;
}

int
vacfileisroot(VacFile *f)
{
	return f == f-&gt;fs-&gt;root;
}

/*
 * The files are reference counted, and while the reference
 * is bigger than zero, each file can be found in its parent's
 * f-&gt;down list (chains via f-&gt;next), so that multiple threads
 * end up sharing a VacFile* when referring to the same file.
 *
 * Each VacFile holds a reference to its parent.
 */
VacFile*
vacfileincref(VacFile *vf)
{
	filemetalock(vf);
	assert(vf-&gt;ref &gt; 0);
	vf-&gt;ref++;
	filemetaunlock(vf);
	return vf;
}

int
vacfiledecref(VacFile *f)
{
	VacFile *p, *q, **qq;

	if(f-&gt;up == nil){
		/* never linked in */
		assert(f-&gt;ref == 1);
		filefree(f);
		return 0;
	}
	
	filemetalock(f);
	f-&gt;ref--;
	if(f-&gt;ref &gt; 0){
		filemetaunlock(f);
		return -1;
	}
	assert(f-&gt;ref == 0);
	assert(f-&gt;down == nil);

	if(f-&gt;source &amp;&amp; vtfilelock(f-&gt;source, -1) &gt;= 0){
		vtfileflush(f-&gt;source);
		vtfileunlock(f-&gt;source);
	}
	if(f-&gt;msource &amp;&amp; vtfilelock(f-&gt;msource, -1) &gt;= 0){
		vtfileflush(f-&gt;msource);
		vtfileunlock(f-&gt;msource);
	}

	/*
	 * Flush f's directory information to the cache.
	 */
	filemetaflush(f, nil);

	p = f-&gt;up;
	qq = &amp;p-&gt;down;
	for(q = *qq; q; q = *qq){
		if(q == f)
			break;
		qq = &amp;q-&gt;next;
	}
	assert(q != nil);
	*qq = f-&gt;next;

	filemetaunlock(f);
	filefree(f);
	vacfiledecref(p);
	return 0;
}


/* 
 * Construct a vacfile for the root of a vac tree, given the 
 * venti file for the root information.  That venti file is a 
 * directory file containing VtEntries for three more venti files:
 * the two venti files making up the root directory, and a 
 * third venti file that would be the metadata half of the 
 * "root's parent".
 *
 * Fossil generates slightly different vac files, due to a now
 * impossible-to-change bug, which contain a VtEntry
 * for just one venti file, that itself contains the expected
 * three directory entries.  Sigh.
 */
VacFile*
_vacfileroot(VacFs *fs, VtFile *r)
{
	int redirected;
	char err[ERRMAX];	
	VtBlock *b;
	VtFile *r0, *r1, *r2;
	MetaBlock mb;
	MetaEntry me;
	VacFile *root, *mr;

	redirected = 0;
Top:
	b = nil;
	root = nil;
	mr = nil;
	r1 = nil;
	r2 = nil;

	if(vtfilelock(r, -1) &lt; 0)
		return nil;
	r0 = vtfileopen(r, 0, fs-&gt;mode);
	if(debug)
		fprint(2, "r0 %p\n", r0);
	if(r0 == nil)
		goto Err;
	r2 = vtfileopen(r, 2, fs-&gt;mode);
	if(debug)
		fprint(2, "r2 %p\n", r2);
	if(r2 == nil){
		/*
		 * some vac files (e.g., from fossil)
		 * have an extra layer of indirection.
		 */
		rerrstr(err, sizeof err);
		if(!redirected &amp;&amp; strstr(err, "not active")){
			redirected = 1;
			vtfileunlock(r);
			r = r0;
			goto Top;
		}
		goto Err;
	}
	r1 = vtfileopen(r, 1, fs-&gt;mode);
	if(debug)
		fprint(2, "r1 %p\n", r1);
	if(r1 == nil)
		goto Err;

	mr = filealloc(fs);
	mr-&gt;msource = r2;
	r2 = nil;

	root = filealloc(fs);
	root-&gt;boff = 0;
	root-&gt;up = mr;
	root-&gt;source = r0;
	r0 = nil;
	root-&gt;msource = r1;
	r1 = nil;

	mr-&gt;down = root;
	vtfileunlock(r);

	if(vtfilelock(mr-&gt;msource, VtOREAD) &lt; 0)
		goto Err1;
	b = vtfileblock(mr-&gt;msource, 0, VtOREAD);
	vtfileunlock(mr-&gt;msource);
	if(b == nil)
		goto Err1;

	if(mbunpack(&amp;mb, b-&gt;data, mr-&gt;msource-&gt;dsize) &lt; 0)
		goto Err1;

	meunpack(&amp;me, &amp;mb, 0);
	if(vdunpack(&amp;root-&gt;dir, &amp;me) &lt; 0)
		goto Err1;
	vtblockput(b);

	return root;
Err:
	vtfileunlock(r);
Err1:
	vtblockput(b);
	if(r0)
		vtfileclose(r0);
	if(r1)
		vtfileclose(r1);
	if(r2)
		vtfileclose(r2);
	if(mr)
		filefree(mr);
	if(root)
		filefree(root);

	return nil;
}

/*
 * Vac directories are a sequence of metablocks, each of which
 * contains a bunch of metaentries sorted by file name.
 * The whole sequence isn't sorted, though, so you still have
 * to look at every block to find a given name.
 * Dirlookup looks in f for an element name elem.
 * It returns a new VacFile with the dir, boff, and mode
 * filled in, but the sources (venti files) are not, and f is 
 * not yet linked into the tree.  These details must be taken
 * care of by the caller.
 *
 * f must be locked, f-&gt;msource must not.
 */
static VacFile*
dirlookup(VacFile *f, char *elem)
{
	int i;
	MetaBlock mb;
	MetaEntry me;
	VtBlock *b;
	VtFile *meta;
	VacFile *ff;
	u32int bo, nb;

	meta = f-&gt;msource;
	b = nil;
	if(vtfilelock(meta, -1) &lt; 0)
		return nil;
	nb = (vtfilegetsize(meta)+meta-&gt;dsize-1)/meta-&gt;dsize;
	for(bo=0; bo&lt;nb; bo++){
		b = vtfileblock(meta, bo, VtOREAD);
		if(b == nil)
			goto Err;
		if(mbunpack(&amp;mb, b-&gt;data, meta-&gt;dsize) &lt; 0)
			goto Err;
		if(mbsearch(&amp;mb, elem, &amp;i, &amp;me) &gt;= 0){
			ff = filealloc(f-&gt;fs);
			if(vdunpack(&amp;ff-&gt;dir, &amp;me) &lt; 0){
				filefree(ff);
				goto Err;
			}
			ff-&gt;qidoffset = f-&gt;qidoffset + ff-&gt;dir.qidoffset;
			vtfileunlock(meta);
			vtblockput(b);
			ff-&gt;boff = bo;
			ff-&gt;mode = f-&gt;mode;
			return ff;
		}
		vtblockput(b);
		b = nil;
	}
	werrstr(ENoFile);
	/* fall through */
Err:
	vtfileunlock(meta);
	vtblockput(b);
	return nil;
}

/*
 * Open the venti file at offset in the directory f-&gt;source.
 * f is locked.
 */
static VtFile *
fileopensource(VacFile *f, u32int offset, u32int gen, int dir, uint mode)
{
	VtFile *r;

	if((r = vtfileopen(f-&gt;source, offset, mode)) == nil)
		return nil;
	if(r == nil)
		return nil;
	if(r-&gt;gen != gen){
		werrstr(ERemoved);
		vtfileclose(r);
		return nil;
	}
	if(r-&gt;dir != dir &amp;&amp; r-&gt;mode != -1){
		werrstr(EBadMeta);
		vtfileclose(r);
		return nil;
	}
	return r;
}

VacFile*
vacfilegetparent(VacFile *f)
{
	if(vacfileisroot(f))
		return vacfileincref(f);
	return vacfileincref(f-&gt;up);
}

/*
 * Given an unlocked vacfile (directory) f,
 * return the vacfile named elem in f.
 * Interprets . and .. as a convenience to callers.
 */
VacFile*
vacfilewalk(VacFile *f, char *elem)
{
	VacFile *ff;

	if(elem[0] == 0){
		werrstr(EBadPath);
		return nil;
	}

	if(!vacfileisdir(f)){
		werrstr(ENotDir);
		return nil;
	}

	if(strcmp(elem, ".") == 0)
		return vacfileincref(f);

	if(strcmp(elem, "..") == 0)
		return vacfilegetparent(f);

	if(filelock(f) &lt; 0)
		return nil;

	for(ff = f-&gt;down; ff; ff=ff-&gt;next){
		if(strcmp(elem, ff-&gt;dir.elem) == 0 &amp;&amp; !ff-&gt;removed){
			ff-&gt;ref++;
			goto Exit;
		}
	}

	ff = dirlookup(f, elem);
	if(ff == nil)
		goto Err;

	if(ff-&gt;dir.mode &amp; ModeSnapshot)
		ff-&gt;mode = VtOREAD;

	if(vtfilelock(f-&gt;source, f-&gt;mode) &lt; 0)
		goto Err;
	if(ff-&gt;dir.mode &amp; ModeDir){
		ff-&gt;source = fileopensource(f, ff-&gt;dir.entry, ff-&gt;dir.gen, 1, ff-&gt;mode);
		ff-&gt;msource = fileopensource(f, ff-&gt;dir.mentry, ff-&gt;dir.mgen, 0, ff-&gt;mode);
		if(ff-&gt;source == nil || ff-&gt;msource == nil)
			goto Err1;
	}else{
		ff-&gt;source = fileopensource(f, ff-&gt;dir.entry, ff-&gt;dir.gen, 0, ff-&gt;mode);
		if(ff-&gt;source == nil)
			goto Err1;
	}
	vtfileunlock(f-&gt;source);

	/* link in and up parent ref count */
	ff-&gt;next = f-&gt;down;
	f-&gt;down = ff;
	ff-&gt;up = f;
	vacfileincref(f);
Exit:
	fileunlock(f);
	return ff;

Err1:
	vtfileunlock(f-&gt;source);
Err:
	fileunlock(f);
	if(ff != nil)
		vacfiledecref(ff);
	return nil;
}

/* 
 * Open a path in the vac file system: 
 * just walk each element one at a time.
 */
VacFile*
vacfileopen(VacFs *fs, char *path)
{
	VacFile *f, *ff;
	char *p, elem[VtMaxStringSize], *opath;
	int n;

	f = fs-&gt;root;
	vacfileincref(f);
	opath = path;
	while(*path != 0){
		for(p = path; *p &amp;&amp; *p != '/'; p++)
			;
		n = p - path;
		if(n &gt; 0){
			if(n &gt; VtMaxStringSize){
				werrstr("%s: element too long", EBadPath);
				goto Err;
			}
			memmove(elem, path, n);
			elem[n] = 0;
			ff = vacfilewalk(f, elem);
			if(ff == nil){
				werrstr("%.*s: %r", utfnlen(opath, p-opath), opath);
				goto Err;
			}
			vacfiledecref(f);
			f = ff;
		}
		if(*p == '/')
			p++;
		path = p;
	}
	return f;
Err:
	vacfiledecref(f);
	return nil;
}

/*
 * Extract the score for the bn'th block in f.
 */
int
vacfileblockscore(VacFile *f, u32int bn, u8int *score)
{
	VtFile *s;
	uvlong size;
	int dsize, ret;

	ret = -1;
	if(filerlock(f) &lt; 0)
		return -1;
	if(vtfilelock(f-&gt;source, VtOREAD) &lt; 0)
		goto out;

	s = f-&gt;source;
	dsize = s-&gt;dsize;
	size = vtfilegetsize(s);
	if((uvlong)bn*dsize &gt;= size)
		goto out1;
	ret = vtfileblockscore(f-&gt;source, bn, score);

out1:
	vtfileunlock(f-&gt;source);
out:
	filerunlock(f);
	return ret;
}

/*
 * Read data from f.
 */
int
vacfileread(VacFile *f, void *buf, int cnt, vlong offset)
{
	int n;

	if(offset &lt; 0){
		werrstr(EBadOffset);
		return -1;
	}
	if(filerlock(f) &lt; 0)
		return -1;
	if(vtfilelock(f-&gt;source, VtOREAD) &lt; 0){
		filerunlock(f);
		return -1;
	}
	n = vtfileread(f-&gt;source, buf, cnt, offset);
	vtfileunlock(f-&gt;source);
	filerunlock(f);
	return n;
}

static int
getentry(VtFile *f, VtEntry *e)
{
	if(vtfilelock(f, VtOREAD) &lt; 0)
		return -1;
	if(vtfilegetentry(f, e) &lt; 0){
		vtfileunlock(f);
		return -1;
	}
	vtfileunlock(f);
	if(vtglobaltolocal(e-&gt;score) != NilBlock){
		werrstr("internal error - data not on venti");
		return -1;
	}
	return 0;
}

/*
 * Get the VtEntries for the data contained in f.
 */
int
vacfilegetentries(VacFile *f, VtEntry *e, VtEntry *me)
{
	if(filerlock(f) &lt; 0)
		return -1;
	if(e &amp;&amp; getentry(f-&gt;source, e) &lt; 0){
		filerunlock(f);
		return -1;
	}
	if(me){
		if(f-&gt;msource == nil)
			memset(me, 0, sizeof *me);
		else if(getentry(f-&gt;msource, me) &lt; 0){
			filerunlock(f);
			return -1;
		}
	}
	filerunlock(f);
	return 0;
}

/*
 * Get the file's size.
 */
int
vacfilegetsize(VacFile *f, uvlong *size)
{
	if(filerlock(f) &lt; 0)
		return -1;
	if(vtfilelock(f-&gt;source, VtOREAD) &lt; 0){
		filerunlock(f);
		return -1;
	}
	*size = vtfilegetsize(f-&gt;source);
	vtfileunlock(f-&gt;source);
	filerunlock(f);

	return 0;
}

/*
 * Directory reading.
 *
 * A VacDirEnum is a buffer containing directory entries.
 * Directory entries contain malloced strings and need to 
 * be cleaned up with vdcleanup.  The invariant in the 
 * VacDirEnum is that the directory entries between
 * vde-&gt;i and vde-&gt;n are owned by the vde and need to
 * be cleaned up if it is closed.  Those from 0 up to vde-&gt;i
 * have been handed to the reader, and the reader must 
 * take care of calling vdcleanup as appropriate.
 */
VacDirEnum*
vdeopen(VacFile *f)
{
	VacDirEnum *vde;
	VacFile *p;

	if(!vacfileisdir(f)){
		werrstr(ENotDir);
		return nil;
	}

	/*
	 * There might be changes to this directory's children
	 * that have not been flushed out into the cache yet.
	 * Those changes are only available if we look at the 
	 * VacFile structures directory.  But the directory reader
	 * is going to read the cache blocks directly, so update them.
	 */
	if(filelock(f) &lt; 0)
		return nil;
	for(p=f-&gt;down; p; p=p-&gt;next)
		filemetaflush(p, nil);
	fileunlock(f);

	vde = vtmallocz(sizeof(VacDirEnum));
	vde-&gt;file = vacfileincref(f);

	return vde;
}

/*
 * Figure out the size of the directory entry at offset.
 * The rest of the metadata is kept in the data half,
 * but since venti has to track the data size anyway,
 * we just use that one and avoid updating the directory
 * each time the file size changes.
 */
static int
direntrysize(VtFile *s, ulong offset, ulong gen, uvlong *size)
{
	VtBlock *b;
	ulong bn;
	VtEntry e;
	int epb;

	epb = s-&gt;dsize/VtEntrySize;
	bn = offset/epb;
	offset -= bn*epb;

	b = vtfileblock(s, bn, VtOREAD);
	if(b == nil)
		goto Err;
	if(vtentryunpack(&amp;e, b-&gt;data, offset) &lt; 0)
		goto Err;

	/* dangling entries are returned as zero size */
	if(!(e.flags &amp; VtEntryActive) || e.gen != gen)
		*size = 0;
	else
		*size = e.size;
	vtblockput(b);
	return 0;

Err:
	vtblockput(b);
	return -1;
}

/*
 * Fill in vde with a new batch of directory entries.
 */
static int
vdefill(VacDirEnum *vde)
{
	int i, n;
	VtFile *meta, *source;
	MetaBlock mb;
	MetaEntry me;
	VacFile *f;
	VtBlock *b;
	VacDir *de;

	/* clean up first */
	for(i=vde-&gt;i; i&lt;vde-&gt;n; i++)
		vdcleanup(vde-&gt;buf+i);
	vtfree(vde-&gt;buf);
	vde-&gt;buf = nil;
	vde-&gt;i = 0;
	vde-&gt;n = 0;

	f = vde-&gt;file;

	source = f-&gt;source;
	meta = f-&gt;msource;

	b = vtfileblock(meta, vde-&gt;boff, VtOREAD);
	if(b == nil)
		goto Err;
	if(mbunpack(&amp;mb, b-&gt;data, meta-&gt;dsize) &lt; 0)
		goto Err;

	n = mb.nindex;
	vde-&gt;buf = vtmalloc(n * sizeof(VacDir));

	for(i=0; i&lt;n; i++){
		de = vde-&gt;buf + i;
		meunpack(&amp;me, &amp;mb, i);
		if(vdunpack(de, &amp;me) &lt; 0)
			goto Err;
		vde-&gt;n++;
		if(!(de-&gt;mode &amp; ModeDir))
		if(direntrysize(source, de-&gt;entry, de-&gt;gen, &amp;de-&gt;size) &lt; 0)
			goto Err;
	}
	vde-&gt;boff++;
	vtblockput(b);
	return 0;
Err:
	vtblockput(b);
	return -1;
}

/*
 * Read a single directory entry from vde into de.
 * Returns -1 on error, 0 on EOF, and 1 on success.
 * When it returns 1, it becomes the caller's responsibility
 * to call vdcleanup(de) to free the strings contained
 * inside, or else to call vdunread to give it back.
 */
int
vderead(VacDirEnum *vde, VacDir *de)
{
	int ret;
	VacFile *f;
	u32int nb;

	f = vde-&gt;file;
	if(filerlock(f) &lt; 0)
		return -1;

	if(vtfilelock2(f-&gt;source, f-&gt;msource, VtOREAD) &lt; 0){
		filerunlock(f);
		return -1;
	}

	nb = (vtfilegetsize(f-&gt;msource)+f-&gt;msource-&gt;dsize-1)/f-&gt;msource-&gt;dsize;

	while(vde-&gt;i &gt;= vde-&gt;n){
		if(vde-&gt;boff &gt;= nb){
			ret = 0;
			goto Return;
		}
		if(vdefill(vde) &lt; 0){
			ret = -1;
			goto Return;
		}
	}

	memmove(de, vde-&gt;buf + vde-&gt;i, sizeof(VacDir));
	vde-&gt;i++;
	ret = 1;

Return:
	vtfileunlock(f-&gt;source);
	vtfileunlock(f-&gt;msource);
	filerunlock(f);

	return ret;
}

/*
 * "Unread" the last directory entry that was read,
 * so that the next vderead will return the same one.
 * If the caller calls vdeunread(vde) it should not call
 * vdcleanup on the entry being "unread".
 */
int
vdeunread(VacDirEnum *vde)
{
	if(vde-&gt;i &gt; 0){
		vde-&gt;i--;
		return 0;
	}
	return -1;
}

/*
 * Close the enumerator.
 */
void
vdeclose(VacDirEnum *vde)
{
	int i;
	if(vde == nil)
		return;
	/* free the strings */
	for(i=vde-&gt;i; i&lt;vde-&gt;n; i++)
		vdcleanup(vde-&gt;buf+i);
	vtfree(vde-&gt;buf);
	vacfiledecref(vde-&gt;file);
	vtfree(vde);
}


/*
 * On to mutation.  If the vac file system has been opened
 * read-write, then the files and directories can all be edited.
 * Changes are kept in the in-memory cache until flushed out
 * to venti, so we must be careful to explicitly flush data 
 * that we're not likely to modify again.
 *
 * Each VacFile has its own copy of its VacDir directory entry
 * in f-&gt;dir, but otherwise the cache is the authoratative source
 * for data.  Thus, for the most part, it suffices if we just 
 * call vtfileflushbefore and vtfileflush when we modify things.
 * There are a few places where we have to remember to write
 * changed VacDirs back into the cache.  If f-&gt;dir *is* out of sync,
 * then f-&gt;dirty should be set.
 *
 * The metadata in a directory is, to venti, a plain data file,
 * but as mentioned above it is actually a sequence of 
 * MetaBlocks that contain sorted lists of VacDir entries.
 * The filemetaxxx routines manipulate that stream.
 */

/*
 * Find space in fp for the directory entry dir (not yet written to disk)
 * and write it to disk, returning NilBlock on failure,
 * or the block number on success.
 *
 * Start is a suggested block number to try.
 * The caller must have filemetalock'ed f and have
 * vtfilelock'ed f-&gt;up-&gt;msource.
 */
static u32int
filemetaalloc(VacFile *fp, VacDir *dir, u32int start)
{
	u32int nb, bo;
	VtBlock *b;
	MetaBlock mb;
	int nn;
	uchar *p;
	int i, n;
	MetaEntry me;
	VtFile *ms;
	
	ms = fp-&gt;msource;
	n = vdsize(dir, VacDirVersion);
	
	/* Look for a block with room for a new entry of size n. */
	nb = (vtfilegetsize(ms)+ms-&gt;dsize-1)/ms-&gt;dsize;
	if(start == NilBlock){
		if(nb &gt; 0)
			start = nb - 1;
		else
			start = 0;
	}
	
	if(start &gt; nb)
		start = nb;
	for(bo=start; bo&lt;nb; bo++){
		if((b = vtfileblock(ms, bo, VtOREAD)) == nil)
			goto Err;
		if(mbunpack(&amp;mb, b-&gt;data, ms-&gt;dsize) &lt; 0)
			goto Err;
		nn = (mb.maxsize*FullPercentage/100) - mb.size + mb.free;
		if(n &lt;= nn &amp;&amp; mb.nindex &lt; mb.maxindex){
			/* reopen for writing */
			vtblockput(b);
			if((b = vtfileblock(ms, bo, VtORDWR)) == nil)
				goto Err;
			mbunpack(&amp;mb, b-&gt;data, ms-&gt;dsize);
			goto Found;
		}
		vtblockput(b);
	}

	/* No block found, extend the file by one metablock. */
	vtfileflushbefore(ms, nb*(uvlong)ms-&gt;dsize);
	if((b = vtfileblock(ms, nb, VtORDWR)) == nil)
		goto Err;
	vtfilesetsize(ms, (nb+1)*ms-&gt;dsize);
	mbinit(&amp;mb, b-&gt;data, ms-&gt;dsize, ms-&gt;dsize/BytesPerEntry);

Found:
	/* Now we have a block; allocate space to write the entry. */
	p = mballoc(&amp;mb, n);
	if(p == nil){
		/* mballoc might have changed block */
		mbpack(&amp;mb);
		werrstr(EBadMeta);
		goto Err;
	}

	/* Figure out where to put the index entry, and write it. */
	mbsearch(&amp;mb, dir-&gt;elem, &amp;i, &amp;me);
	assert(me.p == nil);	/* not already there */
	me.p = p;
	me.size = n;
	vdpack(dir, &amp;me, VacDirVersion);
	mbinsert(&amp;mb, i, &amp;me);
	mbpack(&amp;mb);
	vtblockput(b);
	return bo;

Err:
	vtblockput(b);
	return NilBlock;
}

/*
 * Update f's directory entry in the block cache. 
 * We look for the directory entry by name;
 * if we're trying to rename the file, oelem is the old name.
 *
 * Assumes caller has filemetalock'ed f.
 */
static int
filemetaflush(VacFile *f, char *oelem)
{
	int i, n;
	MetaBlock mb;
	MetaEntry me, me2;
	VacFile *fp;
	VtBlock *b;
	u32int bo;

	if(!f-&gt;dirty)
		return 0;

	if(oelem == nil)
		oelem = f-&gt;dir.elem;

	/*
	 * Locate f's old metadata in the parent's metadata file.
	 * We know which block it was in, but not exactly where
	 * in the block.
	 */
	fp = f-&gt;up;
	if(vtfilelock(fp-&gt;msource, -1) &lt; 0)
		return -1;
	/* can happen if source is clri'ed out from under us */
	if(f-&gt;boff == NilBlock)
		goto Err1;
	b = vtfileblock(fp-&gt;msource, f-&gt;boff, VtORDWR);
	if(b == nil)
		goto Err1;
	if(mbunpack(&amp;mb, b-&gt;data, fp-&gt;msource-&gt;dsize) &lt; 0)
		goto Err;
	if(mbsearch(&amp;mb, oelem, &amp;i, &amp;me) &lt; 0)
		goto Err;

	/*
	 * Check whether we can resize the entry and keep it 
	 * in this block.
	 */
	n = vdsize(&amp;f-&gt;dir, VacDirVersion);
	if(mbresize(&amp;mb, &amp;me, n) &gt;= 0){
		/* Okay, can be done without moving to another block. */

		/* Remove old data */
		mbdelete(&amp;mb, i, &amp;me);

		/* Find new location if renaming */
		if(strcmp(f-&gt;dir.elem, oelem) != 0)
			mbsearch(&amp;mb, f-&gt;dir.elem, &amp;i, &amp;me2);

		/* Pack new data into new location. */
		vdpack(&amp;f-&gt;dir, &amp;me, VacDirVersion);
vdunpack(&amp;f-&gt;dir, &amp;me);
		mbinsert(&amp;mb, i, &amp;me);
		mbpack(&amp;mb);
		
		/* Done */
		vtblockput(b);
		vtfileunlock(fp-&gt;msource);
		f-&gt;dirty = 0;
		return 0;
	}
	
	/*
	 * The entry must be moved to another block.
	 * This can only really happen on renames that
	 * make the name very long.
	 */
	
	/* Allocate a spot in a new block. */
	if((bo = filemetaalloc(fp, &amp;f-&gt;dir, f-&gt;boff+1)) == NilBlock){
		/* mbresize above might have modified block */
		mbpack(&amp;mb);
		goto Err;
	}
	f-&gt;boff = bo;

	/* Now we're committed.  Delete entry in old block. */
	mbdelete(&amp;mb, i, &amp;me);
	mbpack(&amp;mb);
	vtblockput(b);
	vtfileunlock(fp-&gt;msource);

	f-&gt;dirty = 0;
	return 0;

Err:
	vtblockput(b);
Err1:
	vtfileunlock(fp-&gt;msource);
	return -1;
}

/*
 * Remove the directory entry for f.
 */
static int
filemetaremove(VacFile *f)
{
	VtBlock *b;
	MetaBlock mb;
	MetaEntry me;
	int i;
	VacFile *fp;

	b = nil;
	fp = f-&gt;up;
	filemetalock(f);

	if(vtfilelock(fp-&gt;msource, VtORDWR) &lt; 0)
		goto Err;
	b = vtfileblock(fp-&gt;msource, f-&gt;boff, VtORDWR);
	if(b == nil)
		goto Err;

	if(mbunpack(&amp;mb, b-&gt;data, fp-&gt;msource-&gt;dsize) &lt; 0)
		goto Err;
	if(mbsearch(&amp;mb, f-&gt;dir.elem, &amp;i, &amp;me) &lt; 0)
		goto Err;
	mbdelete(&amp;mb, i, &amp;me);
	mbpack(&amp;mb);
	vtblockput(b);
	vtfileunlock(fp-&gt;msource);

	f-&gt;removed = 1;
	f-&gt;boff = NilBlock;
	f-&gt;dirty = 0;

	filemetaunlock(f);
	return 0;

Err:
	vtfileunlock(fp-&gt;msource);
	vtblockput(b);
	filemetaunlock(f);
	return -1;
}

/*
 * That was far too much effort for directory entries.
 * Now we can write code that *does* things.
 */

/*
 * Flush all data associated with f out of the cache and onto venti.
 * If recursive is set, flush f's children too.
 * Vacfiledecref knows how to flush source and msource too.
 */
int
vacfileflush(VacFile *f, int recursive)
{
	int ret;
	VacFile **kids, *p;
	int i, nkids;
	
	if(f-&gt;mode == VtOREAD)
		return 0;

	ret = 0;
	filemetalock(f);
	if(filemetaflush(f, nil) &lt; 0)
		ret = -1;
	filemetaunlock(f);

	if(filelock(f) &lt; 0)
		return -1;

	/*
	 * Lock order prevents us from flushing kids while holding
	 * lock, so make a list and then flush without the lock.
	 */
	nkids = 0;
	kids = nil;
	if(recursive){
		nkids = 0;
		for(p=f-&gt;down; p; p=p-&gt;next)
			nkids++;
		kids = vtmalloc(nkids*sizeof(VacFile*));
		i = 0;
		for(p=f-&gt;down; p; p=p-&gt;next){
			kids[i++] = p;
			p-&gt;ref++;
		}
	}
	if(nkids &gt; 0){
		fileunlock(f);
		for(i=0; i&lt;nkids; i++){
			if(vacfileflush(kids[i], 1) &lt; 0)
				ret = -1;
			vacfiledecref(kids[i]);
		}
		filelock(f);
	}
	free(kids);

	/*
	 * Now we can flush our own data.
	 */	
	vtfilelock(f-&gt;source, -1);
	if(vtfileflush(f-&gt;source) &lt; 0)
		ret = -1;
	vtfileunlock(f-&gt;source);
	if(f-&gt;msource){
		vtfilelock(f-&gt;msource, -1);
		if(vtfileflush(f-&gt;msource) &lt; 0)
			ret = -1;
		vtfileunlock(f-&gt;msource);
	}
	fileunlock(f);

	return ret;
}
		
/*
 * Create a new file named elem in fp with the given mode.
 * The mode can be changed later except for the ModeDir bit.
 */
VacFile*
vacfilecreate(VacFile *fp, char *elem, ulong mode)
{
	VacFile *ff;
	VacDir *dir;
	VtFile *pr, *r, *mr;
	int type;
	u32int bo;

	if(filelock(fp) &lt; 0)
		return nil;

	/*
	 * First, look to see that there's not a file in memory
	 * with the same name.
	 */
	for(ff = fp-&gt;down; ff; ff=ff-&gt;next){
		if(strcmp(elem, ff-&gt;dir.elem) == 0 &amp;&amp; !ff-&gt;removed){
			ff = nil;
			werrstr(EExists);
			goto Err1;
		}
	}

	/*
	 * Next check the venti blocks.
	 */
	ff = dirlookup(fp, elem);
	if(ff != nil){
		werrstr(EExists);
		goto Err1;
	}

	/*
	 * By the way, you can't create in a read-only file system.
	 */
	pr = fp-&gt;source;
	if(pr-&gt;mode != VtORDWR){
		werrstr(EReadOnly);
		goto Err1;
	}

	/*
	 * Okay, time to actually create something.  Lock the two
	 * halves of the directory and create a file.
	 */
	if(vtfilelock2(fp-&gt;source, fp-&gt;msource, -1) &lt; 0)
		goto Err1;
	ff = filealloc(fp-&gt;fs);
	ff-&gt;qidoffset = fp-&gt;qidoffset;	/* hopefully fp-&gt;qidoffset == 0 */
	type = VtDataType;
	if(mode &amp; ModeDir)
		type = VtDirType;
	mr = nil;
	if((r = vtfilecreate(pr, pr-&gt;psize, pr-&gt;dsize, type)) == nil)
		goto Err;
	if(mode &amp; ModeDir)
	if((mr = vtfilecreate(pr, pr-&gt;psize, pr-&gt;dsize, VtDataType)) == nil)
		goto Err;

	/*
	 * Fill in the directory entry and write it to disk.
	 */
	dir = &amp;ff-&gt;dir;
	dir-&gt;elem = vtstrdup(elem);
	dir-&gt;entry = r-&gt;offset;
	dir-&gt;gen = r-&gt;gen;
	if(mode &amp; ModeDir){
		dir-&gt;mentry = mr-&gt;offset;
		dir-&gt;mgen = mr-&gt;gen;
	}
	dir-&gt;size = 0;
	if(_vacfsnextqid(fp-&gt;fs, &amp;dir-&gt;qid) &lt; 0)
		goto Err;
	dir-&gt;uid = vtstrdup(fp-&gt;dir.uid);
	dir-&gt;gid = vtstrdup(fp-&gt;dir.gid);
	dir-&gt;mid = vtstrdup("");
	dir-&gt;mtime = time(0L);
	dir-&gt;mcount = 0;
	dir-&gt;ctime = dir-&gt;mtime;
	dir-&gt;atime = dir-&gt;mtime;
	dir-&gt;mode = mode;
	if((bo = filemetaalloc(fp, &amp;ff-&gt;dir, NilBlock)) == NilBlock)
		goto Err;

	/*
	 * Now we're committed.
	 */
	vtfileunlock(fp-&gt;source);
	vtfileunlock(fp-&gt;msource);
	ff-&gt;source = r;
	ff-&gt;msource = mr;
	ff-&gt;boff = bo;

	/* Link into tree. */
	ff-&gt;next = fp-&gt;down;
	fp-&gt;down = ff;
	ff-&gt;up = fp;
	vacfileincref(fp);

	fileunlock(fp);
	
	filelock(ff);
	vtfilelock(ff-&gt;source, -1);
	vtfileunlock(ff-&gt;source);
	fileunlock(ff);

	return ff;

Err:
	vtfileunlock(fp-&gt;source);
	vtfileunlock(fp-&gt;msource);
	if(r){
		vtfilelock(r, -1);
		vtfileremove(r);
	}
	if(mr){
		vtfilelock(mr, -1);
		vtfileremove(mr);
	}
Err1:
	if(ff)
		vacfiledecref(ff);
	fileunlock(fp);
	return nil;
}

/*
 * Change the size of the file f.
 */
int
vacfilesetsize(VacFile *f, uvlong size)
{
	if(vacfileisdir(f)){
		werrstr(ENotFile);
		return -1;
	}
	
	if(filelock(f) &lt; 0)
		return -1;

	if(f-&gt;source-&gt;mode != VtORDWR){
		werrstr(EReadOnly);
		goto Err;
	}
	if(vtfilelock(f-&gt;source, -1) &lt; 0)
		goto Err;
	if(vtfilesetsize(f-&gt;source, size) &lt; 0){
		vtfileunlock(f-&gt;source);
		goto Err;
	}
	vtfileunlock(f-&gt;source);
	fileunlock(f);
	return 0;

Err:
	fileunlock(f);
	return -1;
}

/*
 * Write data to f.
 */
int
vacfilewrite(VacFile *f, void *buf, int cnt, vlong offset)
{
	if(vacfileisdir(f)){
		werrstr(ENotFile);
		return -1;
	}
	if(filelock(f) &lt; 0)
		return -1;
	if(f-&gt;source-&gt;mode != VtORDWR){
		werrstr(EReadOnly);
		goto Err;
	}
	if(offset &lt; 0){
		werrstr(EBadOffset);
		goto Err;
	}

	if(vtfilelock(f-&gt;source, -1) &lt; 0)
		goto Err;
	if(f-&gt;dir.mode &amp; ModeAppend)
		offset = vtfilegetsize(f-&gt;source);
	if(vtfilewrite(f-&gt;source, buf, cnt, offset) != cnt
	|| vtfileflushbefore(f-&gt;source, offset) &lt; 0){
		vtfileunlock(f-&gt;source);
		goto Err;
	}
	vtfileunlock(f-&gt;source);
	fileunlock(f);
	return cnt;

Err:
	fileunlock(f);
	return -1;
}

/*
 * Set (!) the VtEntry for the data contained in f.
 * This let's us efficiently copy data from one file to another.
 */
int
vacfilesetentries(VacFile *f, VtEntry *e, VtEntry *me)
{
	int ret;

	vacfileflush(f, 0);	/* flush blocks to venti, since we won't see them again */

	if(!(e-&gt;flags&amp;VtEntryActive)){
		werrstr("missing entry for source");
		return -1;
	}
	if(me &amp;&amp; !(me-&gt;flags&amp;VtEntryActive))
		me = nil;
	if(f-&gt;msource &amp;&amp; !me){
		werrstr("missing entry for msource");
		return -1;
	}
	if(me &amp;&amp; !f-&gt;msource){
		werrstr("no msource to set");
		return -1;
	}

	if(filelock(f) &lt; 0)
		return -1;
	if(f-&gt;source-&gt;mode != VtORDWR
	|| (f-&gt;msource &amp;&amp; f-&gt;msource-&gt;mode != VtORDWR)){
		werrstr(EReadOnly);
		fileunlock(f);
		return -1;
	}
	if(vtfilelock2(f-&gt;source, f-&gt;msource, -1) &lt; 0){
		fileunlock(f);
		return -1;
	}
	ret = 0;
	if(vtfilesetentry(f-&gt;source, e) &lt; 0)
		ret = -1;
	else if(me &amp;&amp; vtfilesetentry(f-&gt;msource, me) &lt; 0)
		ret = -1;

	vtfileunlock(f-&gt;source);
	if(f-&gt;msource)
		vtfileunlock(f-&gt;msource);
	fileunlock(f);
	return ret;
}

/*
 * Get the directory entry for f.
 */
int
vacfilegetdir(VacFile *f, VacDir *dir)
{
	if(filerlock(f) &lt; 0)
		return -1;

	filemetalock(f);
	vdcopy(dir, &amp;f-&gt;dir);
	filemetaunlock(f);

	if(!vacfileisdir(f)){
		if(vtfilelock(f-&gt;source, VtOREAD) &lt; 0){
			filerunlock(f);
			return -1;
		}
		dir-&gt;size = vtfilegetsize(f-&gt;source);
		vtfileunlock(f-&gt;source);
	}
	filerunlock(f);

	return 0;
}

/*
 * Set the directory entry for f.
 */
int
vacfilesetdir(VacFile *f, VacDir *dir)
{
	VacFile *ff;
	char *oelem;
	u32int mask;
	u64int size;

	/* can not set permissions for the root */
	if(vacfileisroot(f)){
		werrstr(ERoot);
		return -1;
	}

	if(filelock(f) &lt; 0)
		return -1;
	filemetalock(f);
	
	if(f-&gt;source-&gt;mode != VtORDWR){
		werrstr(EReadOnly);
		goto Err;
	}

	/* On rename, check new name does not already exist */
	if(strcmp(f-&gt;dir.elem, dir-&gt;elem) != 0){
		for(ff = f-&gt;up-&gt;down; ff; ff=ff-&gt;next){
			if(strcmp(dir-&gt;elem, ff-&gt;dir.elem) == 0 &amp;&amp; !ff-&gt;removed){
				werrstr(EExists);
				goto Err;
			}
		}
		ff = dirlookup(f-&gt;up, dir-&gt;elem);
		if(ff != nil){
			vacfiledecref(ff);
			werrstr(EExists);
			goto Err;
		}
		werrstr("");	/* "failed" dirlookup poisoned it */
	}

	/* Get ready... */
	if(vtfilelock2(f-&gt;source, f-&gt;msource, -1) &lt; 0)
		goto Err;
	if(!vacfileisdir(f)){
		size = vtfilegetsize(f-&gt;source);
		if(size != dir-&gt;size){
			if(vtfilesetsize(f-&gt;source, dir-&gt;size) &lt; 0){
				vtfileunlock(f-&gt;source);
				if(f-&gt;msource)
					vtfileunlock(f-&gt;msource);
				goto Err;
			}
		}
	}
	/* ... now commited to changing it. */
	vtfileunlock(f-&gt;source);
	if(f-&gt;msource)
		vtfileunlock(f-&gt;msource);

	oelem = nil;
	if(strcmp(f-&gt;dir.elem, dir-&gt;elem) != 0){
		oelem = f-&gt;dir.elem;
		f-&gt;dir.elem = vtstrdup(dir-&gt;elem);
	}

	if(strcmp(f-&gt;dir.uid, dir-&gt;uid) != 0){
		vtfree(f-&gt;dir.uid);
		f-&gt;dir.uid = vtstrdup(dir-&gt;uid);
	}

	if(strcmp(f-&gt;dir.gid, dir-&gt;gid) != 0){
		vtfree(f-&gt;dir.gid);
		f-&gt;dir.gid = vtstrdup(dir-&gt;gid);
	}

	if(strcmp(f-&gt;dir.mid, dir-&gt;mid) != 0){
		vtfree(f-&gt;dir.mid);
		f-&gt;dir.mid = vtstrdup(dir-&gt;mid);
	}

	f-&gt;dir.mtime = dir-&gt;mtime;
	f-&gt;dir.atime = dir-&gt;atime;

	mask = ~(ModeDir|ModeSnapshot);
	f-&gt;dir.mode &amp;= ~mask;
	f-&gt;dir.mode |= mask &amp; dir-&gt;mode;
	f-&gt;dirty = 1;

	if(filemetaflush(f, oelem) &lt; 0){
		vtfree(oelem);
		goto Err;	/* that sucks */
	}
	vtfree(oelem);

	filemetaunlock(f);
	fileunlock(f);
	return 0;

Err:
	filemetaunlock(f);
	fileunlock(f);
	return -1;
}

/*
 * Set the qid space.
 */
int
vacfilesetqidspace(VacFile *f, u64int offset, u64int max)
{
	int ret;

	if(filelock(f) &lt; 0)
		return -1;
	if(f-&gt;source-&gt;mode != VtORDWR){
		fileunlock(f);
		werrstr(EReadOnly);
		return -1;
	}
	filemetalock(f);
	f-&gt;dir.qidspace = 1;
	f-&gt;dir.qidoffset = offset;
	f-&gt;dir.qidmax = max;
	f-&gt;dirty = 1;
	ret = filemetaflush(f, nil);
	filemetaunlock(f);
	fileunlock(f);
	return ret;
}

/*
 * Check that the file is empty, returning 0 if it is.
 * Returns -1 on error (and not being empty is an error).
 */
static int
filecheckempty(VacFile *f)
{
	u32int i, n;
	VtBlock *b;
	MetaBlock mb;
	VtFile *r;

	r = f-&gt;msource;
	n = (vtfilegetsize(r)+r-&gt;dsize-1)/r-&gt;dsize;
	for(i=0; i&lt;n; i++){
		b = vtfileblock(r, i, VtOREAD);
		if(b == nil)
			return -1;
		if(mbunpack(&amp;mb, b-&gt;data, r-&gt;dsize) &lt; 0)
			goto Err;
		if(mb.nindex &gt; 0){
			werrstr(ENotEmpty);
			goto Err;
		}
		vtblockput(b);
	}
	return 0;

Err:
	vtblockput(b);
	return -1;
}

/*
 * Remove the vac file f.
 */
int
vacfileremove(VacFile *f)
{
	VacFile *ff;

	/* Cannot remove the root */
	if(vacfileisroot(f)){
		werrstr(ERoot);
		return -1;
	}

	if(filelock(f) &lt; 0)
		return -1;
	if(f-&gt;source-&gt;mode != VtORDWR){
		werrstr(EReadOnly);
		goto Err1;
	}
	if(vtfilelock2(f-&gt;source, f-&gt;msource, -1) &lt; 0)
		goto Err1;
	if(vacfileisdir(f) &amp;&amp; filecheckempty(f)&lt;0)
		goto Err;

	for(ff=f-&gt;down; ff; ff=ff-&gt;next)
		assert(ff-&gt;removed);

	vtfileremove(f-&gt;source);
	f-&gt;source = nil;
	if(f-&gt;msource){
		vtfileremove(f-&gt;msource);
		f-&gt;msource = nil;
	}
	fileunlock(f);

	if(filemetaremove(f) &lt; 0)
		return -1;
	return 0;

Err:
	vtfileunlock(f-&gt;source);
	if(f-&gt;msource)
		vtfileunlock(f-&gt;msource);
Err1:
	fileunlock(f);
	return -1;
}

/*
 * Vac file system format.
 */
static char EBadVacFormat[] = "bad format for vac file";

static VacFs *
vacfsalloc(VtConn *z, int bsize, int ncache, int mode)
{
	VacFs *fs;

	fs = vtmallocz(sizeof(VacFs));
	fs-&gt;z = z;
	fs-&gt;bsize = bsize;
	fs-&gt;mode = mode;
	fs-&gt;cache = vtcachealloc(z, bsize, ncache);
	return fs;
}

static int
readscore(int fd, uchar score[VtScoreSize])
{
	char buf[45], *pref;
	int n;

	n = readn(fd, buf, sizeof(buf)-1);
	if(n &lt; sizeof(buf)-1) {
		werrstr("short read");
		return -1;
	}
	buf[n] = 0;

	if(vtparsescore(buf, &amp;pref, score) &lt; 0){
		werrstr(EBadVacFormat);
		return -1;
	}
	if(pref==nil || strcmp(pref, "vac") != 0) {
		werrstr("not a vac file");
		return -1;
	}
	return 0;
}

VacFs*
vacfsopen(VtConn *z, char *file, int mode, int ncache)
{
	int fd;
	uchar score[VtScoreSize];
	char *prefix;
	
	if(vtparsescore(file, &amp;prefix, score) &gt;= 0){
		if(prefix == nil || strcmp(prefix, "vac") != 0){
			werrstr("not a vac file");
			return nil;
		}
	}else{
		fd = open(file, OREAD);
		if(fd &lt; 0)
			return nil;
		if(readscore(fd, score) &lt; 0){
			close(fd);
			return nil;
		}
		close(fd);
	}
	return vacfsopenscore(z, score, mode, ncache);
}

VacFs*
vacfsopenscore(VtConn *z, u8int *score, int mode, int ncache)
{
	VacFs *fs;
	int n;
	VtRoot rt;
	uchar buf[VtRootSize];
	VacFile *root;
	VtFile *r;
	VtEntry e;

	n = vtread(z, score, VtRootType, buf, VtRootSize);
	if(n &lt; 0)
		return nil;
	if(n != VtRootSize){
		werrstr("vtread on root too short");
		return nil;
	}

	if(vtrootunpack(&amp;rt, buf) &lt; 0)
		return nil;

	if(strcmp(rt.type, "vac") != 0) {
		werrstr("not a vac root");
		return nil;
	}

	fs = vacfsalloc(z, rt.blocksize, ncache, mode);
	memmove(fs-&gt;score, score, VtScoreSize);
	fs-&gt;mode = mode;

	memmove(e.score, rt.score, VtScoreSize);
	e.gen = 0;
	e.psize = rt.blocksize;
	e.dsize = rt.blocksize;
	e.type = VtDirType;
	e.flags = VtEntryActive;
	e.size = 3*VtEntrySize;

	root = nil;
	if((r = vtfileopenroot(fs-&gt;cache, &amp;e)) == nil)
		goto Err;
	if(debug)
		fprint(2, "r %p\n", r);
	root = _vacfileroot(fs, r);
	if(debug)
		fprint(2, "root %p\n", root);
	vtfileclose(r);
	if(root == nil)
		goto Err;
	fs-&gt;root = root;
	return fs;
Err:
	if(root)
		vacfiledecref(root);
	vacfsclose(fs);
	return nil;
}

int
vacfsmode(VacFs *fs)
{
	return fs-&gt;mode;
}

VacFile*
vacfsgetroot(VacFs *fs)
{
	return vacfileincref(fs-&gt;root);
}

int
vacfsgetblocksize(VacFs *fs)
{
	return fs-&gt;bsize;
}

int
vacfsgetscore(VacFs *fs, u8int *score)
{
	memmove(score, fs-&gt;score, VtScoreSize);
	return 0;
}

int
_vacfsnextqid(VacFs *fs, uvlong *qid)
{
	++fs-&gt;qid;
	*qid = fs-&gt;qid;
	return 0;
}

void
vacfsjumpqid(VacFs *fs, uvlong step)
{
	fs-&gt;qid += step;
}

/*
 * Set *maxqid to the maximum qid expected in this file system.
 * In newer vac archives, the maximum qid is stored in the
 * qidspace VacDir annotation.  In older vac archives, the root
 * got created last, so it had the maximum qid.
 */
int
vacfsgetmaxqid(VacFs *fs, uvlong *maxqid)
{
	VacDir vd;
	
	if(vacfilegetdir(fs-&gt;root, &amp;vd) &lt; 0)
		return -1;
	if(vd.qidspace)
		*maxqid = vd.qidmax;
	else
		*maxqid = vd.qid;
	vdcleanup(&amp;vd);
	return 0;
}


void
vacfsclose(VacFs *fs)
{
	if(fs-&gt;root)
		vacfiledecref(fs-&gt;root);
	fs-&gt;root = nil;
	vtcachefree(fs-&gt;cache);
	vtfree(fs);
}

/*
 * Create a fresh vac fs.
 */
VacFs *
vacfscreate(VtConn *z, int bsize, int ncache)
{
	VacFs *fs;
	VtFile *f;
	uchar buf[VtEntrySize], metascore[VtScoreSize];
	VtEntry e;
	VtBlock *b;
	MetaBlock mb;
	VacDir vd;
	MetaEntry me;
	int psize;
	int mbsize;
	
	if((fs = vacfsalloc(z, bsize, ncache, VtORDWR)) == nil)
		return nil;
	
	/*
	 * Fake up an empty vac fs.
	 */
	psize = bsize;
	f = vtfilecreateroot(fs-&gt;cache, psize, bsize, VtDirType);
	vtfilelock(f, VtORDWR);
	
	/* Metablocks can't be too big -- they have 16-bit offsets in them. */
	mbsize = bsize;
	if(mbsize &gt;= 56*1024)
		mbsize = 56*1024;

	/* Write metablock containing root directory VacDir. */
	b = vtcacheallocblock(fs-&gt;cache, VtDataType);
	mbinit(&amp;mb, b-&gt;data, mbsize, mbsize/BytesPerEntry);
	memset(&amp;vd, 0, sizeof vd);
	vd.elem = "/";
	vd.mode = 0777|ModeDir;
	vd.uid = "vac";
	vd.gid = "vac";
	vd.mid = "";
	me.size = vdsize(&amp;vd, VacDirVersion);
	me.p = mballoc(&amp;mb, me.size);
	vdpack(&amp;vd, &amp;me, VacDirVersion);
	mbinsert(&amp;mb, 0, &amp;me);
	mbpack(&amp;mb);
	vtblockwrite(b);
	memmove(metascore, b-&gt;score, VtScoreSize);
	vtblockput(b);
	
	/* First entry: empty venti directory stream. */
	memset(&amp;e, 0, sizeof e);
	e.flags = VtEntryActive;
	e.psize = psize;
	e.dsize = bsize;
	e.type = VtDirType;
	memmove(e.score, vtzeroscore, VtScoreSize);
	vtentrypack(&amp;e, buf, 0);
	vtfilewrite(f, buf, VtEntrySize, 0);
	
	/* Second entry: empty metadata stream. */
	e.type = VtDataType;
	e.dsize = mbsize;
	vtentrypack(&amp;e, buf, 0);
	vtfilewrite(f, buf, VtEntrySize, VtEntrySize);

	/* Third entry: metadata stream with root directory. */
	memmove(e.score, metascore, VtScoreSize);
	e.size = mbsize;
	vtentrypack(&amp;e, buf, 0);
	vtfilewrite(f, buf, VtEntrySize, VtEntrySize*2);

	vtfileflush(f);
	vtfileunlock(f);
	
	/* Now open it as a vac fs. */
	fs-&gt;root = _vacfileroot(fs, f);
	if(fs-&gt;root == nil){
		werrstr("vacfileroot: %r");
		vacfsclose(fs);
		return nil;
	}

	return fs;
}

int
vacfssync(VacFs *fs)
{
	uchar buf[1024];
	VtEntry e;
	VtFile *f;
	VtRoot root;

	/* Sync the entire vacfs to disk. */
	if(vacfileflush(fs-&gt;root, 1) &lt; 0)
		return -1;
	if(vtfilelock(fs-&gt;root-&gt;up-&gt;msource, -1) &lt; 0)
		return -1;
	if(vtfileflush(fs-&gt;root-&gt;up-&gt;msource) &lt; 0){
		vtfileunlock(fs-&gt;root-&gt;up-&gt;msource);
		return -1;
	}
	vtfileunlock(fs-&gt;root-&gt;up-&gt;msource);

	/* Prepare the dir stream for the root block. */
	if(getentry(fs-&gt;root-&gt;source, &amp;e) &lt; 0)
		return -1;
	vtentrypack(&amp;e, buf, 0);
	if(getentry(fs-&gt;root-&gt;msource, &amp;e) &lt; 0)
		return -1;
	vtentrypack(&amp;e, buf, 1);
	if(getentry(fs-&gt;root-&gt;up-&gt;msource, &amp;e) &lt; 0)
		return -1;
	vtentrypack(&amp;e, buf, 2);

	f = vtfilecreateroot(fs-&gt;cache, fs-&gt;bsize, fs-&gt;bsize, VtDirType);
	vtfilelock(f, VtORDWR);
	if(vtfilewrite(f, buf, 3*VtEntrySize, 0) &lt; 0
	|| vtfileflush(f) &lt; 0){
		vtfileunlock(f);
		vtfileclose(f);
		return -1;
	}
	vtfileunlock(f);
	if(getentry(f, &amp;e) &lt; 0){
		vtfileclose(f);
		return -1;
	}
	vtfileclose(f);

	/* Build a root block. */
	memset(&amp;root, 0, sizeof root);
	strcpy(root.type, "vac");
	strcpy(root.name, fs-&gt;name);
	memmove(root.score, e.score, VtScoreSize);
	root.blocksize = fs-&gt;bsize;
	memmove(root.prev, fs-&gt;score, VtScoreSize);
	vtrootpack(&amp;root, buf);
	if(vtwrite(fs-&gt;z, fs-&gt;score, VtRootType, buf, VtRootSize) &lt; 0){
		werrstr("writing root: %r");
		return -1;
	}
	if(vtsync(fs-&gt;z) &lt; 0)
		return -1;
	return 0;
}

int
vacfiledsize(VacFile *f)
{
	VtEntry e;

	if(vacfilegetentries(f,&amp;e,nil) &lt; 0)
		return -1;
	return e.dsize;
}

/*
 * Does block b of f have the same SHA1 hash as the n bytes at buf?
 */
int
sha1matches(VacFile *f, ulong b, uchar *buf, int n)
{
	uchar fscore[VtScoreSize];
	uchar bufscore[VtScoreSize];
	
	if(vacfileblockscore(f, b, fscore) &lt; 0)
		return 0;
	n = vtzerotruncate(VtDataType, buf, n);
	sha1(buf, n, bufscore, nil);
	if(memcmp(bufscore, fscore, VtScoreSize) == 0)
		return 1;
	return 0;
}

<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

