<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/venti/srv/lumpcache.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/venti/srv/lumpcache.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include "stdinc.h"
#include "dat.h"
#include "fns.h"

/* #define CHECK(x)	x */
#define CHECK(x)

typedef struct LumpCache	LumpCache;

enum
{
	HashLog		= 9,
	HashSize	= 1&lt;&lt;HashLog,
	HashMask	= HashSize - 1,
};

struct LumpCache
{
	QLock		lock;
	Rendez		full;
	Lump		*free;			/* list of available lumps */
	u32int		allowed;		/* total allowable space for packets */
	u32int		avail;			/* remaining space for packets */
	u32int		now;			/* ticks for usage timestamps */
	Lump		**heads;		/* hash table for finding address */
	int		nheap;			/* number of available victims */
	Lump		**heap;			/* heap for locating victims */
	int		nblocks;		/* number of blocks allocated */
	Lump		*blocks;		/* array of block descriptors */
};

static LumpCache	lumpcache;

static void	delheap(Lump *db);
static int	downheap(int i, Lump *b);
static void	fixheap(int i, Lump *b);
static int	upheap(int i, Lump *b);
static Lump	*bumplump(void);

void
initlumpcache(u32int size, u32int nblocks)
{
	Lump *last, *b;
	int i;

	lumpcache.full.l = &amp;lumpcache.lock;
	lumpcache.nblocks = nblocks;
	lumpcache.allowed = size;
	lumpcache.avail = size;
	lumpcache.heads = MKNZ(Lump*, HashSize);
	lumpcache.heap = MKNZ(Lump*, nblocks);
	lumpcache.blocks = MKNZ(Lump, nblocks);
	setstat(StatLcacheSize, lumpcache.nblocks);

	last = nil;
	for(i = 0; i &lt; nblocks; i++){
		b = &amp;lumpcache.blocks[i];
		b-&gt;type = TWID8;
		b-&gt;heap = TWID32;
		b-&gt;next = last;
		last = b;
	}
	lumpcache.free = last;
	lumpcache.nheap = 0;
}

Lump*
lookuplump(u8int *score, int type)
{
	uint ms;
	Lump *b;
	u32int h;

	ms = 0;
	trace(TraceLump, "lookuplump enter");
	
	h = hashbits(score, HashLog);

	/*
	 * look for the block in the cache
	 */
	qlock(&amp;lumpcache.lock);
	CHECK(checklumpcache());
again:
	for(b = lumpcache.heads[h]; b != nil; b = b-&gt;next){
		if(scorecmp(score, b-&gt;score)==0 &amp;&amp; type == b-&gt;type){
			addstat(StatLcacheHit, 1);
			trace(TraceLump, "lookuplump hit");
			goto found;
		}
	}

	trace(TraceLump, "lookuplump miss");

	/*
	 * missed: locate the block with the oldest second to last use.
	 * remove it from the heap, and fix up the heap.
	 */
	while(lumpcache.free == nil){
		trace(TraceLump, "lookuplump bump");
		CHECK(checklumpcache());
		if(bumplump() == nil){
			CHECK(checklumpcache());
			logerr(EAdmin, "all lump cache blocks in use");
			addstat(StatLcacheStall, 1);
			CHECK(checklumpcache());
			rsleep(&amp;lumpcache.full);
			CHECK(checklumpcache());
			addstat(StatLcacheStall, -1);
			goto again;
		}
		CHECK(checklumpcache());
	}

	/* start timer on cache miss to avoid system call on cache hit */
	ms = msec();

	addstat(StatLcacheMiss, 1);
	b = lumpcache.free;
	lumpcache.free = b-&gt;next;

	/*
	 * the new block has no last use, so assume it happens sometime in the middle
ZZZ this is not reasonable
	 */
	b-&gt;used = (b-&gt;used2 + lumpcache.now) / 2;

	/*
	 * rechain the block on the correct hash chain
	 */
	b-&gt;next = lumpcache.heads[h];
	lumpcache.heads[h] = b;
	if(b-&gt;next != nil)
		b-&gt;next-&gt;prev = b;
	b-&gt;prev = nil;

	scorecp(b-&gt;score, score);
	b-&gt;type = type;
	b-&gt;size = 0;
	b-&gt;data = nil;

found:
	b-&gt;ref++;
	b-&gt;used2 = b-&gt;used;
	b-&gt;used = lumpcache.now++;
	if(b-&gt;heap != TWID32)
		fixheap(b-&gt;heap, b);
	CHECK(checklumpcache());
	qunlock(&amp;lumpcache.lock);


	addstat(StatLumpStall, 1);
	qlock(&amp;b-&gt;lock);
	addstat(StatLumpStall, -1);

	trace(TraceLump, "lookuplump exit");
	addstat2(StatLcacheRead, 1, StatLcacheReadTime, ms ? msec()-ms : 0);
	return b;
}

void
insertlump(Lump *b, Packet *p)
{
	u32int size;

	/*
	 * look for the block in the cache
	 */
	trace(TraceLump, "insertlump enter");
	qlock(&amp;lumpcache.lock);
	CHECK(checklumpcache());
again:

	addstat(StatLcacheWrite, 1);

	/*
	 * missed: locate the block with the oldest second to last use.
	 * remove it from the heap, and fix up the heap.
	 */
	size = packetasize(p);
	while(lumpcache.avail &lt; size){
		trace(TraceLump, "insertlump bump");
		CHECK(checklumpcache());
		if(bumplump() == nil){
			logerr(EAdmin, "all lump cache blocks in use");
			addstat(StatLcacheStall, 1);
			CHECK(checklumpcache());
			rsleep(&amp;lumpcache.full);
			CHECK(checklumpcache());
			addstat(StatLcacheStall, -1);
			goto again;
		}
		CHECK(checklumpcache());
	}
	b-&gt;data = p;
	b-&gt;size = size;
	lumpcache.avail -= size;
	CHECK(checklumpcache());
	qunlock(&amp;lumpcache.lock);
	trace(TraceLump, "insertlump exit");
}

void
putlump(Lump *b)
{
	if(b == nil)
		return;

	trace(TraceLump, "putlump");
	qunlock(&amp;b-&gt;lock);
	qlock(&amp;lumpcache.lock);
	CHECK(checklumpcache());
	if(--b-&gt;ref == 0){
		if(b-&gt;heap == TWID32)
			upheap(lumpcache.nheap++, b);
		trace(TraceLump, "putlump wakeup");
		rwakeupall(&amp;lumpcache.full);
	}
	CHECK(checklumpcache());
	qunlock(&amp;lumpcache.lock);
}

/*
 * remove some lump from use and update the free list and counters
 */
static Lump*
bumplump(void)
{
	Lump *b;
	u32int h;

	/*
	 * remove blocks until we find one that is unused
	 * referenced blocks are left in the heap even though
	 * they can't be scavenged; this is simple a speed optimization
	 */
	CHECK(checklumpcache());
	for(;;){
		if(lumpcache.nheap == 0){
			trace(TraceLump, "bumplump emptyheap");
			return nil;
		}
		b = lumpcache.heap[0];
		delheap(b);
		if(!b-&gt;ref){
			trace(TraceLump, "bumplump wakeup");
			rwakeupall(&amp;lumpcache.full);
			break;
		}
	}

	/*
	 * unchain the block
	 */
	trace(TraceLump, "bumplump unchain");
	if(b-&gt;prev == nil){
		h = hashbits(b-&gt;score, HashLog);
		if(lumpcache.heads[h] != b)
			sysfatal("bad hash chains in lump cache");
		lumpcache.heads[h] = b-&gt;next;
	}else
		b-&gt;prev-&gt;next = b-&gt;next;
	if(b-&gt;next != nil)
		b-&gt;next-&gt;prev = b-&gt;prev;

	if(b-&gt;data != nil){
		packetfree(b-&gt;data);
		b-&gt;data = nil;
		lumpcache.avail += b-&gt;size;
		b-&gt;size = 0;
	}
	b-&gt;type = TWID8;

	b-&gt;next = lumpcache.free;
	lumpcache.free = b;

	CHECK(checklumpcache());
	trace(TraceLump, "bumplump exit");
	return b;
}

void
emptylumpcache(void)
{
	qlock(&amp;lumpcache.lock);
	while(bumplump())
		;
	qunlock(&amp;lumpcache.lock);
}

/*
 * delete an arbitrary block from the heap
 */
static void
delheap(Lump *db)
{
	fixheap(db-&gt;heap, lumpcache.heap[--lumpcache.nheap]);
	db-&gt;heap = TWID32;
}

/*
 * push an element up or down to it's correct new location
 */
static void
fixheap(int i, Lump *b)
{
	if(upheap(i, b) == i)
		downheap(i, b);
}

static int
upheap(int i, Lump *b)
{
	Lump *bb;
	u32int now;
	int p;

	now = lumpcache.now;
	for(; i != 0; i = p){
		p = (i - 1) &gt;&gt; 1;
		bb = lumpcache.heap[p];
		if(b-&gt;used2 - now &gt;= bb-&gt;used2 - now)
			break;
		lumpcache.heap[i] = bb;
		bb-&gt;heap = i;
	}

	lumpcache.heap[i] = b;
	b-&gt;heap = i;
	return i;
}

static int
downheap(int i, Lump *b)
{
	Lump *bb;
	u32int now;
	int k;

	now = lumpcache.now;
	for(; ; i = k){
		k = (i &lt;&lt; 1) + 1;
		if(k &gt;= lumpcache.nheap)
			break;
		if(k + 1 &lt; lumpcache.nheap &amp;&amp; lumpcache.heap[k]-&gt;used2 - now &gt; lumpcache.heap[k + 1]-&gt;used2 - now)
			k++;
		bb = lumpcache.heap[k];
		if(b-&gt;used2 - now &lt;= bb-&gt;used2 - now)
			break;
		lumpcache.heap[i] = bb;
		bb-&gt;heap = i;
	}

	lumpcache.heap[i] = b;
	b-&gt;heap = i;
	return i;
}

static void
findblock(Lump *bb)
{
	Lump *b, *last;
	int h;

	last = nil;
	h = hashbits(bb-&gt;score, HashLog);
	for(b = lumpcache.heads[h]; b != nil; b = b-&gt;next){
		if(last != b-&gt;prev)
			sysfatal("bad prev link");
		if(b == bb)
			return;
		last = b;
	}
	sysfatal("block score=%V type=%#x missing from hash table", bb-&gt;score, bb-&gt;type);
}

void
checklumpcache(void)
{
	Lump *b;
	u32int size, now, nfree;
	int i, k, refed;

	now = lumpcache.now;
	for(i = 0; i &lt; lumpcache.nheap; i++){
		if(lumpcache.heap[i]-&gt;heap != i)
			sysfatal("lc: mis-heaped at %d: %d", i, lumpcache.heap[i]-&gt;heap);
		if(i &gt; 0 &amp;&amp; lumpcache.heap[(i - 1) &gt;&gt; 1]-&gt;used2 - now &gt; lumpcache.heap[i]-&gt;used2 - now)
			sysfatal("lc: bad heap ordering");
		k = (i &lt;&lt; 1) + 1;
		if(k &lt; lumpcache.nheap &amp;&amp; lumpcache.heap[i]-&gt;used2 - now &gt; lumpcache.heap[k]-&gt;used2 - now)
			sysfatal("lc: bad heap ordering");
		k++;
		if(k &lt; lumpcache.nheap &amp;&amp; lumpcache.heap[i]-&gt;used2 - now &gt; lumpcache.heap[k]-&gt;used2 - now)
			sysfatal("lc: bad heap ordering");
	}

	refed = 0;
	size = 0;
	for(i = 0; i &lt; lumpcache.nblocks; i++){
		b = &amp;lumpcache.blocks[i];
		if(b-&gt;data == nil &amp;&amp; b-&gt;size != 0)
			sysfatal("bad size: %d data=%p", b-&gt;size, b-&gt;data);
		if(b-&gt;ref &amp;&amp; b-&gt;heap == TWID32)
			refed++;
		if(b-&gt;type != TWID8){
			findblock(b);
			size += b-&gt;size;
		}
		if(b-&gt;heap != TWID32
		&amp;&amp; lumpcache.heap[b-&gt;heap] != b)
			sysfatal("lc: spurious heap value");
	}
	if(lumpcache.avail != lumpcache.allowed - size){
		fprint(2, "mismatched available=%d and allowed=%d - used=%d space", lumpcache.avail, lumpcache.allowed, size);
		*(int*)0=0;
	}

	nfree = 0;
	for(b = lumpcache.free; b != nil; b = b-&gt;next){
		if(b-&gt;type != TWID8 || b-&gt;heap != TWID32)
			sysfatal("lc: bad free list");
		nfree++;
	}

	if(lumpcache.nheap + nfree + refed != lumpcache.nblocks)
		sysfatal("lc: missing blocks: %d %d %d %d", lumpcache.nheap, refed, nfree, lumpcache.nblocks);
}

<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

