<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/libcontrol/control.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/libcontrol/control.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;draw.h&gt;
#include &lt;thread.h&gt;
#include &lt;mouse.h&gt;
#include &lt;keyboard.h&gt;
#include &lt;control.h&gt;

static int debug = 0;

enum	/* alts */
{
	AKey,
	AMouse,
	ACtl,
	AExit,
	NALT
};

static Controlset **controlset;
int	ncontrolset;
int	ctldeletequits;

char *alignnames[Nalignments] = {
	[Aupperleft] =		"upperleft",
	[Auppercenter] =	"uppercenter",
	[Aupperright] =		"upperright",
	[Acenterleft] =		"centerleft",
	[Acenter] =		"center",
	[Acenterright] =	"centerright",
	[Alowerleft] =		"lowerleft",
	[Alowercenter] =	"lowercenter",
	[Alowerright] =		"lowerright",
};

char *ctltypenames[Ntypes] = {
	[Ctlunknown] =		"unknown",
	[Ctlbox] =			"box",
	[Ctlbutton] =		"button",
	[Ctlentry] =		"entry",
	[Ctlkeyboard] =		"keyboard",
	[Ctllabel] =		"label",
	[Ctlmenu] =		"menu",
	[Ctlradio] =		"radio",
	[Ctlscribble] =		"scribble",
	[Ctlslider] =		"slider",
	[Ctltabs] =			"tabs",
	[Ctltext] =			"text",
	[Ctltextbutton] =	"textbutton",
	[Ctltextbutton3] =	"textbutton3",
	[Ctlgroup] =		"group",		// divider between controls and metacontrols
	[Ctlboxbox] =		"boxbox",
	[Ctlcolumn] =		"column",
	[Ctlrow] =			"row",
	[Ctlstack] =		"stack",
	[Ctltab] =			"tab",
};

static void	_ctlcmd(Controlset*, char*);
static void	_ctlcontrol(Controlset*, char*);

static char*
_mkctlcmd(Control *c, char *fmt, va_list arg)
{
	char *name, *p, *both;

	name = quotestrdup(c-&gt;name);
	if(name == nil)
		ctlerror("quotestrdup in ctlprint failed");
	p = vsmprint(fmt, arg);
	if(p == nil){
		free(name);
		ctlerror("vsmprint1 in ctlprint failed");
	}
	both = ctlmalloc(strlen(name)+strlen(p)+2);
	strcpy(both, name);
	strcat(both, " ");
	strcat(both, p);
	free(name);
	free(p);
	return both;
}

int
ctlprint(Control *c, char *fmt, ...)
{
	int n;
	char *p;
	va_list arg;

	va_start(arg, fmt);
	p = _mkctlcmd(c, fmt, arg);
	va_end(arg);
	n = sendp(c-&gt;controlset-&gt;ctl, p);
	yield();
	return n;
}

void
_ctlprint(Control *c, char *fmt, ...)
{
	char *p;
	va_list arg;

	va_start(arg, fmt);
	p = _mkctlcmd(c, fmt, arg);
	va_end(arg);
	_ctlcmd(c-&gt;controlset, p);
	free(p);
}

int
_ctllookup(char *s, char *tab[], int ntab)
{
	int i;

	for(i=0; i&lt;ntab; i++)
		if(tab[i] != nil &amp;&amp; strcmp(s, tab[i]) == 0)
			return i;
	return -1;
}

static Control*
_newcontrol(Controlset *cs, uint n, char *name, char *type)
{
	Control *c;

	for(c=cs-&gt;controls; c; c=c-&gt;next)
		if(strcmp(c-&gt;name, name) == 0){
			werrstr("control %q already defined", name);
			return nil;
		}
	c = ctlmalloc(n);
	c-&gt;screen = cs-&gt;screen;
	c-&gt;name = ctlstrdup(name);
	c-&gt;type = _ctllookup(type, ctltypenames, Ntypes);
	if (c-&gt;type &lt; 0)
		ctlerror("unknown type: %s", type);
	c-&gt;event = chancreate(sizeof(char*), 64);
	c-&gt;data = chancreate(sizeof(char*), 0);
	c-&gt;size = Rect(1, 1, _Ctlmaxsize, _Ctlmaxsize);
	c-&gt;hidden = 0;
	c-&gt;ctl = nil;
	c-&gt;mouse = nil;
	c-&gt;key = nil;
	c-&gt;exit = nil;
	c-&gt;setsize = nil;

	c-&gt;controlset = cs;
	c-&gt;next = cs-&gt;controls;
	cs-&gt;controls = c;
	return c;
}

static void
controlsetthread(void *v)
{
	Controlset *cs;
	Mouse mouse;
	Control *f;
	int prevbut, n, i;
	Alt alts[NALT+1];
	char tmp[64], *str;
	Rune buf[2][20], *rp;

	cs = v;
	snprint(tmp, sizeof tmp, "controlsetthread 0x%p", cs);
	threadsetname(tmp);

	alts[AKey].c = cs-&gt;kbdc;
	alts[AKey].v = &amp;rp;
	alts[AKey].op = CHANRCV;
	alts[AMouse].c = cs-&gt;mousec;
	alts[AMouse].v = &amp;mouse;
	alts[AMouse].op = CHANRCV;
	alts[ACtl].c = cs-&gt;ctl;
	alts[ACtl].v = &amp;str;
	alts[ACtl].op = CHANRCV;
	alts[AExit].c = cs-&gt;csexitc;
	alts[AExit].v = nil;
	alts[AExit].op = CHANRCV;
	alts[NALT].op = CHANEND;

	cs-&gt;focus = nil;
	prevbut=0;
	n = 0;
	for(;;){
		/* toggle so we can receive in one buffer while client processes the other */
		alts[AKey].v = buf[n];
		rp = buf[n];
		n = 1-n;
		switch(alt(alts)){
		case AKey:
			if(ctldeletequits &amp;&amp; rp[0]=='\177')
				ctlerror("delete");
			for(i=1; i&lt;nelem(buf[0])-1; i++)
				if(nbrecv(cs-&gt;kbdc, rp+i) &lt;= 0)
					break;
			rp[i] = L'\0';
			if(cs-&gt;focus &amp;&amp; cs-&gt;focus-&gt;key)
				cs-&gt;focus-&gt;key(cs-&gt;focus, rp);
			break;
		case AMouse:
			/* is this a focus change? */
			if(prevbut)	/* don't change focus if button was down */
				goto Send;
			if(cs-&gt;focus!=nil &amp;&amp; cs-&gt;focus-&gt;hidden == 0 &amp;&amp; ptinrect(mouse.xy, cs-&gt;focus-&gt;rect))
				goto Send;
			if(cs-&gt;clicktotype == 0)
				goto Change;
			/* click to type: only change if button is down */
			if(mouse.buttons == 0)
				goto Send;
		Change:
			/* change of focus */
			if(cs-&gt;focus != nil)
				_ctlprint(cs-&gt;focus, "focus 0");
			cs-&gt;focus = nil;
			for(f=cs-&gt;actives; f!=nil; f=f-&gt;nextactive)
				if(f-&gt;hidden == 0 &amp;&amp; f-&gt;mouse &amp;&amp; ptinrect(mouse.xy, f-&gt;rect)){
					cs-&gt;focus = f;
					_ctlprint(f, "focus 1");
					if (f-&gt;mouse) {
						if (debug) fprint(2, "f-&gt;mouse %s\n", f-&gt;name);
						f-&gt;mouse(f, &amp;mouse);
					}
					break;
				}
		Send:
			if(cs-&gt;focus &amp;&amp; cs-&gt;focus-&gt;mouse) {
				if (debug) fprint(2, "cs-&gt;focus-&gt;mouse %s\n", cs-&gt;focus-&gt;name);
				cs-&gt;focus-&gt;mouse(cs-&gt;focus, &amp;mouse);
			}
			prevbut=mouse.buttons;
			break;
		case ACtl:
			_ctlcontrol(cs, str);
			free(str);
			break;
		case AExit:
			threadexits(nil);
		}
	}
}

Control*
_createctl(Controlset *cs, char *type, uint size, char *name)
{
	Control *c;

	c = _newcontrol(cs, size, name, type);
	if(c == nil)
		ctlerror("can't create %s control %q: %r", type, name);
	return c;
}

void
closecontrol(Control *c)
{
	Control *prev, *p;

	if(c == nil)
		return;
	if (c == c-&gt;controlset-&gt;focus)
		c-&gt;controlset-&gt;focus = nil;
	if(c-&gt;exit)
		c-&gt;exit(c);

	prev = nil;
	for(p=c-&gt;controlset-&gt;controls; p; p=p-&gt;next){
		if(p == c)
			break;
		prev = p;
	}
	if(p == nil)
		ctlerror("closecontrol: no such control %q %p\n", c-&gt;name, c);
	if(prev == nil)
		c-&gt;controlset-&gt;controls = c-&gt;next;
	else
		prev-&gt;next = c-&gt;next;

	/* is it active? if so, delete from active list */
	prev = nil;
	for(p=c-&gt;controlset-&gt;actives; p; p=p-&gt;nextactive){
		if(p == c)
			break;
		prev = p;
	}
	if(p != nil){
		if(prev == nil)
			c-&gt;controlset-&gt;actives = c-&gt;nextactive;
		else
			prev-&gt;nextactive = c-&gt;nextactive;
	}

	if(!c-&gt;wevent)
		chanfree(c-&gt;event);
	if(!c-&gt;wdata)
		chanfree(c-&gt;data);
	free(c-&gt;name);
	free(c-&gt;format);
	free(c);
}

Control*
controlcalled(char *name)
{
	Control *c;
	int i;

	for(i=0; i&lt;ncontrolset; i++)
		for(c=controlset[i]-&gt;controls; c; c=c-&gt;next)
			if(strcmp(c-&gt;name, name) == 0)
				return c;
	return nil;
}

void
ctlerror(char *fmt, ...)
{
	va_list arg;
	char buf[256];

	va_start(arg, fmt);
	vfprint(2, fmt, arg);
	va_end(arg);
	write(2, "\n", 1);
	threadexitsall(buf);
}

Rune*
_ctlrunestr(char *s)
{
	Rune *r, *ret;

	ret = r = ctlmalloc((utflen(s)+1)*sizeof(Rune));
	while(*s != '\0')
		s += chartorune(r++, s);
	*r = L'\0';
	return ret;
}

char*
_ctlstrrune(Rune *r)
{
	char *s;
	s = ctlmalloc(runestrlen(r)*UTFmax+1);
	sprint(s, "%S", r);
	return s;
}

void*
ctlmalloc(uint n)
{
	void *p;

	p = mallocz(n, 1);
	if(p == nil)
		ctlerror("control allocation failed: %r");
	return p;
}

void*
ctlrealloc(void *p, uint n)
{
	p = realloc(p, n);
	if(p == nil)
		ctlerror("control reallocation failed: %r");
	return p;
}

char*
ctlstrdup(char *s)
{
	char *t;

	t = strdup(s);
	if(t == nil)
		ctlerror("control strdup(%q) failed: %r", s);
	return t;
}

static void
ctokenize(char *s, CParse *cp)
{
	snprint(cp-&gt;str, sizeof cp-&gt;str, "%s", s);
	cp-&gt;args = cp-&gt;pargs;
	cp-&gt;nargs = tokenize(s, cp-&gt;args, nelem(cp-&gt;pargs));
}

static int
ctlparse(CParse *cp, char *s, int hasreceiver)
{
	int i;
	char *t;

	/* keep original string for good error messages */
	strncpy(cp-&gt;str, s, sizeof cp-&gt;str);
	cp-&gt;str[sizeof cp-&gt;str - 1] = '\0';
	ctokenize(s, cp);
	if(cp-&gt;nargs == 0)
		return -1;
	/* strip leading sender name if present */
	cp-&gt;sender = nil;
	i = strlen(cp-&gt;args[0])-1;
	if(cp-&gt;args[0][i] == ':'){
		cp-&gt;sender = cp-&gt;args[0];
		cp-&gt;sender[i] = '\0';
		cp-&gt;args++;
		cp-&gt;nargs--;
	}
	if(hasreceiver){
		if(cp-&gt;nargs-- == 0)
			return -1;
		cp-&gt;receiver = *cp-&gt;args++;
	}else
		cp-&gt;receiver = nil;
	for(i=0; i&lt;cp-&gt;nargs; i++){
		t = cp-&gt;args[i];
		while(*t == '[')	/* %R gives [0 0] [1 1]; atoi will stop at closing ] */
			t++;
		cp-&gt;iargs[i] = atoi(t);
	}
	return cp-&gt;nargs;
}

void
_ctlargcount(Control *c, CParse *cp, int n)
{
	if(cp-&gt;nargs != n)
		ctlerror("%q: wrong argument count in '%s'", c-&gt;name, cp-&gt;str);
}

static void
_ctlcmd(Controlset *cs, char*s)
{
	CParse cp;
	char	*rcvrs[32];
	int	ircvrs[32], n, i, hit;
	Control *c;

//	fprint(2, "_ctlcmd: %s\n", s);
	cp.args = cp.pargs;
	if (ctlparse(&amp;cp, s, 1) &lt; 0)
		ctlerror("bad command string: %q", cp.str);
	if (cp.nargs == 0 &amp;&amp; strcmp(cp.receiver, "sync") == 0){
		chanprint(cs-&gt;data, "sync");
		return;
	}
	if (cp.nargs == 0)
		ctlerror("no command in command string: %q", cp.str);

	n = tokenize(cp.receiver, rcvrs, nelem(rcvrs));

	// lookup type names: a receiver can be a named type or a named control
	for (i = 0; i &lt; n; i++)
		ircvrs[i] = _ctllookup(rcvrs[i], ctltypenames, Ntypes);

	for(c = cs-&gt;controls; c != nil; c = c-&gt;next){
		/* if a control matches on more than one receiver element,
		 * make sure it gets processed once; hence loop through controls
		 * in the outer loop
		 */
		hit = 0;
		for (i = 0; i &lt; n; i++)
			if(strcmp(c-&gt;name, rcvrs[i]) == 0 || c-&gt;type == ircvrs[i])
				hit++;
		if (hit &amp;&amp; c-&gt;ctl)
			c-&gt;ctl(c, &amp;cp);
	}
}

static void
_ctlcontrol(Controlset *cs, char *s)
{
	char *lines[16];
	int i, n;
	char *l;

//	fprint(2, "_ctlcontrol: %s\n", s);
	n = gettokens(s, lines, nelem(lines), "\n");
	for(i=0; i&lt;n; i++){
		l = lines[i];
		while(*l==' ' || *l=='\t')
			l++;
		if(*l != '\0')
			_ctlcmd(cs, l);
	}
}

Rune*
_ctlgetsnarf(void)
{
	int i, n;
	char *sn, buf[512];
	Rune *snarf;

	if(_ctlsnarffd &lt; 0)
		return nil;
	sn = nil;
	i = 0;
	seek(_ctlsnarffd, 0, 0);
	while((n = read(_ctlsnarffd, buf, sizeof buf)) &gt; 0){
		sn = ctlrealloc(sn, i+n+1);
		memmove(sn+i, buf, n);
		i += n;
		sn[i] = 0;
	}
	snarf = nil;
	if(i &gt; 0){
		snarf = _ctlrunestr(sn);
		free(sn);
	}
	return snarf;
}

void
_ctlputsnarf(Rune *snarf)
{
	int fd, i, n, nsnarf;

	if(_ctlsnarffd&lt;0 || snarf[0]==0)
		return;
	fd = open("/dev/snarf", OWRITE);
	if(fd &lt; 0)
		return;
	nsnarf = runestrlen(snarf);
	/* snarf buffer could be huge, so fprint will truncate; do it in blocks */
	for(i=0; i&lt;nsnarf; i+=n){
		n = nsnarf-i;
		if(n &gt;= 256)
			n = 256;
		if(fprint(fd, "%.*S", n, snarf+i) &lt; 0)
			break;
	}
	close(fd);
}

int
_ctlalignment(char *s)
{
	int i;

	i = _ctllookup(s, alignnames, Nalignments);
	if (i &lt; 0)
		ctlerror("unknown alignment: %s", s);
	return i;
}

Point
_ctlalignpoint(Rectangle r, int dx, int dy, int align)
{
	Point p;

	p = r.min;	/* in case of trouble */
	switch(align%3){
	case 0:	/* left */
		p.x = r.min.x;
		break;
	case 1:	/* center */
		p.x = r.min.x+(Dx(r)-dx)/2;
		break;
	case 2:	/* right */
		p.x = r.max.x-dx;
		break;
	}
	switch((align/3)%3){
	case 0:	/* top */
		p.y = r.min.y;
		break;
	case 1:	/* center */
		p.y = r.min.y+(Dy(r)-dy)/2;
		break;
	case 2:	/* bottom */
		p.y = r.max.y - dy;
		break;
	}
	return p;
}

void
controlwire(Control *cfrom, char *name, Channel *chan)
{
	Channel **p;

	p = nil;
	if(strcmp(name, "event") == 0){
		p = &amp;cfrom-&gt;event;
		cfrom-&gt;wevent = 1;
	}else if(strcmp(name, "data") == 0){
		p = &amp;cfrom-&gt;data;
		cfrom-&gt;wdata = 1;
	}else
		ctlerror("%q: unknown controlwire channel %s", cfrom-&gt;name, name);
	chanfree(*p);
	*p = chan;
}

void
_ctlfocus(Control *me, int set)
{
	Controlset *cs;

	cs = me-&gt;controlset;
	if(set){
		if(cs-&gt;focus == me)
			return;
		if(cs-&gt;focus != nil)
			_ctlprint(cs-&gt;focus, "focus 0");
		cs-&gt;focus = me;
	}else{
		if(cs-&gt;focus != me)
			return;
		cs-&gt;focus = nil;
	}
}

static void
resizethread(void *v)
{
	Controlset *cs;
	char buf[64];
	Alt alts[3];

	cs = v;
	snprint(buf, sizeof buf, "resizethread0x%p", cs);
	threadsetname(buf);

	alts[0].c = cs-&gt;resizec;
	alts[0].v = nil;
	alts[0].op = CHANRCV;
	alts[1].c = cs-&gt;resizeexitc;
	alts[1].v = nil;
	alts[1].op = CHANRCV;
	alts[2].op = CHANEND;

	for(;;){
		switch(alt(alts)){
		case 0:
			resizecontrolset(cs);
			break;
		case 1:
			return;
		}
	}
}

void
activate(Control *a)
{
	Control *c;

	for(c=a-&gt;controlset-&gt;actives; c; c=c-&gt;nextactive)
		if(c == a)
			ctlerror("%q already active\n", a-&gt;name);
	
	if (a-&gt;activate){
		a-&gt;activate(a, 1);
		return;
	}
	/* prepend */
	a-&gt;nextactive = a-&gt;controlset-&gt;actives;
	a-&gt;controlset-&gt;actives = a;
}

void
deactivate(Control *a)
{
	Control *c, *prev;

	/* if group, first deactivate kids, then self */
	if (a-&gt;activate){
		a-&gt;activate(a, 0);
		return;
	}
	prev = nil;
	for(c=a-&gt;controlset-&gt;actives; c; c=c-&gt;nextactive){
		if(c == a){
			if(a-&gt;controlset-&gt;focus == a)
				a-&gt;controlset-&gt;focus = nil;
			if(prev != nil)
				prev-&gt;nextactive = a-&gt;nextactive;
			else
				a-&gt;controlset-&gt;actives = a-&gt;nextactive;
			return;
		}
		prev = c;
	}
	ctlerror("%q not active\n", a-&gt;name);
}

static struct
{
	char	*name;
	ulong	color;
}coltab[] = {
	"red",			DRed,
	"green",			DGreen,
	"blue",			DBlue,
	"cyan",			DCyan,
	"magenta",		DMagenta,
	"yellow",			DYellow,
	"paleyellow",		DPaleyellow,
	"darkyellow",		DDarkyellow,
	"darkgreen",		DDarkgreen,
	"palegreen",		DPalegreen,
	"medgreen",		DMedgreen,
	"darkblue",		DDarkblue,
	"palebluegreen",	DPalebluegreen,
	"paleblue",		DPaleblue,
	"bluegreen",		DBluegreen,
	"greygreen",		DGreygreen,
	"palegreygreen",	DPalegreygreen,
	"yellowgreen",		DYellowgreen,
	"medblue",		DMedblue,
	"greyblue",		DGreyblue,
	"palegreyblue",		DPalegreyblue,
	"purpleblue",		DPurpleblue,
	nil,	0
};

void
initcontrols(void)
{
	int i;
	Image *im;

	quotefmtinstall();
	namectlimage(display-&gt;opaque, "opaque");
	namectlimage(display-&gt;transparent, "transparent");
	namectlimage(display-&gt;white, "white");
	namectlimage(display-&gt;black, "black");
	for(i=0; coltab[i].name!=nil; i++){
		im = allocimage(display, Rect(0,0,1,1), RGB24, 1, coltab[i].color);
		namectlimage(im, coltab[i].name);
	}
	namectlfont(font, "font");
	_ctlsnarffd = open("/dev/snarf", OREAD);
}

Controlset*
newcontrolset(Image *im, Channel *kbdc, Channel *mousec, Channel *resizec)
{
	Controlset *cs;

	if(im == nil)
		im = screen;
	if((mousec==nil &amp;&amp; resizec!=nil) || (mousec!=nil &amp;&amp; resizec==nil))
		ctlerror("must specify either or both of mouse and resize channels");

	cs = ctlmalloc(sizeof(Controlset));
	cs-&gt;screen = im;

	if(kbdc == nil){
		cs-&gt;keyboardctl = initkeyboard(nil);
		if(cs-&gt;keyboardctl == nil)
			ctlerror("can't initialize keyboard: %r");
		kbdc = cs-&gt;keyboardctl-&gt;c;
	}
	cs -&gt;kbdc = kbdc;

	if(mousec == nil){
		cs-&gt;mousectl = initmouse(nil, im);
		if(cs-&gt;mousectl == nil)
			ctlerror("can't initialize mouse: %r");
		mousec = cs-&gt;mousectl-&gt;c;
		resizec = cs-&gt;mousectl-&gt;resizec;
	}
	cs-&gt;mousec = mousec;
	cs-&gt;resizec = resizec;
	cs-&gt;ctl = chancreate(sizeof(char*), 64);	/* buffer to prevent deadlock */
	cs-&gt;data = chancreate(sizeof(char*), 0);
	cs-&gt;resizeexitc = chancreate(sizeof(int), 0);
	cs-&gt;csexitc = chancreate(sizeof(int), 0);

	threadcreate(resizethread, cs, 32*1024);
	threadcreate(controlsetthread, cs, 32*1024);

	controlset = ctlrealloc(controlset, (ncontrolset+1)*sizeof(Controlset*));
	controlset[ncontrolset++] = cs;
	return cs;
}

void
closecontrolset(Controlset *cs)
{
	int i;

	sendul(cs-&gt;resizeexitc, 0);
	chanfree(cs-&gt;resizeexitc);
	sendul(cs-&gt;csexitc, 0);
	chanfree(cs-&gt;csexitc);
	chanfree(cs-&gt;ctl);
	chanfree(cs-&gt;data);

	for(i=0; i&lt;ncontrolset; i++)
		if(cs == controlset[i]){
			memmove(controlset+i, controlset+i+1, (ncontrolset-(i+1))*sizeof(Controlset*));
			ncontrolset--;
			goto Found;
		}

	if(i == ncontrolset)
		ctlerror("closecontrolset: control set not found");

    Found:
	while(cs-&gt;controls != nil)
		closecontrol(cs-&gt;controls);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

