<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/cwfs/net.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/cwfs/net.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/* network i/o */

#include "all.h"
#include "io.h"
#include &lt;fcall.h&gt;		/* 9p2000 */
#include &lt;thread.h&gt;

enum {
	Maxfdata	= 8192,
	Nqueue		= 200,		/* queue size (tunable) */

	Netclosed	= 0,		/* Connection state */
	Netopen,
};

/*
 * the kernel file server read packets directly from
 * its ethernet(s) and did all the protocol processing.
 * if the incoming packets were 9p (over il/ip), they
 * were queued for the server processes to operate upon.
 *
 * in user mode, we have one process per incoming connection
 * instead, and those processes get just the data, minus
 * tcp and ip headers, so they just see a stream of 9p messages,
 * which they then queue for the server processes.
 *
 * there used to be more queueing (in the kernel), with separate
 * processes for ethernet input, il input, 9p processing, il output
 * and ethernet output, and queues connecting them.  we now let
 * the kernel's network queues, protocol stacks and processes do
 * much of this work.
 *
 * partly as a result of this, we can now process 9p messages
 * transported via tcp, exploit multiple x86 processors, and
 * were able to shed 70% of the file server's source, by line count.
 *
 * the upshot is that Ether (now Network) is no longer a perfect fit for
 * the way network i/o is done now.  the notion of `connection'
 * is being introduced to complement it.
 */

typedef struct Network Network;
typedef struct Netconn Netconn;
typedef struct Conn9p Conn9p;

/* a network, not necessarily an ethernet */
struct Network {
	int	ctlrno;
	char	iname[NAMELEN];
	char	oname[NAMELEN];

	char	*dialstr;
	char	anndir[40];
	char	lisdir[40];
	int	annfd;			/* fd from announce */
};

/* an open tcp (or other transport) connection */
struct Netconn {
	Queue*	reply;		/* network output */
	char*	raddr;		/* remote caller's addr */
	Chan*	chan;		/* list of tcp channels */

	int	alloc;		/* flag: allocated */

	int	state;
	Conn9p*	conn9p;		/* not reference-counted */

	Lock;
};

/*
 * incoming 9P network connection from a given machine.
 * typically will multiplex 9P sessions for multiple users.
 */
struct Conn9p {
	QLock;
	Ref;
	int	fd;
	char*	dir;
	Netconn*netconn;		/* cross-connection */
	char*	raddr;
};

static Network netif[Maxnets];
static struct {
	Lock;
	Chan*	chan;
} netchans;
static Queue *netoq;		/* only one network output queue is needed */

char *annstrs[Maxnets] = {
	"tcp!*!9fs",
};

/* never returns nil */
static Chan*
getchan(Conn9p *conn9p)
{
	Netconn *netconn;
	Chan *cp, *xcp;

	lock(&amp;netchans);

	/* look for conn9p's Chan */
	xcp = nil;
	for(cp = netchans.chan; cp; cp = netconn-&gt;chan) {
		netconn = cp-&gt;pdata;
		if(!netconn-&gt;alloc)
			xcp = cp;		/* remember free Chan */
		else if(netconn-&gt;raddr != nil &amp;&amp;
		    strcmp(conn9p-&gt;raddr, netconn-&gt;raddr) == 0) {
			unlock(&amp;netchans);
			return cp;		/* found conn9p's Chan */
		}
	}

	/* conn9p's Chan not found; if no free Chan, allocate &amp; fill in one */
	cp = xcp;
	if(cp == nil) {
		cp = fs_chaninit(Devnet, 1, sizeof(Netconn));
		netconn = cp-&gt;pdata;
		netconn-&gt;chan = netchans.chan;
		netconn-&gt;state = Netopen;	/* a guess */
		/* cross-connect netconn and conn9p */
		netconn-&gt;conn9p = conn9p;	/* not reference-counted */
		conn9p-&gt;netconn = netconn;
		netchans.chan = cp;
	}

	/* fill in Chan's netconn */
	netconn = cp-&gt;pdata;
	netconn-&gt;raddr = strdup(conn9p-&gt;raddr);

	/* fill in Chan */
	cp-&gt;send = serveq;
	if (cp-&gt;reply == nil)
		cp-&gt;reply = netoq;
	netconn-&gt;reply = netoq;
	cp-&gt;protocol = nil;
	cp-&gt;msize = 0;
	cp-&gt;whotime = 0;
	strncpy(cp-&gt;whochan, conn9p-&gt;raddr, sizeof cp-&gt;whochan);
//	cp-&gt;whoprint = tcpwhoprint;
	netconn-&gt;alloc = 1;

	unlock(&amp;netchans);
	return cp;
}

static char *
fd2name(int fd)
{
	char data[128];

	if (fd2path(fd, data, sizeof data) &lt; 0)
		return strdup("/GOK");
	return strdup(data);
}

static void
hangupdfd(int dfd)
{
	int ctlfd;
	char *end, *data;

	data = fd2name(dfd);
	close(dfd);

	end = strstr(data, "/data");
	if (end != nil)
		strcpy(end, "/ctl");
	ctlfd = open(data, OWRITE);
	if (ctlfd &gt;= 0) {
		hangup(ctlfd);
		close(ctlfd);
	}
	free(data);
}

void
closechan(int n)
{
	Chan *cp;

	for(cp = chans; cp; cp = cp-&gt;next)
		if(cp-&gt;whotime != 0 &amp;&amp; cp-&gt;chan == n)
			fileinit(cp);
}

void
nethangup(Chan *cp, char *msg, int dolock)
{
	Netconn *netconn;

	netconn = cp-&gt;pdata;
	netconn-&gt;state = Netclosed;

	if(msg != nil)
		print("hangup! %s %s\n", msg, netconn-&gt;raddr);

	fileinit(cp);
	cp-&gt;whotime = 0;
	strcpy(cp-&gt;whoname, "&lt;none&gt;");

	if(dolock)
		lock(&amp;netchans);
	netconn-&gt;alloc = 0;
	free(netconn-&gt;raddr);
	netconn-&gt;raddr = nil;
	if(dolock)
		unlock(&amp;netchans);
}

void
chanhangup(Chan *cp, char *msg, int dolock)
{
	Netconn *netconn = cp-&gt;pdata;
	Conn9p *conn9p = netconn-&gt;conn9p;

	if (conn9p-&gt;fd &gt; 0)
		hangupdfd(conn9p-&gt;fd);	/* drop it */
	nethangup(cp, msg, dolock);
}

/*
 * returns length of next 9p message (including the length) and
 * leaves it in the first few bytes of abuf.
 */
static long
size9pmsg(int fd, void *abuf, uint n)
{
	int m;
	uchar *buf = abuf;

	if (n &lt; BIT32SZ)
		return -1;	/* caller screwed up */

	/* read count */
	m = readn(fd, buf, BIT32SZ);
	if(m != BIT32SZ){
		if(m &lt; 0)
			return -1;
		return 0;
	}
	return GBIT32(buf);
}

static int
readalloc9pmsg(int fd, Msgbuf **mbp)
{
	int m, len;
	uchar lenbuf[BIT32SZ];
	Msgbuf *mb;

	*mbp = nil;
	len = size9pmsg(fd, lenbuf, BIT32SZ);
	if (len &lt;= 0)
		return len;
	if(len &lt;= BIT32SZ || len &gt; IOHDRSZ+Maxfdata){
		werrstr("bad length in 9P2000 message header");
		return -1;
	}
	if ((mb = mballoc(len, nil, Mbeth1)) == nil)
		panic("readalloc9pmsg: mballoc failed");
	*mbp = mb;
	memmove(mb-&gt;data, lenbuf, BIT32SZ);
	len -= BIT32SZ;
	m = readn(fd, mb-&gt;data+BIT32SZ, len);
	if(m &lt; len)
		return 0;
	return BIT32SZ+m;
}

static void
connection(void *v)
{
	int n;
	char buf[64];
	Chan *chan9p;
	Conn9p *conn9p = v;
	Msgbuf *mb;
	NetConnInfo *nci;

	incref(conn9p);			/* count connections */
	nci = getnetconninfo(conn9p-&gt;dir, conn9p-&gt;fd);
	if (nci == nil)
		panic("connection: getnetconninfo(%s, %d) failed",
			conn9p-&gt;dir, conn9p-&gt;fd);
	conn9p-&gt;raddr = nci-&gt;raddr;

	chan9p = getchan(conn9p);
	print("new connection on %s pid %d from %s\n",
		conn9p-&gt;dir, getpid(), conn9p-&gt;raddr);

	/*
	 * reading from a pipe or a network device
	 * will give an error after a few eof reads.
	 * however, we cannot tell the difference
	 * between a zero-length read and an interrupt
	 * on the processes writing to us,
	 * so we wait for the error.
	 */
	while (conn9p-&gt;fd &gt; 0 &amp;&amp; (n = readalloc9pmsg(conn9p-&gt;fd, &amp;mb)) &gt;= 0) {
		if(n == 0)
			continue;
		mb-&gt;param = (uintptr)conn9p;	/* has fd for replies */
		mb-&gt;chan = chan9p;

		assert(mb-&gt;magic == Mbmagic);
		incref(conn9p);			/* &amp; count packets in flight */
		fs_send(serveq, mb);		/* to 9P server processes */
		/* mb will be freed by receiving process */
	}

	rerrstr(buf, sizeof buf);

	qlock(conn9p);
	print("connection hung up from %s\n", conn9p-&gt;dir);
	if (conn9p-&gt;fd &gt; 0)		/* not poisoned yet? */
		hangupdfd(conn9p-&gt;fd);	/* poison the fd */

	nethangup(chan9p, "remote hung up", 1);
	closechan(chan9p-&gt;chan);

	conn9p-&gt;fd = -1;		/* poison conn9p */
	if (decref(conn9p) == 0) {	/* last conn.? turn the lights off */
		free(conn9p-&gt;dir);
		qunlock(conn9p);
		free(conn9p);
	} else
		qunlock(conn9p);

	freenetconninfo(nci);

	if(buf[0] == '\0' || strstr(buf, "hungup") != nil)
		exits("");
	sysfatal("mount read, pid %d", getpid());
}

static void
neti(void *v)
{
	int lisfd, accfd;
	Network *net;
	Conn9p *conn9p;

	net = v;
	print("net%di\n", net-&gt;ctlrno);
	for(;;) {
		lisfd = listen(net-&gt;anndir, net-&gt;lisdir);
		if (lisfd &lt; 0) {
			print("listen %s failed: %r\n", net-&gt;anndir);
			continue;
		}

		/* got new call on lisfd */
		accfd = accept(lisfd, net-&gt;lisdir);
		if (accfd &lt; 0) {
			print("accept %d (from %s) failed: %r\n",
				lisfd, net-&gt;lisdir);
			continue;
		}

		/* accepted that call */
		conn9p = malloc(sizeof *conn9p);
		conn9p-&gt;dir = strdup(net-&gt;lisdir);
		conn9p-&gt;fd = accfd;
		newproc(connection, conn9p, smprint("9P read %s", conn9p-&gt;dir));
		close(lisfd);
	}
}

/* only need one of these for all network connections, thus all interfaces */
static void
neto(void *)
{
	int len, datafd;
	Msgbuf *mb;
	Conn9p *conn9p;

	print("neto\n");
	for(;;) {
		/* receive 9P answer from 9P server processes */
		while((mb = fs_recv(netoq, 0)) == nil)
			continue;

		if(mb-&gt;data == nil) {
			print("neto: pkt nil cat=%d free=%d\n",
				mb-&gt;category, mb-&gt;flags&amp;FREE);
			if(!(mb-&gt;flags &amp; FREE))
				mbfree(mb);
			continue;
		}

		/* send answer back over the network connection in the reply */
		len = mb-&gt;count;
		conn9p = (Conn9p *)mb-&gt;param;
		assert(conn9p);

		qlock(conn9p);
		datafd = conn9p-&gt;fd;
		assert(len &gt;= 0);
		/* datafd &lt; 0 probably indicates poisoning by the read side */
		if (datafd &lt; 0 || write(datafd, mb-&gt;data, len) != len) {
			print( "network write error (%r);");
			print(" closing connection for %s\n", conn9p-&gt;dir);
			nethangup(getchan(conn9p), "network write error", 1);
			if (datafd &gt; 0)
				hangupdfd(datafd);	/* drop it */
			conn9p-&gt;fd = -1;		/* poison conn9p */
		}
		mbfree(mb);
		if (decref(conn9p) == 0)
			panic("neto: zero ref count");
		qunlock(conn9p);
	}
}

void
netstart(void)
{
	int netorun = 0;
	Network *net;

	if(netoq == nil)
		netoq = newqueue(Nqueue, "network reply");
	for(net = &amp;netif[0]; net &lt; &amp;netif[Maxnets]; net++){
		if(net-&gt;dialstr == nil)
			continue;
		sprint(net-&gt;oname, "neto");
		if (netorun++ == 0)
			newproc(neto, nil, net-&gt;oname);
		sprint(net-&gt;iname, "net%di", net-&gt;ctlrno);
		newproc(neti, net, net-&gt;iname);
	}
}

void
netinit(void)
{
	Network *net;

	for (net = netif; net &lt; netif + Maxnets; net++) {
		net-&gt;dialstr = annstrs[net - netif];
		if (net-&gt;dialstr == nil)
			continue;
		net-&gt;annfd = announce(net-&gt;dialstr, net-&gt;anndir);
		/* /bin/service/tcp564 may already have grabbed the port */
		if (net-&gt;annfd &lt; 0)
			sysfatal("can't announce %s: %r", net-&gt;dialstr);
		print("netinit: announced on %s\n", net-&gt;dialstr);
	}
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

